
/*
 *	BIRD -- IP Addresses et Cetera for IPv6
 *
 *	(c) 1999--2000 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#ifndef _BIRD_IPV6_H_
#define _BIRD_IPV6_H_

#include <sys/types.h>
#include <netinet/in.h>
#include "lib/string.h"
#include "lib/bitops.h"
#include "lib/unaligned.h"

typedef struct ipv6_addr {
  u32 addr[4];
} ip6_addr;

struct prefix6 {
  ip6_addr addr;
  unsigned int len;
};

#define _MI6(a,b,c,d) ((ip6_addr) {{ a, b, c, d }})
#define _I0(a) ((a).addr[0])
#define _I1(a) ((a).addr[1])
#define _I2(a) ((a).addr[2])
#define _I3(a) ((a).addr[3])

#define MAX_PREFIX_LENGTH6 128
#define BITS_PER_ip6_addrESS6 128
#define STD_ADDRESS_P_LENGTH6 39
#define SIZE_OF_IP_HEADER6 40

#define IPA6_NONE _MI6(0,0,0,0)

#define ipa6_equal(x,y)  ({ ip6_addr _a=(x), _b=(y); \
			   _I0(_a) == _I0(_b) && \
			   _I1(_a) == _I1(_b) && \
			   _I2(_a) == _I2(_b) && \
			   _I3(_a) == _I3(_b); })
#define ipa6_nonzero(x) ({ ip6_addr _a=(x); (_I0(_a) || _I1(_a) || _I2(_a) || _I3(_a)); })
#define ipa6_and(x,y) ({ ip6_addr _a=(x), _b=(y); \
		     _MI6(_I0(_a) & _I0(_b), \
			 _I1(_a) & _I1(_b), \
			 _I2(_a) & _I2(_b), \
			 _I3(_a) & _I3(_b)); })
#define ipa6_or(x,y)  ({ ip6_addr _a=(x), _b=(y); \
		     _MI6(_I0(_a) | _I0(_b), \
			 _I1(_a) | _I1(_b), \
			 _I2(_a) | _I2(_b), \
			 _I3(_a) | _I3(_b)); })
#define ipa6_xor(x,y) ({ ip6_addr _a=(x), _b=(y); \
		     _MI6(_I0(_a) ^ _I0(_b), \
			 _I1(_a) ^ _I1(_b), \
			 _I2(_a) ^ _I2(_b), \
			 _I3(_a) ^ _I3(_b)); })
#define ipa6_not(x) ({ ip6_addr _a=(x); _MI6(~_I0(_a),~_I1(_a),~_I2(_a),~_I3(_a)); })
#define ipa6_mkmask(x) ipv6_mkmask(x)
#define ipa6_mklen(x) ipv6_mklen(&(x))
#define ipa6_hash(x) ipv6_hash(&(x))
#define ipa6_hton(x) ipv6_hton(&(x))
#define ipa6_ntoh(x) ipv6_ntoh(&(x))
#define ipa6_classify(x) ipv6_classify(&(x))
#define ipa6_has_link_scope(x) ipv6_has_link_scope(&(x))
#define ipa6_opposite_m1(x) ({ ip6_addr _a=(x); _MI6(_I0(_a),_I1(_a),_I2(_a),_I3(_a) ^ 1); })
#define ipa6_opposite_m2(x) ({ ip6_addr _a=(x); _MI6(_I0(_a),_I1(_a),_I2(_a),_I3(_a) ^ 3); })
/* ipa_class_mask don't make sense with IPv6 */
/* ipa_from_u32 and ipa_to_u32 replaced by ipa_build */
#define ipa6_build(a,b,c,d) _MI6(a,b,c,d)
#define ipa6_compare(x,y) ipv6_compare(x,y)
/* ipa_pxlen() requires that x != y */
#define ipa6_pxlen(x, y) ipv6_pxlen(x, y)
#define ipa6_getbit(x, y) ipv6_getbit(x, y)
#define ipa6_put_addr(x, y) ipv6_put_addr(x, y)
#define ipa6_absolutize(x,y) ipv6_absolutize(x,y)

#define ipa6_in_net(x,n,p) (!ipa6_nonzero(ipa6_and(ipa6_xor((n),(x)),ipa6_mkmask(p))))
#define ip6_is_prefix(a,l) (!ipa6_nonzero(ipa6_and(a, ipa6_not(ipa6_mkmask(l)))))
#define ipa6_zero(x) (!ipa6_nonzero(x))

/* In IPv6, SOCK_RAW does not return packet header */
#define ip6_skip_header(x, y) x

ip6_addr ipv6_mkmask(unsigned);
unsigned ipv6_mklen(ip6_addr *);
int ipv6_classify(ip6_addr *);
void ipv6_hton(ip6_addr *);
void ipv6_ntoh(ip6_addr *);
int ipv6_compare(ip6_addr, ip6_addr);
int ipv6_pton(char *a, ip6_addr *o);
int ipv4_pton_u32(char *, u32 *);
void ipv6_absolutize(ip6_addr *, ip6_addr *);

static inline int ipv6_has_link_scope(ip6_addr *a)
{
  return ((a->addr[0] & 0xffc00000) == 0xfe800000);
}

/*
 *  This hash function looks well, but once IPv6 enters
 *  mainstream use, we need to check that it has good
 *  distribution properties on real routing tables.
 */

static inline unsigned ipv6_hash(ip6_addr *a)
{
  /* Returns a 16-bit hash key */
  u32 x = _I0(*a) ^ _I1(*a) ^ _I2(*a) ^ _I3(*a);
  return (x ^ (x >> 16) ^ (x >> 8)) & 0xffff;
}

static inline u32 ipv6_getbit(ip6_addr a, u32 y)
{
  return a.addr[y / 32] & (0x80000000 >> (y % 32));
}

static inline u32 ipv6_pxlen(ip6_addr a, ip6_addr b)
{
  int i = 0;
  i+= (a.addr[i] == b.addr[i]);
  i+= (a.addr[i] == b.addr[i]);
  i+= (a.addr[i] == b.addr[i]);
  i+= (a.addr[i] == b.addr[i]);
  return 32 * i + 31 - u32_log2(a.addr[i] ^ b.addr[i]);
}

static inline byte * ipv6_put_addr(byte *buf, ip6_addr a)
{
  put_u32(buf+0,  _I0(a));
  put_u32(buf+4,  _I1(a));
  put_u32(buf+8,  _I2(a));
  put_u32(buf+12, _I3(a));
  return buf+16;
}

/*
 *  RFC 1883 defines packet precendece, but RFC 2460 replaces it
 *  by generic Traffic Class ID with no defined semantics. Better
 *  not use it yet.
 */
#define IP6_PREC_INTERNET_CONTROL -1

#endif
