/*
 *	BIRD -- BGP Packet Processing
 *
 *	(c) 2000 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#undef LOCAL_DEBUG

#include "nest/bird.h"
#include "nest/iface.h"
#include "nest/protocol.h"
#include "nest/route.h"
#include "nest/attrs.h"
#include "nest/mrtdump.h"
#include "conf/conf.h"
#include "lib/unaligned.h"
#include "lib/socket.h"

#include "nest/cli.h"

#include "bgp.h"

static void bgp_attach_next_hop(ea_list **eal, u16 bgp_family, u16 rt_family, byte *addrdata, int addrlen);
static byte * bgp_create_mp_update(struct bgp_conn *conn, u16 bgp_family, byte *buf);
static int bgp_check_family(u32 afi_safi, u16 *bgp_family, u32 family_mask, int nh_len);

static int bgp_parse_mp_list(struct bgp_conn *conn, byte *x, int len, u16 bgp_family, rta *a, byte *nh_data, int nh_len);

static struct rate_limit rl_rcv_update,  rl_snd_update;

/* bgp_family mappings to {rt_family, bits on the wire, has_label }*/
struct bgp_afi_map bgp_afi_map[] = {
  { RT_IPV4, BGP_AF_IPV4, BGP_SAFI_UNICAST, 0, 0, { bgp_parse_mp_list } },		/* BGP_PF_IPV4 */
  { RT_IPV6, BGP_AF_IPV6, BGP_SAFI_UNICAST, 0, 0, { bgp_parse_mp_list } },		/* BGP_PF_IPV6 */
  { RT_VPNV4, BGP_AF_IPV4, BGP_SAFI_VPN, 1, 1, { bgp_parse_mp_list } },		/* BGP_PF_VPNV4 */
  { RT_VPNV6, BGP_AF_IPV6, BGP_SAFI_VPN, 1, 1, { bgp_parse_mp_list } },		/* BGP_PF_VPNV6 */
  { RT_IPV4, BGP_AF_IPV4, BGP_SAFI_LUNICAST, 1, 0, { bgp_parse_mp_list } },	/* BGP_PF_LIPV4 */
  { RT_IPV6, BGP_AF_IPV6, BGP_SAFI_LUNICAST, 1, 0, { bgp_parse_mp_list } },	/* BGP_PF_LIPV6 */
};

/*
 * MRT Dump format is not semantically specified.
 * We will use these values in appropriate fields:
 *
 * Local AS, Remote AS - configured AS numbers for given BGP instance.
 * Local IP, Remote IP - IP addresses of the TCP connection (0 if no connection)
 *
 * We dump two kinds of MRT messages: STATE_CHANGE (for BGP state
 * changes) and MESSAGE (for received BGP messages).
 *
 * STATE_CHANGE uses always AS4 variant, but MESSAGE uses AS4 variant
 * only when AS4 session is established and even in that case MESSAGE
 * does not use AS4 variant for initial OPEN message. This strange
 * behavior is here for compatibility with Quagga and Bgpdump,
 */

static byte *
mrt_put_bgp4_hdr(byte *buf, struct bgp_conn *conn, int as4)
{
  struct bgp_proto *p = conn->bgp;

  if (as4)
    {
      put_u32(buf+0, p->remote_as);
      put_u32(buf+4, p->local_as);
      buf+=8;
    }
  else
    {
      put_u16(buf+0, (p->remote_as <= 0xFFFF) ? p->remote_as : AS_TRANS);
      put_u16(buf+2, (p->local_as <= 0xFFFF)  ? p->local_as  : AS_TRANS);
      buf+=4;
    }

  put_u16(buf+0, p->neigh ? p->neigh->iface->index : 0);
  put_u16(buf+2, BGP_AF);
  buf+=4;
  buf = ipa_put_addr(buf, conn->sk ? conn->sk->daddr : IPA_NONE);
  buf = ipa_put_addr(buf, conn->sk ? conn->sk->saddr : IPA_NONE);

  return buf;
}

static void
mrt_dump_bgp_packet(struct bgp_conn *conn, byte *pkt, unsigned len)
{
  byte buf[BGP_MAX_PACKET_LENGTH + 128];
  byte *bp = buf + MRTDUMP_HDR_LENGTH;
  int as4 = conn->bgp->as4_session;

  bp = mrt_put_bgp4_hdr(bp, conn, as4);
  memcpy(bp, pkt, len);
  bp += len;
  mrt_dump_message(&conn->bgp->p, BGP4MP, as4 ? BGP4MP_MESSAGE_AS4 : BGP4MP_MESSAGE,
		   buf, bp-buf);
}

static inline u16
convert_state(unsigned state)
{
  /* Convert state from our BS_* values to values used in MRTDump */
  return (state == BS_CLOSE) ? 1 : state + 1;
}

void
mrt_dump_bgp_state_change(struct bgp_conn *conn, unsigned old, unsigned new)
{
  byte buf[128];
  byte *bp = buf + MRTDUMP_HDR_LENGTH;

  bp = mrt_put_bgp4_hdr(bp, conn, 1);
  put_u16(bp+0, convert_state(old));
  put_u16(bp+2, convert_state(new));
  bp += 4;
  mrt_dump_message(&conn->bgp->p, BGP4MP, BGP4MP_STATE_CHANGE_AS4, buf, bp-buf);
}

static byte *
bgp_create_notification(struct bgp_conn *conn, byte *buf)
{
  struct bgp_proto *p = conn->bgp;

  BGP_TRACE(D_PACKETS, "Sending NOTIFICATION(code=%d.%d)", conn->notify_code, conn->notify_subcode);
  buf[0] = conn->notify_code;
  buf[1] = conn->notify_subcode;
  memcpy(buf+2, conn->notify_data, conn->notify_size);
  return buf + 2 + conn->notify_size;
}

static byte *
bgp_put_cap_mp(struct bgp_conn *conn UNUSED, byte *buf, u16 bgp_family)
{
  *buf++ = 1;		/* Capability 1: Multiprotocol extensions */
  *buf++ = 4;		/* Capability data length */
  put_u16(buf, bgp_afi_map[bgp_family].afi);
  buf += 2;
  *buf++ = 0;		/* RFU */
  *buf++ = bgp_afi_map[bgp_family].safi;	/* and SAFI */
  return buf;
}

static byte *
bgp_put_cap_rr(struct bgp_conn *conn UNUSED, byte *buf)
{
  *buf++ = 2;		/* Capability 2: Support for route refresh */
  *buf++ = 0;		/* Capability data length */
  return buf;
}

static byte *
bgp_put_cap_as4(struct bgp_conn *conn, byte *buf)
{
  *buf++ = 65;		/* Capability 65: Support for 4-octet AS number */
  *buf++ = 4;		/* Capability data length */
  put_u32(buf, conn->bgp->local_as);
  return buf + 4;
}

static byte *
bgp_create_open(struct bgp_conn *conn, byte *buf)
{
  struct bgp_proto *p = conn->bgp;
  byte *cap;
  int cap_len;
  u32 fmask, i, b;

  BGP_TRACE(D_PACKETS, "Sending OPEN(ver=%d,as=%d,hold=%d,id=%08x)",
	    BGP_VERSION, p->local_as, p->cf->hold_time, p->local_id);
  buf[0] = BGP_VERSION;
  put_u16(buf+1, (p->local_as < 0xFFFF) ? p->local_as : AS_TRANS);
  put_u16(buf+3, p->cf->hold_time);
  put_u32(buf+5, p->local_id);

  if (conn->start_state == BSS_CONNECT_NOCAP)
    {
      BGP_TRACE(D_PACKETS, "Skipping capabilities");
      buf[9] = 0;
      return buf + 10;
    }

  /* Skipped 3 B for length field and Capabilities parameter header */
  cap = buf + 12;

  BGP_TRACE(D_PACKETS, "MP-BGP capability flags: %X", conn->want_families);

  for (fmask = conn->want_families, b = 1, i = 0; b <= fmask; b *= 2, i++)
  {
    if (!(fmask & b))
      continue;

    fmask &= ~b;
    /* All families are known to us (since peer caps is ANDede with our caps) */
    DBG("Sending bgp_family %d AFI/SAFI %d/%d\n", i, bgp_afi_map[i].afi, bgp_afi_map[i].safi);
    cap = bgp_put_cap_mp(conn, cap, i);
  }

  if (p->cf->enable_refresh)
    cap = bgp_put_cap_rr(conn, cap);

  if (conn->want_as4_support)
    cap = bgp_put_cap_as4(conn, cap);

  cap_len = cap - buf - 12;
  if (cap_len > 0)
    {
      buf[9]  = cap_len + 2;	/* Optional params len */
      buf[10] = 2;		/* Option: Capability list */
      buf[11] = cap_len;	/* Option length */
      return cap;
    }
  else
    {
      buf[9] = 0;		/* No optional parameters */
      return buf + 10;
    }
}

/**
 * bgp_encode_prefixes - encode bucket of prefixes in given bgp_family into network foemat
 * @p: protocol instance
 * @w: target buffer
 * @bgp_family: family we are encoding into
 * @bam: pointer to bgp_afi_map record
 * @has_label: true if we have to prepend label before prefix
 * @buck: bucket to encode
 * @remains: buffer size
 *
 */
static unsigned int
bgp_encode_prefixes(struct bgp_proto *p, byte *w, u16 bgp_family, u16 has_label, struct bgp_afi_map *bam, struct bgp_bucket *buck, unsigned int remains)
{
  byte *start, *xlen;
  int bytes, rt_size, i, plen;
  u32 label;
  struct bgp_prefix *px;
  struct bgp_vpn_prefix *pvx;

  /*
   * Prefix length encoding:
   * 1) prefix length in bits for 'simple' AFs
   *  +---------------------------+
   *  |   Length (1 octet)        |
   *  +---------------------------+
   *  |   Prefix (variable)       |
   *  +---------------------------+
   *
   * 2) prefix length + label length for VPN/Labeled unicast families
   *  +---------------------------+
   *  |   Length (1 octet)        |
   *  +---------------------------+
   *  |   Label (3 octets)        |
   *  +---------------------------+
   *  .............................
   *  +---------------------------+
   *  |   Prefix (variable)       |
   *  +---------------------------+
   *
   */

  if (EMPTY_LIST(buck->prefixes))
    return 0;

  start = w;
  /* rt_addsize may be bigger than real size required */
  rt_size = rt_addrsize(bam->rt_family) + 1;
  if (has_label)
    rt_size += 3 * BGP_MAX_LABELS; /* Add labelstack info */

  while (!EMPTY_LIST(buck->prefixes) && (remains >= rt_size))
    {
      px = SKIP_BACK(struct bgp_prefix, bucket_node, HEAD(buck->prefixes));
      DBG("\tDequeued route %F\n", &px->n);
      /* Check if we need to prepend label before prefix */
      if (has_label)
      {
	pvx = (struct bgp_vpn_prefix *)px;
	/* Put label stack first */
	xlen = w++;
	plen = 0;

	if (has_label == 2)
	  {
	    /* Label withdraw. Send special label */
	    DBG(" * Writing withdraw label\n");
	    put_u16(w, 0x8000);
	    w += 2;
	    *w++ = '\0';
	    plen += 3 * 8;
	  }
	else
	  {
	    for (i = 0; i < BGP_MAX_LABELS; i++)
	      {
		/* 
		 * RFC 3107, sec. 3
		 * The Label field carries one or more labels (that corresponds to
		 * the stack of labels [MPLS-ENCAPS]).  Each label is encoded as 3
	    	 * octets, where the high-order 20 bits contain the label value,
    		 * and the low order bit contains "Bottom of Stack" (as defined in
		 */
		label = pvx->labels[i];
		DBG(" * Writing label %d%s\n", (label & 0xFFFFF), ((label >> 24) & 0x1) ? " [EOS]" : "");
		put_u16(w, (label & 0xFFFFF) >> 4);
		w += 2;
		*w++ = ((label >> 24) & 0x1) | ((label & 0xF) << 4);
		plen += 3 * 8;
		/* Check if end of stack is encoded */
		if ((label >> 24) & 0x1)
		  break;
	      }
	  }

	if (bam->has_rd)
	  {
	    /* Note pxlength field is sum of 1) real prefix length, 2) RD length, 3) labelstack length */
	    *xlen = px->n.pxlen + 8 * 8 + plen;
	    /* Get _address_ bytes to copy. Note we have to copy address WITH RD */
	    bytes = (px->n.pxlen + 7) / 8 + 8;
	  }
	else
	  {
	    *xlen = px->n.pxlen + plen;
	    bytes = (px->n.pxlen + 7) / 8;
	  }
	/* Count bytes we've spent on labels */
	remains -= plen / 8;
      }
      else
      {
	/* IPv6 or other similar AF */
	*w++ = px->n.pxlen;
	/* Get _address_ bytes to copy. */
	bytes = (px->n.pxlen + 7) / 8;
      }
      
      /* 
       * Write 'full' address to buffer (we must have space for it since while() guarantees this)
       * increase buffer pointer by bytes we really needs to store (e.g. 2 bytes for 10.1.0.0/16 )
       */
      put_addr(w, bam->rt_family, FPREFIX(&px->n));
      w += bytes;
      remains -= bytes + 1;
      rem_node(&px->bucket_node);
      fib_delete(&p->bfd[bgp_family].prefix_fib, px);
    }
  return w - start;
}

static void
bgp_flush_prefixes(struct bgp_proto *p, struct bgp_bucket *buck, u16 bgp_family)
{
  while (!EMPTY_LIST(buck->prefixes))
    {
      struct bgp_prefix *px = SKIP_BACK(struct bgp_prefix, bucket_node, HEAD(buck->prefixes));
      log(L_ERR "%s: - route %F skipped", p->p.name, &px->n);
      rem_node(&px->bucket_node);
      fib_delete(&p->bfd[bgp_family].prefix_fib, px);
    }
}

static byte *
bgp_create_update(struct bgp_conn *conn, byte *buf)
{
  struct bgp_proto *p = conn->bgp;
  struct bgp_family_data *bfd;
  struct bgp_bucket *buck;
  int remains = BGP_MAX_PACKET_LENGTH - BGP_HEADER_LENGTH - 4;
  byte *w;
  int wd_size = 0;
  int r_size = 0;
  int a_size = 0;

  bfd = &p->bfd[BGP_PF_IPV4];

  w = buf+2;
  if ((buck = bfd->withdraw_bucket) && !EMPTY_LIST(buck->prefixes))
    {
      DBG("Withdrawn routes:\n");
      wd_size = bgp_encode_prefixes(p, w, BGP_PF_IPV4, 0, &bgp_afi_map[BGP_PF_IPV4], buck, remains);
      w += wd_size;
      remains -= wd_size;
    }
  put_u16(buf, wd_size);

  if (remains >= 3072)
    {
      while ((buck = (struct bgp_bucket *) HEAD(bfd->bucket_queue))->send_node.next)
	{
	  if (EMPTY_LIST(buck->prefixes))
	    {
	      DBG("Deleting empty bucket %p\n", buck);
	      rem_node(&buck->send_node);
	      bgp_free_bucket(p, bfd, buck);
	      continue;
	    }

	  DBG("Processing bucket %p\n", buck);
	  a_size = bgp_encode_attrs(p, BGP_PF_IPV4, w+2, buck->eattrs, 2048);

	  if (a_size < 0)
	    {
	      log(L_ERR "%s: Attribute list too long, skipping corresponding routes", p->p.name);
	      bgp_flush_prefixes(p, buck, BGP_PF_IPV4);
	      rem_node(&buck->send_node);
	      bgp_free_bucket(p, bfd, buck);
	      continue;
	    }

	  put_u16(w, a_size);
	  w += a_size + 2;
	  r_size = bgp_encode_prefixes(p, w, BGP_PF_IPV4, 0, &bgp_afi_map[BGP_PF_IPV4], buck, remains - a_size);
	  w += r_size;
	  break;
	}
    }
  if (!a_size)				/* Attributes not already encoded */
    {
      put_u16(w, 0);
      w += 2;
    }
  if (wd_size || r_size)
    {
      BGP_TRACE_RL(&rl_snd_update, D_PACKETS, "Sending UPDATE");
      return w;
    }
  else
    return NULL;
}


/**
 * bgp_encode_nh - write next hop address into supplied buffer
 * @p: pointer to BGP protocol
 * @bgp_family: BGP nexthop family
 * @nh: nexthop data attribuute 
 * @data: data to write to
 * @datalen: buffer data length
 * @nh_len: written next hop length
 *
 */
static int
bgp_encode_nh(struct bgp_proto *p, u16 bgp_family, eattr *nh, byte *data, int datalen, int *nh_len)
{
  ip4_addr *v4;
  int rlen;
  ip6_addr *v6;
#ifdef IPV6
  int second;
  ip6_addr *sv6, ip_ll;
  neighbor *n;
#endif
  DBG("Encoding next-hop for BGP family %d\n", bgp_family);

  switch (bgp_family)
  {
    case BGP_PF_VPNV4:
      if (datalen < 12)
        return -1;
  
      v4 = (ip4_addr *)nh->u.ptr->data;
      /* Write empty RD */
      put_u32(data, 0);
      put_u32(data + 4, 0);
      /* Write IPv4 address itself */
      put_u32(data + 8, *v4);
      *nh_len = 12;
      break;

    case BGP_PF_LIPV6:
    case BGP_PF_VPNV6:
      if (!(nh->flags & BAF_IPV4MAPPED))
	{
	  log(L_ERR "Non IPv4-mapped nexthop, skipping");
	  return -1;
	}

      v6 = (ip6_addr *)nh->u.ptr->data;
      rlen = (bgp_family == BGP_PF_VPNV6) ? 8 + 16 : 16;
      *nh_len = rlen;
      /* Zero RD data */
      memset(data, 0, rlen);
      put_addr((bgp_family == BGP_PF_VPNV6) ? data + 8 : data, RT_IPV6, v6);
      break;
#ifdef IPV6
    case BGP_PF_IPV6:
      v6 = (ip6_addr *)nh->u.ptr->data;
      rlen = 16;
      second = (nh->u.ptr->length == rlen);
      ip_ll = IPA_NONE;
    
      /* XXX: we should ensure source_address family is IPv6 */
      if (memcmp(v6, &p->source_addr, rt_addrsize(RT_IPV6)) == 0)
        ip_ll = p->local_link;
      else
        {
          /* 
  	 * If we send a route with 'third party' next hop destinated 
           * in the same interface, we should also send a link local 
           * next hop address. We use the received one (stored in the 
           * other part of BA_NEXT_HOP eattr). If we didn't received
           * it (for example it is a static route), we can't use
           * 'third party' next hop and we have to use local IP address
           * as next hop. Sending original next hop address without
           * link local address seems to be a natural way to solve that
           * problem, but it is contrary to RFC 2545 and Quagga does not
           * accept such routes.
	   *
	   * There are two cases, either we have global IP, or
	   * IPA_NONE if the neighbor is link-local. For IPA_NONE,
	   * we suppose it is on the same iface, see bgp_update_attrs().
           */
    
	  if (ipa_zero(ip) || same_iface(p, &ip))
    	{
    	  if (second && ipa_nonzero(*sv6))
    	    ip_ll = *sv6;
    	  else
    	    {
    	      switch (p->cf->missing_lladdr)
    		{
    		case MLL_SELF:
    		  v6 = &p->source_addr;
    		  ip_ll = p->local_link;
    		  break;
    		case MLL_DROP:
    		  log(L_ERR "%s: Missing link-local next hop address, skipping corresponding routes", p->p.name);
  		  *nh_len = 0;
  		  return -1;
    		case MLL_IGNORE:
    		  break;
    		}
    	    }
    	}
      }
  
      /* Write NH. Most typical situation */
      if (ipa_nonzero(ip_ll))
        {
  	if (datalen < rlen * 2)
  	  return -1;
  	*nh_len = rlen * 2;
  	put_addr(data, RT_IPV6, v6);
  	put_addr(data + rlen, RT_IPV6, &ip_ll);
        }
      else
        {
  	if (datalen < rlen)
  	  return -1;
  	*nh_len = rlen;
  	put_addr(data, RT_IPV6, v6);
        }
      break;
#endif
    default:
      /* unsupported family */
      return -1;
  }

  return 0;
}


static inline int
same_iface(struct bgp_proto *p, ip_addr *ip)
{
  neighbor *n = neigh_find(&p->p, ip, 0);
  return n && p->neigh && n->iface == p->neigh->iface;
}

static byte *
bgp_create_mp_update(struct bgp_conn *conn, u16 bgp_family, byte *buf)
{
  struct bgp_proto *p = conn->bgp;
  struct bgp_family_data *bfd;
  struct bgp_bucket *buck;
  int size, attr_len, remains, nh_len;
  u16 bgp_afi, bgp_safi;
  struct bgp_afi_map *bam;
  u8 has_label;
  byte *w, *tmp, *tstart;
  byte nh_data[64];
  static byte attr_data[2048];
  ea_list *ea;
  eattr *nh;

  remains = BGP_MAX_PACKET_LENGTH - BGP_HEADER_LENGTH - 4;

  /* bgp_family is assumed to be correct */
  bfd = &p->bfd[bgp_family];

  /* Get AFI/SAFI */
  bgp_afi = bgp_afi_map[bgp_family].afi;
  bgp_safi = bgp_afi_map[bgp_family].safi;
  bam = &bgp_afi_map[bgp_family];
  has_label = bgp_afi_map[bgp_family].has_label;

  DBG("MP-UPDATE create requested for family %d AFI/SAFI: %d/%d\n", bgp_family, bgp_afi, bgp_safi);
  /*
   *   UPDATE message header 
   *
   *   +-----------------------------------------------------+
   *   |   Withdrawn Routes Length (2 octets)                |
   *   +-----------------------------------------------------+
   *   |   Withdrawn Routes (variable)                       |
   *   +-----------------------------------------------------+
   *   |   Total Path Attribute Length (2 octets)            |
   *   +-----------------------------------------------------+
   *   |   Path Attributes (variable)                        |
   *   +-----------------------------------------------------+
   *   |   Network Layer Reachability Information (variable) |
   *   +-----------------------------------------------------+
   */

  /* IPv4 withdrawal route length */
  put_u16(buf, 0);
  /* Point w to beginning of path attributes data */
  w = buf + 4;

  /*
   *  MP_UNREACH_NLRI attribute encoding
   *
   *  +---------------------------------------------------------+
   *  | Address Family Identifier (2 octets)                    |
   *  +---------------------------------------------------------+
   *  | Subsequent Address Family Identifier (1 octet)          |
   *  +---------------------------------------------------------+
   *  | Withdrawn Routes (variable)                             |
   *  +---------------------------------------------------------+
   *
   */

  if ((buck = bfd->withdraw_bucket) && !EMPTY_LIST(buck->prefixes))
    {
      DBG("Withdrawn routes:\n");
      /* 
       * XXX: we pass attribute DATA length in bgp_attach_attr_wa().
       * Total attribute length can be up to `remain` bytes. Attribute header
       * is 3 (or 4) bytes, not 8
       */
      ea = NULL;
      /* Allocate temporary data from linpool */
      tmp = bgp_attach_attr_wa(&ea, bgp_linpool, BA_MP_UNREACH_NLRI, remains - 8);
      put_u16(tmp, bgp_afi);
      tmp += 2;
      *tmp++ = bgp_safi;
      /* Send special has_label value to indicate withdraw */
      ea->attrs[0].u.ptr->length = 3 + bgp_encode_prefixes(p, tmp, bgp_family, has_label ? 2 : 0, bam, buck, remains - 11);
      /* Write single attribute into real buffer */
      size = bgp_encode_attrs(p, bgp_family, w, ea, remains);
      ASSERT(size >= 0);
      w += size;
      remains -= size;
    }

  if (remains < 3072)
    goto done;

  while ((buck = (struct bgp_bucket *) HEAD(bfd->bucket_queue))->send_node.next)
  {
    if (EMPTY_LIST(buck->prefixes))
      {
        DBG("Deleting empty bucket %p\n", buck);
        rem_node(&buck->send_node);
        bgp_free_bucket(p, bfd, buck);
        continue;
      }

    DBG("Processing bucket %p\n", buck);

    /* Let's construct next hop */
    nh = ea_find(buck->eattrs, EA_CODE(EAP_BGP, BA_NEXT_HOP));
    ASSERT(nh);
    /* NH is saved in temporary buffer for the sake of simplicity */
    if (bgp_encode_nh(p, bgp_family, nh, nh_data, sizeof(nh_data), &nh_len) == -1)
      {
	log(L_ERR "%s: Missing link-local next hop address, skipping corresponding routes", p->p.name);
	/* Skip prefixes and free bucket */
	bgp_flush_prefixes(p, buck, bgp_family);
	rem_node(&buck->send_node);
	bgp_free_bucket(p, bfd, buck);
	continue;
      }

    /* 
     * Save path attributes to static buffer to permit MP_REACH_NLRI to be the first
     * path attribute as draft-ietf-idr-optional-transitive-03 recommends 
     */
    if ((attr_len = bgp_encode_attrs(p, bgp_family, attr_data, buck->eattrs, 2048)) < 0)
      {
        log(L_ERR "%s: Attribute list too long, skipping corresponding routes", p->p.name);
        bgp_flush_prefixes(p, buck, bgp_family);
        rem_node(&buck->send_node);
        bgp_free_bucket(p, bfd, buck);
        continue;
      }
    
    /* Valid bucket found, search stopped */
    break;
  }

  /* Check for end of list.  tail-end has NULL ->next pointer */
  if (!buck->send_node.next)
    goto done;

  /* 
   * Start preparing NLRI:
   * We already get all PATH attributes encoded and NH encoded.
   */

  /*
   *  MP_REACH_NLRI attribute encoding 
   *
   *  +---------------------------------------------------------+
   *  | Address Family Identifier (2 octets)                    |
   *  +---------------------------------------------------------+
   *  | Subsequent Address Family Identifier (1 octet)          |
   *  +---------------------------------------------------------+
   *  | Length of Next Hop Network Address (1 octet)            |
   *  +---------------------------------------------------------+
   *  | Network Address of Next Hop (variable)                  |
   *  +---------------------------------------------------------+
   *  | Reserved (1 octet)                                      |
   *  +---------------------------------------------------------+
   *  | Network Layer Reachability Information (variable)       |
   *  +---------------------------------------------------------+
   */
  ea = NULL;
  tmp = tstart = bgp_attach_attr_wa(&ea, bgp_linpool, BA_MP_REACH_NLRI, remains - 8 - attr_len);

  put_u16(tmp, bgp_afi);
  tmp += 2;
  *tmp++ = bgp_safi;
  /* Save nexthop length and copy nexthop data */
  *tmp++ = nh_len;
  memcpy(tmp, nh_data, nh_len);
  tmp += nh_len;
  /* Add reserved bit */
  *tmp++ = 0;

  tmp += bgp_encode_prefixes(p, tmp, bgp_family, has_label, bam, buck, remains - 8 - (nh_len + 5) - attr_len);
  /* Count attribute legnth */
  ea->attrs[0].u.ptr->length = tmp - tstart;
  size = bgp_encode_attrs(p, bgp_family, w, ea, remains - attr_len);
  ASSERT(size >= 0);
  w += size;
  /* Copy other path attributes */
  memcpy(w, attr_data, attr_len);
  w += attr_len;

done:
  /* Count total Path attribute length */
  size = w - (buf+4);
  put_u16(buf+2, size);
  lp_flush(bgp_linpool);
  if (size)
    {
      BGP_TRACE_RL(&rl_snd_update, D_PACKETS, "Sending UPDATE");
      return w;
    }
  else
    return NULL;
}

static byte *
bgp_create_route_refresh(struct bgp_conn *conn, u16 bgp_family, byte *buf)
{
  struct bgp_proto *p = conn->bgp;
  BGP_TRACE(D_PACKETS, "Sending ROUTE-REFRESH for AFI/SAFI %d/%d", bgp_afi_map[bgp_family].afi, bgp_afi_map[bgp_family].safi);

  put_u16(buf, bgp_afi_map[bgp_family].afi);
  buf += 2;
  *buf++ = 0;		/* RFU */
  *buf++ = bgp_afi_map[bgp_family].safi;
  return buf;
}

static void
bgp_create_header(byte *buf, unsigned int len, unsigned int type)
{
  memset(buf, 0xff, 16);		/* Marker */
  put_u16(buf+16, len);
  buf[18] = type;
}

/**
 * bgp_fire_tx - transmit packets
 * @conn: connection
 *
 * Whenever the transmit buffers of the underlying TCP connection
 * are free and we have any packets queued for sending, the socket functions
 * call bgp_fire_tx() which takes care of selecting the highest priority packet
 * queued (Notification > Keepalive > Open > Update), assembling its header
 * and body and sending it to the connection.
 */
static int
bgp_fire_tx(struct bgp_conn *conn)
{
  struct bgp_proto *p = conn->bgp;
  unsigned int s = conn->packets_to_send;
  u32 um;
  sock *sk = conn->sk;
  byte *buf, *pkt, *end;
  int i, type;

  if (!sk)
    {
      conn->packets_to_send = 0;
      return 0;
    }
  buf = sk->tbuf;
  pkt = buf + BGP_HEADER_LENGTH;

  if (s & (1 << PKT_SCHEDULE_CLOSE))
    {
      /* We can finally close connection and enter idle state */
      bgp_conn_enter_idle_state(conn);
      return 0;
    }
  if (s & (1 << PKT_NOTIFICATION))
    {
      s = 1 << PKT_SCHEDULE_CLOSE;
      type = PKT_NOTIFICATION;
      end = bgp_create_notification(conn, pkt);
    }
  else if (s & (1 << PKT_KEEPALIVE))
    {
      s &= ~(1 << PKT_KEEPALIVE);
      type = PKT_KEEPALIVE;
      end = pkt;			/* Keepalives carry no data */
      BGP_TRACE(D_PACKETS, "Sending KEEPALIVE");
      bgp_start_timer(conn->keepalive_timer, conn->keepalive_time);
    }
  else if (s & (1 << PKT_OPEN))
    {
      s &= ~(1 << PKT_OPEN);
      type = PKT_OPEN;
      end = bgp_create_open(conn, pkt);
    }
  else if (s & (1 << PKT_ROUTE_REFRESH))
    {
      type = PKT_ROUTE_REFRESH;
      /* Request AFs supported by both sides */
      conn->refresh_mask &= conn->peer_families;
      if (conn->refresh_mask == 0)
	{
	  s &= ~(1 << PKT_ROUTE_REFRESH);
	  return 0;
	}
      else
	{
	  /* Assume non-zero update mask */
	  for (i = 0, um = conn->refresh_mask; !(um & 1); i++)
	    um = um >> 1;
	  end = bgp_create_route_refresh(conn, i, pkt);
	  conn->refresh_mask &= ~(1 << i);
	  if (conn->refresh_mask == 0)
	    s &= ~(1 << PKT_ROUTE_REFRESH);
	}
    }
  else if (s & (1 << PKT_UPDATE))
    {
      /* Check update for given families. Assume _some_ bit is set */
      for (i = 0, um = conn->update_mask; !(um & 1); i++)
	um = um >> 1;
      type = PKT_UPDATE;
      end = (i == BGP_PF_IPV4) ? bgp_create_update(conn, pkt) : bgp_create_mp_update(conn, i, pkt);
      if (!end)
	{
	  /* Remove update bit from selected family */
	  conn->update_mask &= ~(1 << i);
	  /* Zero packets_to_send bitmask IFF update_mask is set to zero */
	  if (conn->update_mask == 0)
	    conn->packets_to_send = 0;
          DBG("UPDATE: Update mask=%X bgp_family=%X packets to send=%d\n", conn->update_mask, i, conn->packets_to_send);
	  return conn->packets_to_send;
	}
      DBG("UPDATE: Update mask=%X packets to send=%d\n", conn->update_mask, conn->packets_to_send);
    }
  else
    return 0;
  conn->packets_to_send = s;
  bgp_create_header(buf, end - buf, type);
  return sk_send(sk, end - buf);
}

/**
 * bgp_schedule_packet - schedule a packet for transmission
 * @conn: connection
 * @type: packet type
 *
 * Schedule a packet of type @type to be sent as soon as possible.
 */
void
bgp_schedule_packet(struct bgp_conn *conn, int type)
{
  DBG("BGP: Scheduling packet type %d\n", type);
  conn->packets_to_send |= 1 << type;
  if (conn->sk && conn->sk->tpos == conn->sk->tbuf)
    ev_schedule(conn->tx_ev);
}

void
bgp_kick_tx(void *vconn)
{
  struct bgp_conn *conn = vconn;

  DBG("BGP: kicking TX\n");
  while (bgp_fire_tx(conn) > 0)
    ;
}

void
bgp_tx(sock *sk)
{
  struct bgp_conn *conn = sk->data;

  DBG("BGP: TX hook\n");
  while (bgp_fire_tx(conn) > 0)
    ;
}

/* Capatibility negotiation as per RFC 2842 */

void
bgp_parse_capabilities(struct bgp_conn *conn, byte *opt, int len)
{
  // struct bgp_proto *p = conn->bgp;
  int cl, error = 0;
  u16 afi, safi, bgp_family;

  while (len > 0)
    {
      if (len < 2 || len < 2 + opt[1])
	goto err;
      
      cl = opt[1];

      switch (opt[0])
	{
	case 1:	/* MP-BGP capability, RFC 4760 */
	  if (cl != 4)
	    goto err;
	  afi = get_u16(opt + 2);
	  safi = *((unsigned char *)(opt + 5));
	  DBG("Found AFI/SAFI %d/%d capability\n", afi, safi);
	  /*
	   * Check if afi/safi is known to us, Use catch-all family mask
	   * to save peer capabilities
	   */
	  if (bgp_check_family(AFI_SAFI(afi, safi), &bgp_family, 0xFFFFFFFF, 0))
	    {
	      /* We don't even know this afi/safi, send notification */
	      log(L_ERR "%s: unknown AFI/SAFI %d/%d, rejecting OPEN", conn->bgp->p.name, afi, safi);
	      error = 8;
	      goto err;
	    }

	  DBG("AFI/SAFI %d/%d mapped to bgp_family %d\n", afi, safi, bgp_family);
	  /* Set connection capabilities */
	  conn->peer_families |= 1 << bgp_family;
	  break;

	case 2:	/* Route refresh capability, RFC 2918 */
	  if (cl != 0)
	    goto err;
	  conn->peer_refresh_support = 1;
	  break;

	case 65: /* AS4 capability, RFC 4893 */ 
	  if (cl != 4)
	    goto err;
	  conn->peer_as4_support = 1;
	  if (conn->want_as4_support)
	    conn->advertised_as = get_u32(opt + 2);
	  break;

	  /* We can safely ignore all other capabilities */
	}
      len -= 2 + cl;
      opt += 2 + cl;
    }

  return;

err:
  bgp_error(conn, 2, error, NULL, 0);
  return;
}

static int
bgp_parse_options(struct bgp_conn *conn, byte *opt, int len)
{
  struct bgp_proto *p = conn->bgp;
  int ol;

  while (len > 0)
    {
      if (len < 2 || len < 2 + opt[1])
	{ bgp_error(conn, 2, 0, NULL, 0); return 0; }
#ifdef LOCAL_DEBUG
      {
	int i;
	DBG("\tOption %02x:", opt[0]);
	for(i=0; i<opt[1]; i++)
	  DBG(" %02x", opt[2+i]);
	DBG("\n");
      }
#endif

      ol = opt[1];
      switch (opt[0])
	{
	case 2:
	  if (conn->start_state == BSS_CONNECT_NOCAP)
	    BGP_TRACE(D_PACKETS, "Ignoring received capabilities");
	  else
	    bgp_parse_capabilities(conn, opt + 2, ol);
	  break;

	default:
	  /*
	   *  BGP specs don't tell us to send which option
	   *  we didn't recognize, but it's common practice
	   *  to do so. Also, capability negotiation with
	   *  Cisco routers doesn't work without that.
	   */
	  bgp_error(conn, 2, 4, opt, ol);
	  return 0;
	}
      len -= 2 + ol;
      opt += 2 + ol;
    }
  return 0;
}

static void
bgp_rx_open(struct bgp_conn *conn, byte *pkt, int len)
{
  struct bgp_conn *other;
  struct bgp_proto *p = conn->bgp;
  unsigned hold;
  u16 base_as;
  u32 id, bgp_family;

  /* Check state */
  if (conn->state != BS_OPENSENT)
    { bgp_error(conn, 5, 0, NULL, 0); return; }

  /* Check message contents */
  if (len < 29 || len != 29 + pkt[28])
    { bgp_error(conn, 1, 2, pkt+16, 2); return; }
  if (pkt[19] != BGP_VERSION)
    { bgp_error(conn, 2, 1, pkt+19, 1); return; } /* RFC 1771 says 16 bits, draft-09 tells to use 8 */
  conn->advertised_as = base_as = get_u16(pkt+20);
  hold = get_u16(pkt+22);
  id = get_u32(pkt+24);
  BGP_TRACE(D_PACKETS, "Got OPEN(as=%d,hold=%d,id=%08x)", conn->advertised_as, hold, id);

  if (bgp_parse_options(conn, pkt+29, pkt[28]))
    return;

  if (hold > 0 && hold < 3)
    { bgp_error(conn, 2, 6, pkt+22, 2); return; }

  if (!id || id == 0xffffffff || id == p->local_id)
    { bgp_error(conn, 2, 3, pkt+24, -4); return; }

  if ((conn->advertised_as != base_as) && (base_as != AS_TRANS))
    log(L_WARN "%s: Peer advertised inconsistent AS numbers", p->p.name);

  if (conn->advertised_as != p->remote_as)
    {
      if (conn->peer_as4_support)
	{
	  u32 val = htonl(conn->advertised_as);
	  bgp_error(conn, 2, 2, (byte *) &val, 4);
	}
      else
	bgp_error(conn, 2, 2, pkt+20, 2);

      return;
    }

  /*
   * Check if MP capabilities are the same
   * XXX: Should we account IPv4 capability as well?
   */

  bgp_family = (conn->want_families | conn->peer_families) &
	~(conn->want_families & conn->peer_families);

  if (bgp_family & ~BGP_PF_IPV4_CAP)
    {
      /* Save wanted families to get valid error message */
      bgp_family = conn->want_families & conn->peer_families;
      conn->families = conn->want_families & ~bgp_family;
      if (conn->families)
	log(L_ERR "%s: the following remote BGP families [%s ] is not configured%s",
	    conn->bgp->p.name, bgp_show_peer_caps(conn),
	    p->cf->enforce_capabilities ? ", rejecting OPEN" : "");
      conn->families = conn->peer_families & ~bgp_family;
      if (conn->families)
	log(L_ERR "%s: the following local BGP families [%s ] is not configured%s",
	    conn->bgp->p.name, bgp_show_peer_caps(conn),
	    p->cf->enforce_capabilities ? ", rejecting OPEN" : "");

      if (p->cf->enforce_capabilities)
	{
	  bgp_error(conn, 2, 8, NULL, 0);
	  return;
	}
    }

  /* Merge common mask */
  conn->families = conn->want_families & conn->peer_families;

  /* Check the other connection */
  other = (conn == &p->outgoing_conn) ? &p->incoming_conn : &p->outgoing_conn;
  switch (other->state)
    {
    case BS_IDLE:
    case BS_CONNECT:
    case BS_ACTIVE:
    case BS_OPENSENT:
    case BS_CLOSE:
      break;
    case BS_OPENCONFIRM:
      if ((p->local_id < id) == (conn == &p->incoming_conn))
	{
	  /* Should close the other connection */
	  BGP_TRACE(D_EVENTS, "Connection collision, giving up the other connection");
	  bgp_error(other, 6, 7, NULL, 0);
	  break;
	}
      /* Fall thru */
    case BS_ESTABLISHED:
      /* Should close this connection */
      BGP_TRACE(D_EVENTS, "Connection collision, giving up this connection");
      bgp_error(conn, 6, 7, NULL, 0);
      return;
    default:
      bug("bgp_rx_open: Unknown state");
    }

  /* Update our local variables */
  conn->hold_time = MIN(hold, p->cf->hold_time);
  conn->keepalive_time = p->cf->keepalive_time ? : conn->hold_time / 3;
  p->remote_id = id;
  p->as4_session = conn->want_as4_support && conn->peer_as4_support;

  DBG("BGP: Hold timer set to %d, keepalive to %d, AS to %d, ID to %x, AS4 session to %d\n", conn->hold_time, conn->keepalive_time, p->remote_as, p->remote_id, p->as4_session);

  bgp_schedule_packet(conn, PKT_KEEPALIVE);
  bgp_start_timer(conn->hold_timer, conn->hold_time);
  bgp_conn_enter_openconfirm_state(conn);
}

static int
bgp_set_next_hop(struct bgp_proto *p, rta *a, u16 bgp_family)
{
  struct eattr *nh = ea_find(a->eattrs, EA_CODE(EAP_BGP, BA_NEXT_HOP));
  ip_addr *nexthop = (ip_addr *) nh->u.ptr->data;
  rtable *rt;

  if (p->cf->gw_mode == GW_DIRECT)
    {
      neighbor *ng = NULL;

      if (ipa_nonzero(*nexthop))
	ng = neigh_find(&p->p, nexthop, 0);
#if 0
      /* XXX: fixme, introduced in 53ffbff3 */
      else if (second)	/* GW_DIRECT -> single_hop -> p->neigh != NULL */
	ng = neigh_find2(&p->p, nexthop + 1, p->neigh->iface, 0);
#endif

      /* Fallback */
      if (!ng)
	ng = p->neigh;

      if (ng->scope == SCOPE_HOST)
	return 0;

      a->dest = RTD_ROUTER;
      a->gw = ng->addr;
      a->iface = ng->iface;
      a->hostentry = NULL;
      a->igp_metric = 0;
    }
  else /* GW_RECURSIVE */
    {
      rt = BT(p, bgp_afi_map[bgp_family].rt_family, table);
      DBG("Selected table %s for bgp AFI/SAFI %d/%d\n", rt->name, bgp_afi_map[bgp_family].afi, bgp_afi_map[bgp_family].safi);

      switch (bgp_family)
      {
	case BGP_PF_LIPV6:
	case BGP_PF_VPNV6:
		/* VPNv6 nexthop can be an IPv4-mapped address (See RFC 4659 3.2.1.2). */
		if (nh->flags & BAF_IPV4MAPPED)
		{
		  /* Nexthop is IPv4-mapped address, let's extract IPv4 address */
		  DBG("Setting v4-mapped nexthop in table %s\n", p->igp_table->name);
		  rta_set_recursive_next_hop(rt, a, p->igp_table, (ip_addr *)((byte *)nexthop + 12), (ip_addr *)((byte *)nexthop + 12));
		  break;
		}

		/* XXX: Pure IPv6 nexthops are not currently supported */
		a->dest = RTD_UNREACHABLE;
#if 0
		second = (nh->u.ptr->length == 16) ? nexthop : (ip_addr *)((byte *)nexthop + 16);
		rta_set_recursive_next_hop(rt, a, p->igp_table, nexthop, second);
#endif
		break;
	case BGP_PF_IPV4:
	case BGP_PF_VPNV4:
        	rta_set_recursive_next_hop(rt, a, p->igp_table, nexthop, nexthop);
		break;
      }
    }

  return 1;
}

#define DECODE_PREFIX(pp, ll) do {		\
  int b = *pp++;				\
  int q;					\
  ll--;						\
  if (b > BITS_PER_IP_ADDRESS) { err=10; goto done; } \
  q = (b+7) / 8;				\
  if (ll < q) { err=1; goto done; }		\
  memcpy(&prefix, pp, q);			\
  pp += q;					\
  ll -= q;					\
  ipa_ntoh(prefix);				\
  prefix = ipa_and(prefix, ipa_mkmask(b));	\
  pxlen = b;					\
} while (0)

static int
bgp_parse_mp_prefix(byte *data, int datalen, u16 bgp_family, u16 rt_family, byte *addrnew, int addrlen, int *addrmask, int has_label, byte *lbuf, unsigned int *llen)
{
  uint8_t pxlen, mask;
  int pfull, plast, lfull = 0, label, ind = 0, is_bos = 0;
  u32 *labels = (u32 *)lbuf;
  char addrdata[64];

  if (datalen <= 0)
    return -1;

  pxlen = *((uint8_t *)data++);
  DBG("PARSE-MP-PREFIX: length=%d\n", pxlen);

  /*
   * pxlen means:
   * 1) prefix length in bits for 'simple' AFs
   *  +---------------------------+
   *  |   Length (1 octet)        |
   *  +---------------------------+
   *  |   Prefix (variable)       |
   *  +---------------------------+
   *
   * 2) prefix length + label length for VPN families
   *  +---------------------------+
   *  |   Length (1 octet)        |
   *  +---------------------------+
   *  |   Label (3 octets)        |
   *  +---------------------------+
   *  .............................
   *  +---------------------------+
   *  |   Prefix (variable)       |
   *  +---------------------------+
   *
   */

  /* Extract label stack first */
  if (has_label)
  {
    while ((is_bos == 0) && (pxlen > 24) && (ind < BGP_MAX_LABELS))
    {
      is_bos = data[2] & 0x01;
      /* Encode BOS as lower bit in upper byte */
      label = (get_u16(data) << 4) + ((data[2] & 0xF0) >> 4) + (is_bos << 24);

      data += 3;
      lfull += 3;
      pxlen -= 24;

      /*
       * Check if label is 0 (Juniper sends such labels) or
       * Special withdraw value (cisco does this)
       *
       * Note RFC 3107 states 'A BGP speaker can withdraw a previously advertised route
       * ..
       * or (b) listing the NLRI of the previously advertised route in the Withdrawn Routes 
       * field of an Update message. The label information carried (as part of NLRI) in the
       * Withdrawn Routes field should be set to 0x800000'
       *
       * Since keyword 'should' is used it is possible that some implementations
       * chose to send original labelstack with the route, so we can't
       * blindly assume that stack depth in case of withdrawal is always 1 
       */
      if ((!label) || (label == 0x80000))
      {
        DBG("label: %d bos: %d pxlen: %d, forsing BOS\n", label & 0xFFFFF, is_bos, pxlen);
	is_bos = 1;
	break;
      }
      labels[ind++] = label;

      DBG("label: %d bos: %d pxlen: %d\n", label & 0xFFFFF, is_bos, pxlen);
    }

    if ((is_bos == 0) && (pxlen <= 24))
      return -1;

    /* Prepare an attribute with label */
    *llen = ind * sizeof(u32);
  }

  /* Check if enough data is available */
  if (pxlen > (addrlen << 3))
    return -1;

  DBG("New prefix, BGP=%d, pxlen=%d\n", bgp_family, pxlen);
  /* Get integer number of bytes and remaining bits */
  pfull = pxlen / 8;
  plast = pxlen % 8;

  if (datalen < pfull + (plast ? 1 : 0) + 1)
    return -1;

  /* zero addrdata */
  memset(addrdata, 0, sizeof(addrdata));

  /* Copy full data octets via memcpy */
  if (pfull > 0)
  {
    memcpy(addrdata, data, pfull);
    data += pfull;
  }

  /* Copy last octet if needed */
  if (plast)
  {
    /* Make inverted mask for bits we need (11100000 for 3)  */
    mask = ((1 << (8 - (plast % 8))) - 1);
    /* Get needed value for last octet */
    mask = *((uint8_t *)data) & ~mask;
    /* save it */
    *((uint8_t *)(addrdata + pfull)) = mask;
  }

  /* convert to host format. Padding data is zeroed in bgp_parse_mp_list() */
  get_addr(addrdata, rt_family, addrnew);

  /* Save real prefix length */
  if (addrmask)
    {
      switch (bgp_family)
	{
	case BGP_PF_VPNV4:
	case BGP_PF_VPNV6:
	  *addrmask = pxlen - (8 << 3); /* RD length in bits */
	  break;
	default:
	  *addrmask = pxlen;
	}
    }

  /* Return total (possibly labeled) prefix legnth */
  return 1 + lfull + pfull + (plast ? 1 : 0);
}

/**
 * bgp_check_family - check if given update is valid.
 * @afi_safi: afi << 16 + safi
 * @bgp_family: pointer to resulting BGP_PF_*
 * @family_mask: consiled accepted mask with all families
 * @nh_len: nexthop field length
 *
 * Returns: 0 for valid data.
 * 1: invalid nhop length
 * 2: unsupported afi/safi
 * 3: afi/safi not within family_mask
 */
static int
bgp_check_family(u32 afi_safi, u16 *bgp_family, u32 family_mask, int nh_len)
{
  int bad = 1;

  switch (afi_safi)
  {
    case AFI_SAFI(BGP_AF_IPV4, BGP_SAFI_UNICAST):
      *bgp_family = BGP_PF_IPV4;
      if ((nh_len == 0) || (nh_len == 4))
        bad = 0;
      break;

    case AFI_SAFI(BGP_AF_IPV6, BGP_SAFI_UNICAST):
      *bgp_family = BGP_PF_IPV6;
      if ((nh_len == 0) || (nh_len == 16) || (nh_len == 32))
        bad = 0;
      break;
    
    case AFI_SAFI(BGP_AF_IPV4, BGP_SAFI_LUNICAST):
      *bgp_family = BGP_PF_LIPV4;
      if ((nh_len == 0) || (nh_len == 4))
        bad = 0;
      break;

    case AFI_SAFI(BGP_AF_IPV6, BGP_SAFI_LUNICAST):
      *bgp_family = BGP_PF_LIPV6;
      if ((nh_len == 0) || (nh_len == 16))
        bad = 0;
      break;

    case AFI_SAFI(BGP_AF_IPV4, BGP_SAFI_VPN):
      *bgp_family = BGP_PF_VPNV4;
      if ((nh_len == 0) || (nh_len == 8 + 4))
        bad = 0;
      break;

    case AFI_SAFI(BGP_AF_IPV6, BGP_SAFI_VPN):
      *bgp_family = BGP_PF_VPNV6;
      if ((nh_len == 0) || (nh_len == 8 + 16))
        bad = 0;
      break;

    default:
      bad = 2;
      break;
  }

  /* Throw error if AFI/SAFI is not supported */
  if ((!bad) && (!(family_mask & (1 << *bgp_family))))
    bad = 3;

  DBG("bgp_get_family(%d,%d) -> bgp_family=%d bad=%d\n", afi_safi  >> 8, afi_safi % 256, *bgp_family, bad);

  return bad;
}

static int
bgp_parse_mp_list(struct bgp_conn *conn, byte *x, int len, u16 bgp_family, rta *a, byte *nh_data, int nh_len)
{
  struct bgp_proto *p = conn->bgp;
  int xlen, has_label, need_label, err = 0;
  u16 nh_family, rt_family, rt_size;
  int addrmask;
  rtable *rt;
  struct announce_hook *h;
  net *n;
  byte lattrs[sizeof(ea_list) + sizeof(eattr) + sizeof(struct adata) + sizeof(u32) * BGP_MAX_LABELS], *labels;
  ea_list *ea, *ea_tmp;
  struct adata *ad;
  rta *newa, *ta = NULL;
  /* sizeof(addrdata) nees to be >= sizeof(maximum supported BGP PF prefix) */
  byte addrnew[64];
#ifdef LOCAL_DEBUG
  byte addr1[64];
#endif

  DBG("bgp_parse_mp_list: bgp_family=%d len=%d\n", bgp_family, len);

  if (len <= 0)
    return 0;

  rt_family = bgp_afi_map[bgp_family].rt_family;
  has_label = bgp_afi_map[bgp_family].has_label;
  need_label = has_label;
  nh_family = bgp_family;

  rt_size = rt_addrsize(rt_family);

  /* Prepare pointers to label data */
  ea = (ea_list *)lattrs;
  ad = (struct adata *)(lattrs + sizeof(ea_list) + sizeof(eattr));
  labels = ad->data;

  /* Enforce label for non-labeled IPv6 unicast */
  if ((bgp_family == BGP_PF_IPV6) && p->cf->enforce_labeled)
  {
    need_label = 1;
    /* Prepare default static label + BOS */
    *((u32 *)labels) = MPLS_LABEL_IPV6NULL | (1 << 24);
    ad->length = sizeof(u32);
    /* Prepare v4-mapped nexthop */
    byte fake_nh[16];
    memset(fake_nh, 0, sizeof(fake_nh));
    fake_nh[10] = 0xFF;
    fake_nh[11] = 0xFF;
    put_addr(&fake_nh[12], RT_IPV4, &p->cf->remote_ip);

    nh_data = fake_nh;
    nh_len = sizeof(fake_nh);
    nh_family = BGP_PF_LIPV6;
#ifdef LOCAL_DEBUG
    addr_print(addr1, sizeof(addr1), RT_IPV6, nh_data);
      DBG("Fake nexthop set to %s\n", addr1);
#endif
  }

  if (a)
  {
    /* 
     * New routes advertising.
     * We need to create NH attribute.
     * Ipv4 nexthop may already exist in attrs 
     * (RFC doesn't forbid it) but we don't care:
     * bgp_attach_attr() adds new list with nexthop attr
     * at the beginning so ea_do_prune (called by ea_sort from rta_lookup)
     * removes 'old' nexthop attribute if it exists.
     *
     * We can get either cached or uncached rta in @a.
     * If storing MPLS labels is not required by BGP family
     * we're saving cached rta in @ta. Overwise, we make
     * cached attribute immediately after every prefix parsing.
     */

    if (a->aflags & RTAF_CACHED)
    {
      /* 
       * This should not happen since this means we've got
       * an IPv4 next hop set.
       */
      /* Save nexhthop attribute to ea_tmp */
      ea_tmp = NULL;
      bgp_attach_next_hop(&ea_tmp, nh_family, rt_family, nh_data, nh_len);
      ta = rta_copy_with_ea(a, ea_tmp);
      bgp_set_next_hop(p, ta, nh_family);
      /* Get persistent cached attributes. uc=1 */
      rta_lookup_self(ta);
    }
   else
    {
      /*
       * Attributes are not cached (a0 is passed from rx_update), let's add next hop
       */
      bgp_attach_next_hop(&a->eattrs, nh_family, rt_family, nh_data, nh_len);
      bgp_set_next_hop(p, a, nh_family);
      /* If label is not used we can make cached attribute */
      ta = need_label ? a : rta_lookup(a);
    }

    /* Prepare attribute list with single label attribute */
    if (need_label)
    {
      ea->next = NULL;
      ea->flags = 0;
      ea->count = 1;
      ea->attrs[0].id = EA_CODE(EAP_BGP, BA_MP_LABEL);
      ea->attrs[0].flags = BAF_OPTIONAL | BAF_INTERNAL;
      ea->attrs[0].type = EAF_TYPE_INT_SET;
      ea->attrs[0].u.ptr = ad;
    }
  }

  rt = BT(p, rt_family, table);
  h = BT(p, rt_family, ahook);
  DBG("Starting MP-parse. rt_family=%d total_len=%d table=%s action=%s\n", rt_family, len, rt->name, ta ? "ADD" : "WITHDRAW");
  /* Zero addrnew before use */
  memset(addrnew, 0, sizeof(addrnew));
  while (len > 0)
  {
    if ((xlen = bgp_parse_mp_prefix(x, len, bgp_family, rt_family, addrnew, rt_size, &addrmask, has_label, labels, &ad->length)) < 0)
    {
      log(L_ERR "prefix parse error, ignoring MP block");
      err = -1;
      goto done;
    }

    if (xlen == 0)
      break; /* End of list */

    len -= xlen;
    x += xlen;

    if (!ta)
    {
      /* Route withdrawal */
#ifdef LOCAL_DEBUG
      addr_print(addr1, sizeof(addr1), rt_family, addrnew);
      DBG("BGP prefix %s/%d, xlen: %d, len: %d, fam: %d\n", addr1, addrmask, xlen, len, rt_family);
#endif
      if (n = fib_find(&rt->fib, addrnew, addrmask))
	rte_update2(h, n, NULL, &p->p);

      continue;
    }

    if (need_label)
    {
      newa = rta_copy_with_ea(ta, ea);
      rta_lookup_self(newa);
      /* usage count is set to 1, so it gets freed by rte_update() if needed  */
      /* Clear EA to permit its reusage */
      ea->next = NULL;
    }
   else
    newa = rta_clone(ta);

    rte *e = rte_get_temp(newa);
    e->net = fib_get(&rt->fib, addrnew, addrmask);
    e->pflags = 0;
    e->u.bgp.suppressed = 0;
#ifdef LOCAL_DEBUG
    addr_print(addr1, sizeof(addr1), rt_family, addrnew);
    DBG("BGP prefix %s/%d, xlen: %d, len: %d, fam: %d\n", addr1, addrmask, xlen, len, rt_family);
#endif
    rte_update2(h, e->net, e, &p->p);
  }

done:
  if (ta && (ta->aflags & RTAF_CACHED))
    rta_free(ta);

  /*
   * Send optional attribute error.
   * XXX: Check x/len for validness
   */
  if (err)
    bgp_error(conn, 3, 9, x, len);

  return -1;
}


static int
bgp_do_rx_mp_update(struct bgp_conn *conn, byte *attrs, int attr_len, rta *a, byte *mp_reach, int mp_reach_len, byte *mp_unreach, int mp_unreach_len)
{
  byte *x;
  int len, nlen;
  u16 af, sub;
  u16 bgp_family;
  int res;

  DBG("MP-UPDATE started REACH_LEN=%d UNREACH_LEN=%d\n", mp_reach_len, mp_unreach_len);

  /*
   *  MP_UNREACH_NLRI attribute encoding
   *
   *  +---------------------------------------------------------+
   *  | Address Family Identifier (2 octets)                    |
   *  +---------------------------------------------------------+
   *  | Subsequent Address Family Identifier (1 octet)          |
   *  +---------------------------------------------------------+
   *  | Withdrawn Routes (variable)                             |
   *  +---------------------------------------------------------+
   *
   */
  if (mp_unreach)
  {
      len = mp_unreach_len;
      x = mp_unreach;
      if (len < 2 + 1)
	return 9;
      af = get_u16(x);
      sub = x[2];
      x += 3;
      len -= 3;
      DBG("\tNLRI AF=%d sub=%d len=%d\n", af, sub, len);

      if (len < 0)
	return 9;

      /* Check if we support this AFI/SAFI and NH length is correct */
      res = bgp_check_family((af << 8) + sub, &bgp_family, conn->families, 0);
      if (res)
      {
	if (res = 3)
	  return conn->bgp->cf->enforce_capabilities ? -1 : 0;
	return -1;
      }

      /* bgp_error() is called by RX hook if needed */
      if (bgp_afi_map[bgp_family].hook.rx(conn, x, len, bgp_family, NULL, NULL, 0) != 0)
	return 0; /* bgp_error() is already called */
  }

  /*
   *  MP_REACH_NLRI attribute encoding 
   *
   *  +---------------------------------------------------------+
   *  | Address Family Identifier (2 octets)                    |
   *  +---------------------------------------------------------+
   *  | Subsequent Address Family Identifier (1 octet)          |
   *  +---------------------------------------------------------+
   *  | Length of Next Hop Network Address (1 octet)            |
   *  +---------------------------------------------------------+
   *  | Network Address of Next Hop (variable)                  |
   *  +---------------------------------------------------------+
   *  | Reserved (1 octet)                                      |
   *  +---------------------------------------------------------+
   *  | Network Layer Reachability Information (variable)       |
   *  +---------------------------------------------------------+
   *
   */
  if (mp_reach)
  {
      len = mp_reach_len;
      x = mp_reach;
      if (len < 2 + 1 + 1)
	return 9;
      af = get_u16(x);
      sub = x[2];
      x += 3;
      len -= 4;
      DBG("\tNLRI AF=%d sub=%d len=%d\n", af, sub, len);

      /* Nexthop data follows */
      nlen = *x++;

      if (len < nlen + 1)
	return 9;

      /* Check if we support this AFI/SAFI and NH length is correct */
      res = bgp_check_family((af << 8) + sub, &bgp_family, conn->families, nlen);
      if (res)
      {
	if (res = 3)
	  return conn->bgp->cf->enforce_capabilities ? -1 : 0;
	return -1;
      }

      /* bgp_error() is called by RX hook if needed */
      if (bgp_afi_map[bgp_family].hook.rx(conn, x + nlen + 1, len - nlen - 1, bgp_family, a, x, nlen) != 0)
	return 0;
  }

  return 0;
}

static void
bgp_do_rx_update(struct bgp_conn *conn,
		 byte *withdrawn, int withdrawn_len,
		 byte *nlri, int nlri_len,
		 byte *attrs, int attr_len)
{
  struct bgp_proto *p = conn->bgp;
  net *n;
  rta *a0, *a = NULL;
  struct attr_helper ah;
  ip_addr prefix;
  int pxlen, err = 0;

  /* Withdraw routes */
  while (withdrawn_len)
    {
      DECODE_PREFIX(withdrawn, withdrawn_len);
      DBG("Withdraw %I/%d\n", prefix, pxlen);
      if (n = net_find(p->p.table, prefix, pxlen))
	rte_update(p->p.table, n, &p->p, &p->p, NULL);
    }

  DBG("UPDATE Checking: attr_len=%d nlri_len=%d\n", attr_len, nlri_len);
  /* nlri_len CAN be null for MP-BGP updates, but attribute length cannot */
  if (!attr_len && !nlri_len)		/* shortcut */
    return;

  memset(&ah, 0, sizeof(ah));
  a0 = bgp_decode_attrs(conn, attrs, attr_len, bgp_linpool, nlri_len, &ah);

  if (conn->state != BS_ESTABLISHED)	/* fatal error during decoding */
    return;

  /* Set hext hop and lookup attributes IFF nlri_len is NOT null (and nexthop presence is checked in bgp_decode_attrs() ) */
  if (a0 && nlri_len && bgp_set_next_hop(p, a0, BGP_PF_IPV4))
    a = rta_lookup(a0);

  while (nlri_len)
    {
      DECODE_PREFIX(nlri, nlri_len);
      DBG("Add %I/%d\n", prefix, pxlen);

      if (a)
	{
	  rte *e = rte_get_temp(rta_clone(a));
	  e->net = net_get(p->p.table, prefix, pxlen);
	  e->pflags = 0;
	  e->u.bgp.suppressed = 0;
	  rte_update(p->p.table, e->net, &p->p, &p->p, e);
	}
      else
	{
	  /* Forced withdraw as a result of soft error */
	  if (n = net_find(p->p.table, prefix, pxlen))
	    rte_update(p->p.table, n, &p->p, &p->p, NULL);
	}
    }

  /* Check if we need to call MP parser */
  if (ah.mp_reach || ah.mp_unreach)
    err = bgp_do_rx_mp_update(conn, attrs, attr_len, a ? a : a0, ah.mp_reach, ah.mp_reach_len, ah.mp_unreach, ah.mp_unreach_len);

done:
    if (a)
      rta_free(a);

    if (err)
      bgp_error(conn, 3, err, NULL, 0);

  return;
}

static void
bgp_attach_next_hop(ea_list **eal, u16 bgp_family, u16 rt_family, byte *addrdata, int addrlen)
{
  byte *nh;
  byte v6mask[12] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF };

  /* 
   * Nexthops are transmitted with zero rd in VPNvX families.
   * Let us store real nexthop values (e.g. skip RD prefix)
   */
  if (bgp_family == BGP_PF_VPNV4)
    addrlen = 4;
  else if (bgp_family == BGP_PF_VPNV6)
  {
    if (addrlen == 8 + 16)
      addrlen = 16;
    else
      addrlen = 32;
  }

  /* Addrlen is already checked and is valid for given family */
  nh = bgp_attach_attr_wa(eal, bgp_linpool, BA_NEXT_HOP, addrlen);

  /* zero attribute */
  memset(nh, 0, addrlen);

  switch (bgp_family)
    {
    case BGP_PF_IPV6:
      get_addr(addrdata, rt_family, nh);
      /* Set V4mapped hint if nhop is v4mapped */
      if (!memcmp(addrdata, v6mask, sizeof(v6mask)))
	(*eal)->attrs[0].flags |= BAF_IPV4MAPPED;

      /* IPv6-related hack with link-local addresses */
      if (addrlen == 16 * 2)
	get_addr(addrdata + 16, RT_IPV6, nh + 16);
      break;

    case BGP_PF_VPNV4:
      get_addr(addrdata + 8, RT_IPV4, nh);
      break;

    case BGP_PF_LIPV6:
    case BGP_PF_VPNV6:
      if (bgp_family == BGP_PF_VPNV6)
	addrdata += 8; /* Skip RD */

      get_addr(addrdata, RT_IPV6, nh);
      /* Set V4mapped hint if nhop is v4mapped */
      if (!memcmp(addrdata, v6mask, sizeof(v6mask)))
	(*eal)->attrs[0].flags |= BAF_IPV4MAPPED;

      break;

    default:
      get_addr(addrdata, rt_family, nh);
  }
}


static void
bgp_rx_update(struct bgp_conn *conn, byte *pkt, int len)
{
  struct bgp_proto *p = conn->bgp;
  byte *withdrawn, *attrs, *nlri;
  int withdrawn_len, attr_len, nlri_len;

  BGP_TRACE_RL(&rl_rcv_update, D_PACKETS, "Got UPDATE");

  /* Workaround for some BGP implementations that skip initial KEEPALIVE */
  if (conn->state == BS_OPENCONFIRM)
    bgp_conn_enter_established_state(conn);

  if (conn->state != BS_ESTABLISHED)
    { bgp_error(conn, 5, 0, NULL, 0); return; }
  bgp_start_timer(conn->hold_timer, conn->hold_time);

  /* Find parts of the packet and check sizes */
  if (len < 23)
    {
      bgp_error(conn, 1, 2, pkt+16, 2);
      return;
    }
  withdrawn = pkt + 21;
  withdrawn_len = get_u16(pkt + 19);
  if (withdrawn_len + 23 > len)
    goto malformed;
  attrs = withdrawn + withdrawn_len + 2;
  attr_len = get_u16(attrs - 2);
  if (withdrawn_len + attr_len + 23 > len)
    goto malformed;
  nlri = attrs + attr_len;
  nlri_len = len - withdrawn_len - attr_len - 23;
  if (!attr_len && nlri_len)
    goto malformed;
  DBG("Sizes: withdrawn=%d, attrs=%d, NLRI=%d\n", withdrawn_len, attr_len, nlri_len);

  lp_flush(bgp_linpool);

  bgp_do_rx_update(conn, withdrawn, withdrawn_len, nlri, nlri_len, attrs, attr_len);
  return;

malformed:
  bgp_error(conn, 3, 1, NULL, 0);
}

static struct {
  byte major, minor;
  byte *msg;
} bgp_msg_table[] = {
  { 1, 0, "Invalid message header" },
  { 1, 1, "Connection not synchronized" },
  { 1, 2, "Bad message length" },
  { 1, 3, "Bad message type" },
  { 2, 0, "Invalid OPEN message" },
  { 2, 1, "Unsupported version number" },
  { 2, 2, "Bad peer AS" },
  { 2, 3, "Bad BGP identifier" },
  { 2, 4, "Unsupported optional parameter" },
  { 2, 5, "Authentication failure" },
  { 2, 6, "Unacceptable hold time" },
  { 2, 7, "Required capability missing" }, /* [RFC3392] */
  { 2, 8, "No supported AFI/SAFI" }, /* This error msg is nonstandard */
  { 3, 0, "Invalid UPDATE message" },
  { 3, 1, "Malformed attribute list" },
  { 3, 2, "Unrecognized well-known attribute" },
  { 3, 3, "Missing mandatory attribute" },
  { 3, 4, "Invalid attribute flags" },
  { 3, 5, "Invalid attribute length" },
  { 3, 6, "Invalid ORIGIN attribute" },
  { 3, 7, "AS routing loop" },		/* Deprecated */
  { 3, 8, "Invalid NEXT_HOP attribute" },
  { 3, 9, "Optional attribute error" },
  { 3, 10, "Invalid network field" },
  { 3, 11, "Malformed AS_PATH" },
  { 4, 0, "Hold timer expired" },
  { 5, 0, "Finite state machine error" },
  { 6, 0, "Cease" }, /* Subcodes are according to [RFC4486] */
  { 6, 1, "Maximum number of prefixes reached" },
  { 6, 2, "Administrative shutdown" },
  { 6, 3, "Peer de-configured" },
  { 6, 4, "Administrative reset" },
  { 6, 5, "Connection rejected" },
  { 6, 6, "Other configuration change" },
  { 6, 7, "Connection collision resolution" },
  { 6, 8, "Out of Resources" }
};

/**
 * bgp_error_dsc - return BGP error description
 * @code: BGP error code
 * @subcode: BGP error subcode
 *
 * bgp_error_dsc() returns error description for BGP errors
 * which might be static string or given temporary buffer.
 */
const char *
bgp_error_dsc(unsigned code, unsigned subcode)
{
  static char buff[32];
  unsigned i;
  for (i=0; i < ARRAY_SIZE(bgp_msg_table); i++)
    if (bgp_msg_table[i].major == code && bgp_msg_table[i].minor == subcode)
      {
	return bgp_msg_table[i].msg;
      }

  bsprintf(buff, "Unknown error %d.%d", code, subcode);
  return buff;
}

void
bgp_log_error(struct bgp_proto *p, u8 class, char *msg, unsigned code, unsigned subcode, byte *data, unsigned len)
{
  const byte *name;
  byte *t, argbuf[36];
  unsigned i;

  /* Don't report Cease messages generated by myself */
  if (code == 6 && class == BE_BGP_TX)
    return;

  name = bgp_error_dsc(code, subcode);
  t = argbuf;
  if (len)
    {
      *t++ = ':';
      *t++ = ' ';

      if ((code == 2) && (subcode == 2) && ((len == 2) || (len == 4)))
	{
	  /* Bad peer AS - we would like to print the AS */
	  t += bsprintf(t, "%d", (len == 2) ? get_u16(data) : get_u32(data));
	  goto done;
	}
      if (len > 16)
	len = 16;
      for (i=0; i<len; i++)
	t += bsprintf(t, "%02x", data[i]);
    }
 done:
  *t = 0;
  log(L_REMOTE "%s: %s: %s%s", p->p.name, msg, name, argbuf);
}

static void
bgp_rx_notification(struct bgp_conn *conn, byte *pkt, int len)
{
  struct bgp_proto *p = conn->bgp;
  if (len < 21)
    {
      bgp_error(conn, 1, 2, pkt+16, 2);
      return;
    }

  unsigned code = pkt[19];
  unsigned subcode = pkt[20];
  int err = (code != 6);

  bgp_log_error(p, BE_BGP_RX, "Received", code, subcode, pkt+21, len-21);
  bgp_store_error(p, conn, BE_BGP_RX, (code << 16) | subcode);

#ifndef IPV6
  if ((code == 2) && ((subcode == 4) || (subcode == 7))
      /* Error related to capability:
       * 4 - Peer does not support capabilities at all.
       * 7 - Peer request some capability. Strange unless it is IPv6 only peer.
       */
      && (p->cf->capabilities == 2)
      /* Capabilities are not explicitly enabled or disabled, therefore heuristic is used */
      && (conn->start_state == BSS_CONNECT)
      /* Failed connection attempt have used capabilities */
      && (p->cf->remote_as <= 0xFFFF))
      /* Not possible with disabled capabilities */
    {
      /* We try connect without capabilities */
      log(L_WARN "%s: Capability related error received, retry with capabilities disabled", p->p.name);
      p->start_state = BSS_CONNECT_NOCAP;
      err = 0;
    }
#endif

  bgp_conn_enter_close_state(conn);
  bgp_schedule_packet(conn, PKT_SCHEDULE_CLOSE);

  if (err) 
    {
      bgp_update_startup_delay(p);
      bgp_stop(p, 0);
    }
}

static void
bgp_rx_keepalive(struct bgp_conn *conn)
{
  struct bgp_proto *p = conn->bgp;

  BGP_TRACE(D_PACKETS, "Got KEEPALIVE");
  bgp_start_timer(conn->hold_timer, conn->hold_time);
  switch (conn->state)
    {
    case BS_OPENCONFIRM:
      bgp_conn_enter_established_state(conn);
      break;
    case BS_ESTABLISHED:
      break;
    default:
      bgp_error(conn, 5, 0, NULL, 0);
    }
}

static void
bgp_rx_route_refresh(struct bgp_conn *conn, byte *pkt, int len)
{
  struct bgp_proto *p = conn->bgp;

  BGP_TRACE(D_PACKETS, "Got ROUTE-REFRESH");

  if (conn->state != BS_ESTABLISHED)
    { bgp_error(conn, 5, 0, NULL, 0); return; }

  if (!p->cf->enable_refresh)
    { bgp_error(conn, 1, 3, pkt+18, 1); return; }

  if (len != (BGP_HEADER_LENGTH + 4))
    { bgp_error(conn, 1, 2, pkt+16, 2); return; }

  /* FIXME - we ignore AFI/SAFI values, as we support
     just one value and even an error code for an invalid
     request is not defined */

  proto_request_feeding(&p->p);
}


/**
 * bgp_rx_packet - handle a received packet
 * @conn: BGP connection
 * @pkt: start of the packet
 * @len: packet size
 *
 * bgp_rx_packet() takes a newly received packet and calls the corresponding
 * packet handler according to the packet type.
 */
static void
bgp_rx_packet(struct bgp_conn *conn, byte *pkt, unsigned len)
{
  byte type = pkt[18];

  DBG("BGP: Got packet %02x (%d bytes)\n", type, len);

  if (conn->bgp->p.mrtdump & MD_MESSAGES)
    mrt_dump_bgp_packet(conn, pkt, len);

  switch (type)
    {
    case PKT_OPEN:		return bgp_rx_open(conn, pkt, len);
    case PKT_UPDATE:		return bgp_rx_update(conn, pkt, len);
    case PKT_NOTIFICATION:      return bgp_rx_notification(conn, pkt, len);
    case PKT_KEEPALIVE:		return bgp_rx_keepalive(conn);
    case PKT_ROUTE_REFRESH:	return bgp_rx_route_refresh(conn, pkt, len);
    default:			bgp_error(conn, 1, 3, pkt+18, 1);
    }
}

/**
 * bgp_rx - handle received data
 * @sk: socket
 * @size: amount of data received
 *
 * bgp_rx() is called by the socket layer whenever new data arrive from
 * the underlying TCP connection. It assembles the data fragments to packets,
 * checks their headers and framing and passes complete packets to
 * bgp_rx_packet().
 */
int
bgp_rx(sock *sk, int size)
{
  struct bgp_conn *conn = sk->data;
  byte *pkt_start = sk->rbuf;
  byte *end = pkt_start + size;
  unsigned i, len;

  DBG("BGP: RX hook: Got %d bytes\n", size);
  while (end >= pkt_start + BGP_HEADER_LENGTH)
    {
      if ((conn->state == BS_CLOSE) || (conn->sk != sk))
	return 0;
      for(i=0; i<16; i++)
	if (pkt_start[i] != 0xff)
	  {
	    bgp_error(conn, 1, 1, NULL, 0);
	    break;
	  }
      len = get_u16(pkt_start+16);
      if (len < BGP_HEADER_LENGTH || len > BGP_MAX_PACKET_LENGTH)
	{
	  bgp_error(conn, 1, 2, pkt_start+16, 2);
	  break;
	}
      if (end < pkt_start + len)
	break;
      bgp_rx_packet(conn, pkt_start, len);
      pkt_start += len;
    }
  if (pkt_start != sk->rbuf)
    {
      memmove(sk->rbuf, pkt_start, end - pkt_start);
      sk->rpos = sk->rbuf + (end - pkt_start);
    }
  return 0;
}
