/*
 *	BIRD -- Layer3 VPN Protocol Configuration
 *
 *	(c) 2011 Yandex, LLC
 *      Author: Alexander V. Chernikov <melifaro@yandex-team.ru>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

/**
 * DOC: l3vpn
 *
 * The l3vpn protocol works very much like pipe. 
 * It connects to VPNvX routing table (filled by RR bgp session) and local
 * 'client' one sending changes from one to other. update filters have to be
 * specified. Usually filtering is done via 'rt' attributes.
 *
 * To avoid loops, l3vpn keeps a `being updated' flag in each routing
 * table. l3vpn protocol is not intended to work with pipe protocol.
 */

#define LOCAL_DEBUG

#include "nest/bird.h"
#include "nest/iface.h"
#include "nest/protocol.h"
#include "nest/route.h"
#include "conf/conf.h"
#include "filter/filter.h"
#include "lib/string.h"
#include "lib/alloca.h"
#include "lib/krt.h"
#include "lib/krt_mpls.h"

#include "proto/bgp/bgp.h"
#include "l3vpn.h"

static void l3vpn_close_lib(struct l3vpn_proto *p);
static int l3vpn_open_lib(struct l3vpn_proto *p);
static void l3vpn_rel_notify(struct proto *provider, struct proto *requester, struct service_reply *repl);

static u32
l3vpn_get_label(struct l3vpn_proto *p, net *n UNUSED)
{
  return p->label;
}

/**
 * l3vpn_write_labelstack - writes MPLS labelstack for L3vpn route
 * @me: pointer to destination labelstack data
 * @bgp_label: pointer to eattr with BA_MP_LABEL data
 * @nhop_label: pointer to eattr with LA_MPLSENTRY. Ca be NULL
 *
 */
static void
l3vpn_write_labelstack(struct bird_mpls_entry *me, eattr *bgp_label, eattr *nhop_label)
{
  struct bird_mpls_entry *nme = NULL;
  struct adata *ad;
  int i, j = 0;
  u32 *label;

  memset(me, 0, sizeof(struct bird_mpls_entry));

  /* Nexthop label can be NULL (LDP is not ready) */
  if (nhop_label != NULL)
  {
    ad = nhop_label->u.ptr;
    nme = (struct bird_mpls_entry *)(ad + 1);
    for (i = 0; i < MPLS_STACK_DEPTH; i++, j++)
    {
      me->stack[j] = nme->stack[i] & 0xFFFFF;
      if (nme->stack[i] & MPLS_STACK_BOS)
	break;
    }
  }

  /* BGP label can not be NULL */
  ad = bgp_label->u.ptr;
  for (label = (u32 *)(ad + 1); j < MPLS_STACK_DEPTH; label++, j++)
  {
    me->stack[j] = *label;
    if (*label & MPLS_STACK_BOS)
      break;
  }

  /* BOS already written by us, let's write the rest fields */
  me->action = KMPLS_OP_PUSH;
  me->depth = j + 1;
}

static void
l3vpn_rt_notify(struct proto *P, rtable *src_table, net *n, rte *new, rte *old, ea_list *attrs)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;
  rtable *dest;
  int order = (src_table == P->table) ? PE_TO_VPN : VPN_TO_PE;
  struct announce_hook *h;
  struct proto_stats *stats;
  struct proto *src;
  ea_list *tmpa, *eal;
  eattr *ea;
  eattr *bgp_label, *nhop_label;
  struct adata *ad;
  struct hostentry *he;
  char buf[32];

  vpn4_addr vpn;
  void *addr;

  net *nn;
  rte *e;
  rta *newa;

  if (!new && !old)
    return;

  /* Do address conversion */
  if (order == VPN_TO_PE)
  {
    memcpy(&vpn, FPREFIX(&n->n), src_table->fib.addr_size);
    /* Search for IP address */
    addr = &vpn.addr;
    h = P->main_ahook;
    stats = &P->stats;
  }
  else
  {
    /* PE -> VPN */
    vpn.rd = p->rd;
    memcpy(&vpn.addr, FPREFIX(&n->n), src_table->fib.addr_size);
    /* Search for VPN address */
    addr = &vpn;
    h = p->vpn_ahook;
    stats = &p->vpn_stats;
  }

  dest = h->table;

  if (dest->pipe_busy)
  {
    log(L_ERR "l3vpn loop detected when sending %F to table %s", &n->n, dest->name);
    return;
  }

  addr_print(buf, sizeof(buf), dest->fib.addr_type, addr);
  log(L_ERR "Converting %F -> %s/%d", &n->n, buf, n->n.pxlen);

  nn = fib_get(&dest->fib, addr, n->n.pxlen);

  if (new)
  {
    /* 
     * Regadless of direction we have to add 1 attribute:
     * for PE_TO_VPN we're adding BA_MP_LABEL
     * for VPN_TO_PE we're adding LA_MPLSENTRY
     */
    tmpa = alloca(sizeof(ea_list) + sizeof(eattr));
    memset(tmpa, 0, sizeof(ea_list) + sizeof(eattr));
    tmpa->count = 1;

    ea = &tmpa->attrs[0];

    /* Setup external storage for attribute */
    ad = alloca(sizeof(struct adata) + MAX(sizeof(struct bird_mpls_entry), sizeof(u32)));
    ea->u.ptr = ad;

   /* 
    * Client route, we should pass all base/extended attributes (which are not filtered out)
    * Filtering attributes (like BGP communities passed from client) is a filter task
    * Add MPLS label if needed 
    */
    if (order == PE_TO_VPN)
    {
      /* Prepare BA_MP_LABEL attribute data */
      ad->length = sizeof(u32);
      /* Some BGP internals have to be used here */
      ea->id = EA_CODE(EAP_BGP, BA_MP_LABEL);
      ea->type = EAF_TYPE_INT_SET;
      ea->flags = BAF_OPTIONAL;

      /* Let's write label */
      u32 *label = (u32 *)ad->data;
      *label = l3vpn_get_label(p, n) | MPLS_STACK_BOS;
    }
   else /* VPN_TO_PE */
    {
      ad->length = sizeof(struct bird_mpls_entry);
      memset(ad->data, 0, sizeof(struct bird_mpls_entry));
      /* Some LDP internals have to be used here */
      ea->id = EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL);
      ea->type = EAF_TEMP | EAF_TYPE_OPAQUE | EAF_ORIGINATED;

      /* 
       * Nexthop resolution is already done in BGP code (this is mandatory to 
       * make best path algorithm to work). We got valid nexthop address in
       * new->attrs and original route attributes in new->attrs->hostentry->source
       *
       * Nexthop update is handled in BGP code, too. We get route update event,
       * so there is no need to make any recursive resolving here
       */

      /*
       * Find mandatory BA_MP_LABEL attribute in VPNvX route attribute
       */
      if ((bgp_label = ea_find(attrs, EA_CODE(EAP_BGP, BA_MP_LABEL))) == NULL)
      {
        DBG("No BA_MP_LABEL in converted VPNvX route (%F), skipping\n", &n->n);
        return;
      }

      if ((he = new->attrs->hostentry) == NULL)
      {
	DBG("Nexthop for route %F is not recursive\n", &n->n);
	nhop_label = NULL;
      }
     else
      {
	if ((nhop_label = ea_find(he->src->eattrs, EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL))) == NULL)
	{
	  /* 
	   * This can be perfectly valid situation - if this nexthop is directly connected to
	   * us and sends Ipmplicit null for its loopback address.
	   */
	  DBG(" No labelstack attribute for route to %F via %I, this can possibly cause blackhole\n", &n->n, he->addr);
	}
      }
    
      l3vpn_write_labelstack((struct bird_mpls_entry *)ad->data, bgp_label, nhop_label);
    }

    /* Make new attributes */
    eal = new->attrs->eattrs;
    /* Copy with (possibly) modified eattrs */
    new->attrs->eattrs = attrs;
    newa = rta_copy_with_ea(new->attrs, tmpa);
    /* Put original pointer back  */
    new->attrs->eattrs = eal;
    newa->aflags = 0;
    newa->hostentry = NULL;

    e = rte_get_temp(newa);
    e->net = nn;
    e->pflags = 0;

    /* Copy protocol specific embedded attributes. */
    memcpy(&(e->u), &(new->u), sizeof(e->u));
    e->pref = new->pref;
    e->pflags = new->pflags;

    src = new->attrs->proto;
   /* Finally, do attribute lookup */
   rta_lookup_self(newa);
  }
  else
  {
    e = NULL;
    src = old->attrs->proto;
  }

  src_table->pipe_busy = 1;
  rte_update2(h, nn, e, &p->p);
  src_table->pipe_busy = 0;
}

static int
l3vpn_import_control(struct proto *P, rte **ee, ea_list **ea UNUSED, struct linpool *p UNUSED)
{
  struct proto *pp = (*ee)->sender->proto;

  if (pp == P)
    return -1;	/* Avoid local loops automatically */

  return 0;
}

static int
l3vpn_reload_routes(struct proto *P)
{
  /*
   * Because the l3vpn protocol feeds routes from both routing tables
   * together, both directions are reloaded during refeed and 'reload
   * out' command works like 'reload' command. For symmetry, we also
   * request refeed when 'reload in' command is used.
   */
  proto_request_feeding(P);
  return 1;
}

/**
 * l3vpn_run - starts instance, connect to routing tables
 * @p: pointer to protocol instance
 *
 * protocol state is not changed, proto_notify_state()
 * needs to be called
 */
static void
l3vpn_run(struct l3vpn_proto *p)
{
  struct proto *P = &p->p;
  struct announce_hook *a;

  DBG("l3vpn_run() for %s\n", P->name);

  /* Connect protocol to the main table */
  a = proto_add_announce_hook(P, P->table, &P->stats);
  P->main_ahook = a;

  /* Connect the protocol to vpn table. */
  a = proto_add_announce_hook(P, p->vpn_table, &p->vpn_stats);
  p->vpn_ahook = a;
}

#if 0
static void
l3vpn_announce_label(struct l3vpn_proto *p, u32 label, struct iface *iface, ip_addr *nexthop)
{
  rta *a, A;
  net *n;
  rte *e;


      DBG("dev_if_notify: %s:%I going up\n", ad->iface->name, ad->ip);
      bzero(&A, sizeof(A));
      A.proto = p;
      A.source = RTS_DEVICE;
      A.scope = SCOPE_UNIVERSE;
      A.cast = RTC_UNICAST;
      A.dest = RTD_DEVICE;
      A.iface = ad->iface;
      A.eattrs = NULL;
      a = rta_lookup(&A);
      n = net_get(p->table, ad->prefix, ad->pxlen);
      e = rte_get_temp(a);
      e->net = n;
      e->pflags = 0;
      rte_update(p->table, n, p, p, e);
  /* Prepare attribute */
  bzero(&A, sizeof(A));
  A.proto = &p->proto;
  A.source = RTS_DEVICE;
  A.scope = SCOPE_UNIVERSE;
  A.cast = RTC_UNICAST;
  A.dest = nexthop ? RTD_ROUTER : RTD_DEVICE;
  A.flags = 0;
  //A.gw = re->nexthop;
  //A.iface = iface;

  /* Prepare label extended attribute */
  
  /* Allocate temporary attribute for LA_MPLSENTRY */	
  tmpa = alloca(sizeof(ea_list) + sizeof(eattr));
  memset(tmpa, 0, sizeof(ea_list) + sizeof(eattr));
  tmpa->count = 1;

  /* Setup external storage for sockaddr  */
  ad = alloca(sizeof(struct adata) + sizeof(struct bird_mpls_entry));
  ad->length = sizeof(struct bird_mpls_entry);
  m = (bird_mpls_entry *)(ad + 1);
  
  /* fill in valid label structure */
  m->local_label = label;
  m->action = KMPLS_OP_L3VPN;


	if (re->me.stack[0] == MPLS_LABEL_IMPLNULL)
	{
		m->depth = 0;
		m->action = KMPLS_OP_POP;
	}
	else
	{
		m->depth = 1;
		m->action = KMPLS_OP_SWAP;
	}

	/* Finish attribute setup */
	ea = (eattr *)(tmpa + 1);
	ea->id = EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL);
	ea->type = EAF_TEMP | EAF_TYPE_OPAQUE | EAF_ORIGINATED;
	ea->u.ptr = ad;

	A.eattrs = tmpa;

	a = rta_lookup(&A);

	/* Get new prefix */
	memset(&lme, 0, sizeof(lme));
	lme.local_label = label;
	n = fib_get(&pl->mpls_table->fib, &lme, 0);

	e = rte_get_temp(a);
	e->net = n;
	e->pflags = 0;

	/* Setup protocol-specific attributes */
	e->u.ldp.router_id = re->ls->lsr_id;
	//e->u.ldp.ltype = lt->type; /* XXX */

	LOG_DB_DBG("Installing label %s with nexthop %I",
		krt_print_mplslbl(label), re->nexthop);

	/* Do announce */
	rte_update2(pl->mpls_ahook, n, e, &pl->proto);

	re->flags |= LDP_RE_HASLABEL;

	return LDP_STATUS_SUCCESS;
    }
}
#endif

/**
 * l3vpn_rel_notify - dispatches forced label release
 * @protived: protocol withdrawing service
 * @requester: our protocol
 * @repl: label information
 */
static void
l3vpn_rel_notify(struct proto *provider, struct proto *requester, struct service_reply *repl)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *)requester;

  DBG("LIB database close request from %s\n", p->lib_proto->name);

  if ((repl->request != SRCLASS_LABEL) || (repl->subclass != SRSUB_LABEL_CLOSEDB))
    return;
/*
  if (repl->t != p->mpls_table) 
    return;
*/

  proto_notify_state(&p->p, PS_DOWN);
}

static u32
l3vpn_request_label(struct l3vpn_proto *p, ip_addr *nexthop)
{
  struct service_request *req;
  struct krt_mpls_label_request lr;
  struct krt_mpls_label_reply lrepl;
  size_t buflen;
  byte *rbuf;

  req = &lr.req;

  req->request = SRCLASS_LABEL;
  req->subclass = SRSUB_LABEL_REQ;
  req->t = p->mpls_table;
  req->p = &p->p;
  req->target = p->lib_proto;
  req->relfunc = NULL;

  lr.op = KMPLS_OP_L3VPN;
  lr.flags = 0;
  lr.data = 0;

  memset(&lrepl, 0, sizeof(lrepl));
  buflen = sizeof(lrepl);
  rbuf = (byte *)&lrepl;

  DBG("Requesting label..");

  if (proto_request_service((struct service_request *)&lr, &rbuf, &buflen) != SR_OK)
    return 0;

  DBG("LIB database returned label %d", lrepl.label);

  return lrepl.label;
}

static int
l3vpn_open_lib(struct l3vpn_proto *p)
{
  struct service_request req;
  struct service_reply repl;
  byte *prepl;
  size_t buflen;
  u32 label;

  req.request = SRCLASS_LABEL;
  req.subclass = SRSUB_LABEL_OPENDB;
  req.t = p->mpls_table;
  req.p = &p->p;
  req.target = NULL;
  req.relfunc = l3vpn_rel_notify;

  buflen = sizeof(repl);
  prepl = (byte *)&repl;

  DBG("Opening LIB database..");

  if (proto_request_service(&req, &prepl, &buflen) != SR_OK)
    return 0;

  DBG("LIB database opened: provider %s", repl.p->name);
  p->lib_proto = repl.p;

  /* Request per-VPN label */
  if (!(label = l3vpn_request_label(p, NULL)))
    {
      /* Label request failed, rollback */
      l3vpn_close_lib(p);
    }

  p->label = label;

  return 1;
}

static void
l3vpn_close_lib(struct l3vpn_proto *p)
{
  struct service_request req;

  if (!p->lib_proto)
    return;

  req.request = SRCLASS_LABEL;
  req.subclass = SRSUB_LABEL_CLOSEDB;
  req.t = p->mpls_table;
  req.p = &p->p;
  req.target = p->lib_proto;

  DBG("Closing LIB database..\n");

  proto_request_service(&req, NULL, NULL);
  p->lib_proto = NULL;
  p->label = 0;
}

static void
l3vpn_open_timer(timer *t)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *)t->data;

  if (l3vpn_open_lib(p))
  {
    /* Found LDP protocol instance, let's start up */
    l3vpn_run(p);
    proto_notify_state(&p->p, PS_UP);
    return;
  }

  /* Not found. restart timer */
  tm_start(t, 2);
}

static int
l3vpn_start(struct proto *P)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;
  struct l3vpn_config *c = (struct l3vpn_config *) P->cf;
  timer *t;

  DBG("l3vpn_start() called for %s\n", P->name);

  p->vpn_table = c->vpn_table->table;
  p->mpls_table = c->mpls_table->table;
  p->rd = c->rd;

  /* Clean up the secondary table stats */
  bzero(&P->stats, sizeof(struct proto_stats));
  bzero(&p->vpn_stats, sizeof(struct proto_stats));
  bzero(&p->mpls_stats, sizeof(struct proto_stats));

  rt_lock_table(P->table);
  rt_lock_table(p->vpn_table);
  rt_lock_table(p->mpls_table);

  /* Request label */
  if (!l3vpn_open_lib(p))
    {
    /* 
     * We should try to re-request label from LDP periodically.
     * .. and show custom protocol status in CLI (like BGP do)
     */
      t = tm_new(P->pool);
      t->hook = l3vpn_open_timer;
      t->data = p;

      t->randomize = 1;
      /* Start periodic LIB database checks */
      tm_start(t, 2);
      return PS_START;
    }

  /* run protocol */
  l3vpn_run(p);

  return PS_UP;
}

static void
l3vpn_cleanup(struct proto *P)
{
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;

  DBG("l3vpn_cleanup() called for %s\n", P->name);

  l3vpn_close_lib(p);

  rt_unlock_table(p->mpls_table);
  rt_unlock_table(p->vpn_table);
  rt_unlock_table(P->table);

  bzero(&P->stats, sizeof(struct proto_stats));
  bzero(&p->vpn_stats, sizeof(struct proto_stats));
  bzero(&p->mpls_stats, sizeof(struct proto_stats));

  p->mpls_ahook = NULL;
  p->vpn_ahook = NULL;
}

static struct proto *
l3vpn_init(struct proto_config *C)
{
  struct proto *P = proto_new(C, sizeof(struct l3vpn_proto));

  P->accept_ra_types = RA_OPTIMAL;
  P->rt_notify = l3vpn_rt_notify;
  P->import_control = l3vpn_import_control;
  P->reload_routes = l3vpn_reload_routes;

  return P;
}

static void
l3vpn_postconfig(struct proto_config *C)
{
  struct l3vpn_config *c = (struct l3vpn_config *) C;

  if (!c->vpn_table)
    cf_error("Name of vpn table not specified");

  if (!c->mpls_table)
    cf_error("Name of mpls table not specified");

  if ((c->vpn_table->addr_type != RT_VPNV4) && (c->vpn_table->addr_type != RT_VPNV6))
    cf_error("VPN Table %s type is not VPN", c->vpn_table->name);

  if (c->mpls_table->addr_type != RT_MPLS)
    cf_error("MPLS Table %s type is not MPLS", c->mpls_table->name);

  if ((C->table->addr_type != RT_IPV4) && (C->table->addr_type != RT_IPV6))
    cf_error("Client table %s type is not IPv4/IPv6", C->table->name);

  if (((c->vpn_table->addr_type == RT_VPNV4) && (C->table->addr_type != RT_IPV4)) ||
	  ((c->vpn_table->addr_type == RT_VPNV6) && (C->table->addr_type != RT_IPV6)))
    cf_error("Address families are different for VPN table %s and client table %s",
	c->vpn_table->name, C->table->name);

  /* 
   * No checks for IGP table is done since VPNv6 table can containg IPv4 nexthops
   */
}

static void
l3vpn_copy_config(struct proto_config *dest, struct proto_config *src)
{
  /* Just a shallow copy, not many items here */
  proto_copy_rest(dest, src, sizeof(struct l3vpn_config));
}

static void
l3vpn_get_status(struct proto *P, byte *buf)
{
  char rdbuf[20];
  struct l3vpn_proto *p = (struct l3vpn_proto *) P;

  switch (P->proto_state)
  {
    case PS_START:
      bsprintf(buf, "LIB query..");
      break;
    case PS_UP:
      addr_print_rd(rdbuf, sizeof(rdbuf), p->rd);
      bsprintf(buf, "%s <[%s]> %s", P->table->name, rdbuf, p->vpn_table->name);
      break;
    default:
      buf[0] = '\0';
  }
}

extern int proto_reconfig_type;

static int
l3vpn_reconfigure(struct proto *P, struct proto_config *new)
{
  struct proto_config *old = P->cf;
  struct l3vpn_proto *p = (struct l3vpn_proto *)P;
  struct l3vpn_config *o = (struct l3vpn_config *) P->cf;
  struct l3vpn_config *n = (struct l3vpn_config *) new;

  if (((o->vpn_table->table != n->vpn_table->table)) || (o->rd != n->rd))
    return 0;

  if (o->mpls_table->table != n->mpls_table->table)
    return 0;

  /* Update output filters in ahooks */
  if (P->main_ahook)
    {
      P->main_ahook->out_filter = new->out_filter;
      P->main_ahook->in_limit = new->in_limit;
    }

  if (p->vpn_ahook)
    {
      p->vpn_ahook->out_filter = new->in_filter;
      p->vpn_ahook->in_limit = new->out_limit;
    }

  if ((P->proto_state != PS_UP) || (proto_reconfig_type == RECONFIG_SOFT))
    return 1;
  
  if ((new->preference != old->preference)
      || ! filter_same(new->in_filter, old->in_filter)
      || ! filter_same(new->out_filter, old->out_filter))
    proto_request_feeding(P);

  return 1;
}


struct protocol proto_l3vpn = {
  name:		"l3vpn",
  template:	"l3vpn%d",
  multitable:	1,
  postconfig:	l3vpn_postconfig,
  init:		l3vpn_init,
  start:	l3vpn_start,
  cleanup:	l3vpn_cleanup,
  reconfigure:	l3vpn_reconfigure,
  copy_config: 	l3vpn_copy_config,
  get_status:	l3vpn_get_status,
};
