/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011-2013 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * LDP cli functions and various related helpers
 *
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_database.h"
#include "lib/krt.h"
#include "lib/krt_mpls.h"
#include "nest/cli.h"

void
ldp_get_route_info(rte *e, byte *buf, ea_list *attrs)
{
  eattr *ea;
  struct adata *ad;
  bird_mpls_entry *me;
  char lbuf[32];

  if ((ea = ea_find(attrs, EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL))) == NULL)
    return;

  ad = ea->u.ptr;
  me = (bird_mpls_entry *)(ad + 1);

  addr_print_me(lbuf, sizeof(lbuf), me, 0);

    buf += bsprintf(buf, " %s (%s)", krt_print_mplsop(e->u.ldp.ltype), lbuf);

  if (e->u.ldp.router_id)
    buf += bsprintf(buf, " [%R]", e->u.ldp.router_id);
}

#define CLI_OUT(fmt, ...)  cli_msg(-1006, fmt, ##__VA_ARGS__)
#define CLI_END()    cli_msg(0, "")
static int
ldp_check_protocol(struct proto *p)
{
  if (p == NULL)
    {
      CLI_OUT("Please specify protocol instance");
      return 0;
    }

  if (p->proto != &proto_ldp)
    {
      CLI_OUT("Incorrect protocol type: got %s instead of %s", p->proto->name, proto_ldp.name);
      return 0;
    }

  return 1;
}

void
ldp_show_neighbour(struct proto *p)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;
  struct ldp_iface *li;
  ldp_hello *lh, *lh_next;
  ldp_neighbour *ln;
  ldp_session *ls, *ls_next;
  time_t tm;

  if (!ldp_check_protocol(p))
    return;

/*
    Peer LDP Ident: 10.0.0.88:0; Local LDP Ident 10.2.33.4:0
        TCP connection: 10.0.0.88.646 - 10.2.33.4.11018
        State: Oper; Msgs sent/rcvd: 18/9; Downstream
        Up time: 00:00:07
        LDP discovery sources:
          FastEthernet0/0.5, Src IP addr: 10.1.5.88
*/
  WALK_LIST_DELSAFE(ls, ls_next, pl->session_list)
  {
    ls = SKIP_BACK(ldp_session, n_list, ls);
    ln = ls->ln;

    CLI_OUT("    Peer LDP Ident: %R:%d; Local LDP Ident %R:%d", ls->lsr_id, ls->space_id, pl->router_id, 0);
    if ((ln->sk != NULL) && (ls->state >= LDP_SSM_INITIALIZED))
      {
	CLI_OUT("         TCP connection: %I.%d - %I.%d", ln->sk->daddr, ln->sk->dport, ln->sk->faddr, 0);
	CLI_OUT("         State: %s; Msgs sent/rcvd: %d/%d; %s",
	  ldp_session_textstatus(ls->state), ls->messages_out, ls->messages_in, "Downstream");
	tm = now - ls->last_state;
	CLI_OUT("         Up time: %02d:%02d:%02d", tm / 3600, (tm % 3600) / 60, tm % 60);
      }

    CLI_OUT("         LDP discovery sources:");
    WALK_LIST_DELSAFE(lh, lh_next, ln->hello_list)
      {
	lh = SKIP_BACK(ldp_hello, n_lsr, lh);
	li = lh->li;
	CLI_OUT("           %s, Src IP addr: %I", li->iface->name, lh->remote);
      }
  }
  CLI_END();
}

void
ldp_show_interfaces(struct proto *p)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;
  struct ldp_iface *li, *li_next;
  struct iface *i;
  ldp_hello *lh, *lh_next;
  char buf[50];

  if (!ldp_check_protocol(p))
    return;

/*
    Interface em0, IP addr: 10.1.5.88
        State: Operational
  LDP neighbours:
    Ident: 10.0.0.88:0; From: 10.2.33.4
*/
  WALK_LIST_DELSAFE(li, li_next, pl->iface_list)
    {
      i = li->iface;

      if (i->addr != NULL)
	bsnprintf(buf, sizeof(buf), "%I/%d", i->addr->ip, i->addr->pxlen);
      else
	bsnprintf(buf, sizeof(buf), "%s", "<UNAVAIL>");
      CLI_OUT("    Interface %s, IP addr: %s", i->name, buf);
      CLI_OUT("         State: %s; bird state: %s",
	ldp_iface_textstatus(li->status), (i->flags & IF_UP) ? "UP" : "DOWN");
      CLI_OUT("         LDP discovery sources:");
      WALK_LIST_DELSAFE(lh, lh_next, pl->hello_list)
	{
	  lh = SKIP_BACK(ldp_hello, n_list, lh);
	  if (lh->li != li)
	    continue;
	  CLI_OUT("           %s, Src IP addr: %I", i->name, lh->remote);
	}
    }
  CLI_END();
}


static void
ldp_show_flags(ldp_rib_entry *re, char *pbuf, int len)
{
  int i, l;
  char *text, *pbuf_start = pbuf;

  for (i = 1; i < 65536; i *= 2)
    {
      switch (re->flags & i)
	{
	  case LDP_RE_DIRECT: { text = "DIRECT"; break; }
	  case LDP_RE_NEXTHOP: { text = "NHOP"; break; }
	  case LDP_RE_LDP: { text = "SES"; break; }
	  case LDP_RE_HASLABEL: { text = "LABEL"; break; }
	  case LDP_RE_BEST: { text = "BEST"; break; }
	  case LDP_RE_INSTALLED: { text = "RIB"; break; }
	  default: text = "";
	}

      if (!text[0])
	continue;

      l = bsnprintf(pbuf, len, "%s,", text);
      pbuf += l;
      len -= l;
    }

  /* Cut last ',' */
  if (pbuf_start != pbuf)
    *(pbuf - 1) = '\0';
}

void
ldp_show_bindings(struct proto *p)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;
  ldp_session *ls;
  ldp_rib_group *rg;
  ldp_rib_entry *re, *re_next;
  char fbuf[50], buf[100], *pbuf;

  if (!ldp_check_protocol(p))
    return;

/*
  tib entry: 10.0.0.0/24, rev 97
        local binding:  tag: 18
        remote binding: tsr: 10.2.33.88:0, tag: exp-null
*/

  /* Free all additionally allocated stuff */
  FIB_WALK(&pl->rib, rtmp)
    {
      rg = (ldp_rib_group *)rtmp;

      CLI_OUT("  Prefix %F", &rg->n);
      WALK_LIST_DELSAFE(re, re_next, rg->entries)
	{
	  ldp_show_flags(re, fbuf, sizeof(fbuf));
	  pbuf = buf;
	  pbuf += bsnprintf(pbuf, sizeof(buf), "    Entry flags: [%s]", fbuf);
	  ls = re->ls;
	  if (ls)
	    pbuf += bsnprintf(pbuf, sizeof(buf), ", session %R:%d", ls->lsr_id, ls->space_id);

	  if (re->flags & LDP_RE_HASLABEL)
	    pbuf += bsnprintf(pbuf, sizeof(buf), ", label %d", re->me.local_label);

	  CLI_OUT("%s", buf);
	}
    }
  FIB_WALK_END;

  CLI_END();
}

