/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * Various databases are mananaged hered:
 * RIB (Remote label bindings)
 * Addrmap (Nexthop <> neighbour mapping)
 *
 *
 * RIB structure:
 *
 * RIB tracks label data associated with given prefix for each neighbor.
 * RIB is very much like usual routing table:
 * it has fib (pl->rib), fib_node is really
 * "ldp_rib_group" ("net" in general routing code).
 * Rib entries are stored in linked list with best entry stored first
 * (exactly tha same as rte)
 *
 * Rte is explicitly marked best iff current best route points to the LDP
 * neighbor which announces this rte.
 *
 *
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_database.h"

static int ldp_addrmap_hash(ip_addr addr);
static void ldp_rib_init_prefix(struct fib_node *n);

int
ldp_rib_init(struct proto_ldp *pl)
{
  fib_init(&pl->rib, MAIN_POOL(pl), sizeof(struct _ldp_rib_group),
    0, ldp_rib_init_prefix);

  return LDP_OK;
}

void
ldp_rib_free(struct proto_ldp *pl, int last UNUSED)
{
  ldp_rib_group *rg;
  ldp_rib_entry *re;

  /* Free all additionally allocated stuff */
  FIB_WALK(&pl->rib, rtmp)
  {
    rg = (ldp_rib_group *)rtmp;

    WALK_LIST_FIRST(re, rg->entries)
    {
      rem_node(NODE re);
      mb_free(re);
    }
  }
  FIB_WALK_END;

  fib_free(&pl->rib);

  return;
}

/**
 * ldp_rib_add_entry - Adds FEC prefix to given database
 * @pl: pointer to protocol instance
 * @prefix: prefix IP address
 * @pxlen: prefix mask
 * @ls: pointer to ldp session (if exists)
 *
 * Adds RIB group entry if not exists
 * Adds RIB entry to group entry list and neighbour prefix list
 *
 * Returns pointer to created RIB entry or NULL in case of error
 */
ldp_rib_entry *
ldp_rib_add_entry(struct proto_ldp *pl, ip_addr *prefix, int pxlen, ldp_session *ls)
{
  ldp_rib_group *rg;
  ldp_rib_entry *re;

  if ((re = mb_allocz(MAIN_POOL(pl), sizeof(ldp_rib_entry))) == NULL)
    return NULL;

  rg = ldp_rib_get_group(pl, prefix, pxlen);

  /* Add to prefix group list */
  add_tail(&rg->entries, NODE re);
  re->rg = rg;

  /* Add to per-session list */
  if (ls != NULL)
    {
      LOG_DB_DBG("++ RIB %F in session %R:%d", &rg->n, ls->lsr_id, ls->space_id);
      ldp_rib_link_session(pl, re, ls);
    }
  else
    LOG_DB_DBG("++ RIB %F", &rg->n);

  return re;
}

/**
 * ldp_rib_set_best - mark RIB entry as best (e.g. installed)
 * @pl: pointer to protocol instance
 * @lrd: pointer to route distinguisher
 * @re: pointer to ldp session if exists
 *
 * Moves RIB entry on the first place in appropriate group 
 * and marks it as BEST
 *
 */
void
ldp_rib_set_best(struct proto_ldp *pl, ldp_rib_entry *re)
{
  re->flags |= LDP_RE_BEST;
  rem_node(NODE re);
  add_head(&re->rg->entries, NODE re);
}

void
ldp_rib_try_gc(struct proto_ldp *pl, ldp_rib_group *rg)
{
  if (!EMPTY_LIST(rg->entries))
    return;

  fib_delete(&pl->rib, rg);
}

/*
 * Link rib entry to given LDP session
 */
void
ldp_rib_link_session(struct proto_ldp *pl, ldp_rib_entry *re, ldp_session *ls)
{
  re->ls = ls;
  re->flags |= LDP_RE_LDP;
  add_tail(&ls->prefix_list, &re->n_lss);
}

/*
 * Unlink rib entry from given session
 */
void
ldp_rib_unlink_session(struct proto_ldp *pl, ldp_rib_entry *re)
{
  re->ls = NULL;
  re->flags &= ~LDP_RE_LDP;
  rem_node(&re->n_lss);
}

/*
 * Remove entry if it is not used by LDP session or routing code
 */
int
ldp_rib_gc_entry(struct proto_ldp *pl, ldp_rib_entry *re)
{

  if (!(re->flags & LDP_RE_USED))
    {
      ldp_rib_del_entry(pl, re);
      return 1;
    }

  return 0;
}

/*
 * For every rib entry associated with session:
 * 1) withdraws any MPLS and/or nexthop route
 * 2) clears LDP session route
 * 3) deletes rte if possible
 */
void
ldp_rib_flush_session(struct proto_ldp *pl, ldp_session *ls)
{
  ldp_rib_group *rg;
  ldp_rib_entry *re;

  WALK_LIST_FIRST(re, ls->prefix_list)
    {
      re = SKIP_BACK(ldp_rib_entry, n_lss, re);
      rg = re->rg;

      /* Withdraw announces if any */
      ldp_rt_withdraw_re(pl, re);

      /* Remove prefix from given session */
      ldp_rib_unlink_session(pl, re);
      /* Check if we can delete @re */
      ldp_rib_gc_entry(pl, re);
    }
}


/**
 * ldp_rib_del_entry - Deletes FEC prefix from RIB database
 * @pl: pointer to protocol instance
 * @re: pointer to RIB entry
 *
 * Deletes RIB entry from group entry list and from neighbour prefix list
 * Drops RIB group entry if no entries remains
 *
 * Returns LDP_OK in case of successful deletion or LDP_ERR otherwise
 */
int
ldp_rib_del_entry(struct proto_ldp *pl, ldp_rib_entry *re)
{
  ldp_rib_group *rg = re->rg;
  ldp_session *ls = re->ls;

  /* Remove from session prefix list */
  if (ls != NULL)
    ldp_rib_unlink_session(pl, re);

  /* Remove from group */
  rem_node(NODE re);
  mb_free(re);

  if (ls != NULL)
    LOG_DB_DBG("-- RIB %F in session %I:%d", FPREFIX(&rg->n),
        ls->lsr_id, ls->space_id);
  else
    LOG_DB_DBG("-- RIB %F", FPREFIX(&rg->n));

  /* Check if we need to free rib group */
  ldp_rib_try_gc(pl, rg);

  return LDP_OK;
}

ldp_rib_group *
ldp_rib_search_group(struct proto_ldp *pl, ip_addr *prefix, int pxlen)
{
  return fib_find(&pl->rib, prefix, pxlen);
}

ldp_rib_group *
ldp_rib_get_group(struct proto_ldp *pl, ip_addr *prefix, int pxlen)
{
  return fib_get(&pl->rib, prefix, pxlen);
}

/**
 * ldp_rib_findlink - find RIB entry for given session
 * @pl: pointer to protocol instance
 * @prefix: pointer to prefix
 * @pxlen: prefix length
 * @ls: pointer to LDP session
 *
 * Returns re if it is either:
 * 1) already linked to given session
 * 2) re has nexthop and its nexhop is mapped to given session via addrmap
 *
 * Otherwise returns NULL
 */
ldp_rib_entry *
ldp_rib_findlink(struct proto_ldp *pl, ip_addr *prefix, int pxlen, ldp_session *ls)
{
  ldp_rib_group *rg;
  ldp_rib_entry *re;

  if ((rg = ldp_rib_search_group(pl, prefix, pxlen)) == NULL)
    return NULL;

  WALK_LIST(re, rg->entries)
    {
      if ((re->flags & LDP_RE_LDP) && (re->ls == ls))
	return re;

      if ((re->flags & LDP_RE_NEXTHOP) &&
	  (ldp_addrmap_find_session(pl, re->nexthop) == ls))
	{
	  if (!(re->flags & LDP_RE_LDP))
	    ldp_rib_link_session(pl, re, ls);
	  return re;
        }
    }

  return NULL;
}

/**
 * ldp_rib_findlink_nh - find RIB entry for given nexthop
 * @pl: pointer to protocol instance
 * @re: pointer to RIB group
 * @nexhtop: IPv4 nexthop
 *
 * Returns re if it either:
 * 1) has the same nexthop (empty for direct routes)
 * 2) re is linked to session and given nexthop maps to the same session via addrmap
 *
 * Otherwise returns NULL
 */
ldp_rib_entry *
ldp_rib_findlink_nh(struct proto_ldp *pl, ldp_rib_group *rg, ip_addr nexthop)
{
  ldp_rib_entry *re;
  ldp_session *ls = NULL;
  int flag = LDP_RE_DIRECT;

  if (nexthop != 0)
    {
      ls = ldp_addrmap_find_session(pl, nexthop);
      flag = LDP_RE_NEXTHOP;
    }

  WALK_LIST(re, rg->entries)
    {
      if ((re->flags & flag) && (re->nexthop == nexthop))
	return re;

      if ((re->flags & LDP_RE_LDP) && (re->ls == ls))
	return re;
    }

  return NULL;
}

static void
ldp_rib_init_prefix(struct fib_node *n)
{
  ldp_rib_group *rg = (ldp_rib_group *)n;

  init_list(&rg->entries);
}


/************************************************************
 *                                                          *
 *                    ADDRMAP DATABASE                      *
 *                                                          *
 ***********************************************************/
#define ADDRMAP_HASH_SIZE  65536

/*
 * Initialize nexthop mapping database
 */
int
ldp_addrmap_init(struct proto_ldp *pl)
{
  list *alist;
  int i;

  if ((alist = mb_alloc(MAIN_POOL(pl), sizeof(list) * ADDRMAP_HASH_SIZE)) == NULL)
    {
      LOG_ERR("Error allocating memory for addrmap");
      return LDP_ERR;
    }

  pl->addr_hash = alist;

  for (i = 0; i < ADDRMAP_HASH_SIZE; i++, alist++)
    init_list(alist);

  return LDP_OK;
}

/*
 * Shutdown addrmap database freeing all remaining nexthops
 */
void
ldp_addrmap_shutdown(struct proto_ldp *pl, int last UNUSED)
{
  list *alist;
  int i;
  ldp_addrmap *la;

  for (i = 0, alist = pl->addr_hash; i < ADDRMAP_HASH_SIZE; i++, alist++)
    {
      WALK_LIST_FIRST(la, *alist)
	{
	  rem_node(NODE la);
	  rem_node(&la->n_lss);
	  mb_free(la);
	}
    }

  mb_free(pl->addr_hash);
  pl->addr_hash = NULL;
}

ldp_session *
ldp_addrmap_find_session(struct proto_ldp *pl, ip_addr addr)
{
  ldp_addrmap *la;

  WALK_LIST(la, pl->addr_hash[ldp_addrmap_hash(addr)])
  {
    if (ipa_equal(la->addr, addr))
      return la->ls;
  }

  return NULL;
}

int
ldp_addrmap_add(struct proto_ldp *pl, ldp_session *ls, ip_addr addr)
{
  ldp_addrmap *la;

  if ((la = mb_alloc(MAIN_POOL(pl), sizeof(ldp_addrmap))) == NULL)
    return LDP_ERR;

  la->ls = ls;
  la->addr = addr;
  /* Add to hash */
  add_tail(&pl->addr_hash[ldp_addrmap_hash(la->addr)], NODE la);
  /* Add to per-session list */
  add_tail(&ls->addr_list, &la->n_lss);

  return LDP_OK;
}

int
ldp_addrmap_del(struct proto_ldp *pl, ldp_session *ls, ip_addr addr)
{
  ldp_addrmap *la;

  WALK_LIST(la, pl->addr_hash[ldp_addrmap_hash(addr)])
    {
      if (!ipa_equal(la->addr, addr))
	continue;

      if (la->ls != ls)
	continue;

	rem_node(NODE la);
	rem_node(&la->n_lss);
	mb_free(la);
	return LDP_OK;
    }

  return LDP_ERR;
}

void
ldp_addrmap_flush_session(struct proto_ldp *pl, ldp_session *ls)
{
  ldp_addrmap *la;

  WALK_LIST_FIRST(la, ls->addr_list)
  {
    la = SKIP_BACK(ldp_addrmap, n_lss, la);
    rem_node(NODE la);
    rem_node(&la->n_lss);
    mb_free(la);
  }
}

static int 
ldp_addrmap_hash(ip_addr addr)
{
  return ipa_hash(addr);
}

