/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * LDP hello/neighbour lists management code
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_proto.h"

static ldp_hello *ldp_hello_find(struct proto_ldp *pl, struct ldp_iface *li, ip_addr remote, int space_id);
static int ldp_hello_preprocess(struct ldp_iface *li, ip_addr remote, char *data, int size);
static void ldp_hello_hold_timeout_hook(timer *t);
static void ldp_hello_shout_loud(timer *t);
static void ldp_hello_shout_targeted(timer *t);
static int ldp_hello_hash(ip_addr addr);
static int ldp_hello_rx_hook(sock *sk, int size);
static void ldp_hello_err_hook(sock *sk, int err);

/*
 * Sets up hello module discovery and targeted discovery timers
 */
int
ldp_hellos_init(struct proto_ldp *pl)
{
  timer *t;
  struct ldp_config *c = PL_TO_CF(pl);
  list *l;
  int i;

  /* Setup discovery timer */
  t = tm_new(MAIN_POOL(pl));
  t->hook = ldp_hello_shout_loud;
  t->data = pl;
  t->recurrent = c->dsend_interval;
  pl->dhello_timer = t;
  tm_start(t, 5); /* Start in 5 seconds */
  /* Setup targeted hello timer */
  t = tm_new(MAIN_POOL(pl));
  t->hook = ldp_hello_shout_targeted;
  t->data = pl;
  t->recurrent = c->tsend_interval;
  pl->thello_timer = t;
  tm_start(t, 5); /* Start in 5 seconds */

  /* Allocate and init hello hash */
  if ((pl->hello_hash = mb_alloc(MAIN_POOL(pl), sizeof(list) * LDP_HASH_SIZE)) == NULL)
    return LDP_ERR;

  for (i = 0, l = pl->hello_hash; i < LDP_HASH_SIZE; i++, l++)
    init_list(l);

  init_list(&pl->hello_list);

  return LDP_OK;
}

/*
 * ldp_hellos_shutdown - Shuts down timers
 * @pl: pointer to protocol instance
 * @last: non-zero if this is last protocol instance
 */
void
ldp_hellos_shutdown(struct proto_ldp *pl, int last UNUSED)
{
  ldp_hello *lh, *lh_next;

  rfree(pl->dhello_timer);
  rfree(pl->thello_timer);

  pl->dhello_timer = pl->thello_timer = NULL;

  /* No hello adjacencies should be left  */
  WALK_LIST_DELSAFE(lh, lh_next, pl->hello_list)
    {
      lh = SKIP_BACK(ldp_hello, n_list, lh);
      ldp_hello_shutdown(lh);
    }

  mb_free(pl->hello_hash);
  pl->hello_hash = NULL;
}


/*
 * ldp_hello_add = Creates new hello adjacency based on interface/remote address
 * @li: pointer to LDP-enable interface
 * @remote: remote address
 */
ldp_hello *
ldp_hello_add(struct ldp_iface *li, ip_addr remote)
{
  ldp_hello *lh;
  struct proto_ldp *pl = li->pl;
  timer *t;

  if ((lh = mb_allocz(MAIN_POOL(pl), sizeof(ldp_hello) + LDP_HELLO_BASE_BUFFER)) == NULL)
    return NULL;

  lh->datasize = LDP_HELLO_BASE_BUFFER;
  lh->data = (byte *)(lh + 1);
  lh->last_packet = now;
  t = tm_new(MAIN_POOL(pl));
  t->hook = ldp_hello_hold_timeout_hook;
  t->data = lh;
  lh->hold_timer = t;
  lh->li = li;
  li->refcount++;
  lh->remote = remote;

  /* Add to hash */
  add_tail(&pl->hello_hash[ldp_hello_hash(lh->remote)], NODE lh);
  /* Add to global list */
  add_tail(&pl->hello_list, &lh->n_list);

  LDBG("Added hello adjacency with %I (%p) to global list", remote, lh);

  return lh;
}


/**
 * ldp_hello_shutdown - Close hello adjacency
 * @ln: pointer to hello
 *
 * Removes adjacency from all lists and deletes it
 */
void
ldp_hello_shutdown(ldp_hello *lh)
{
  ldp_neighbour *ln = lh->ln;
  struct ldp_iface *li = lh->li;

  LOG_HL_VRB("Shutting down hello adjacency with %I (%R:%d) on %s",
    lh->remote, ln->lsr_id, lh->space_id, li->iface_name);

  /* Stop timer */
  rfree(lh->hold_timer);

  /* Delete from hash */
  rem_node(NODE lh);
  /* Delete from global list */
  rem_node(&lh->n_list);

  /* Unlink hello from ldp neihbour */
  rem_node(&lh->n_lsr);

  /* Free data buffer if allocated separately */
  if (lh->flags & LDP_FLAG_CUSTOM_BUF)
    mb_free(lh->data);

  if (!--li->refcount)
    {
      /* Check if we can delete stale interface */
      if (_IS_DELETED(li))
	mb_free(li);
    }

  mb_free(lh);
}

/**
 * ldp_hello_build - build hello message in gived buffer
 * @pl: pointer to protocol instance
 * @databuf: pointer to target data buffer
 * @datasize: buffer length
 *
 * Creates PDU with single LDP_MSG_HELLO message.
 * Common options TLV and IPv4 Transport address TLV are inserted
 *
 * Returns PDU full length or LDP_ERR in case of error
 */
int
ldp_hello_build(byte *databuf, int datasize, struct proto_ldp *pl)
{
  byte *data;
  int length;

  if (datasize < LDP_PDU_HEADER_SIZE + LDP_MESSAGE_HEADER_SIZE +
	LDP_TLV_IPV4TRANSPORT_SIZE + LDP_TLV_CHELLO_PARAMS_SIZE)
    return LDP_ERR;

  data = databuf + LDP_PDU_HEADER_SIZE + LDP_MESSAGE_HEADER_SIZE;

  /* Add mandatory hello options TLV */
  length = ldp_tlv_put_chello_params(data, pl, 0);
  /* Add our transport address */
  length += ldp_tlv_put_ipv4transport(data + length, pl);
  /* Fill in message header */
  length = ldp_message_put_header(databuf + LDP_PDU_HEADER_SIZE, NULL, 0,
    LDP_MSG_HELLO, length, NULL);
  /* Put PDU header */
  length = ldp_pdu_put_header(databuf, length, pl->space_id);

  LLOG_HL_DB2("Hello build, total length: %d", length);

  return length;
}

/**
 * ldp_hello_setup_socket - Sets up per-interface RX/TX UDP sockets
 * @li:  pointer to LDP interface
 * @direction: LDP_SOCK_RX (inbound) or LDP_SOCK_TX (outbound)
 *
 * Sets up multicast receiver in case of RX or transmitter socket
 * in case of TX
 */
int
ldp_hello_setup_socket(struct ldp_iface *li, int direction)
{
  sock *sk;
  char *tdir;
  struct proto_ldp *pl = li->pl;

  if (li->flags & LDP_FLAG_SOCKETUP)
    return LDP_ERR; /* Already setup */

  if ((sk = sk_new(MAIN_POOL(pl))) == NULL)
    {
      LOG_HL_ERR("Error creating %s socket on interface %s",
	(direction == LDP_SOCK_RX) ? "RX" : "TX", li->iface_name);
      return LDP_ERR;
    }
  sk->type = SK_UDP;
  if (direction == LDP_SOCK_RX)
  {
    sk->sport = LDP_PORT;
    sk->saddr = LDP_ALL_ROUTERS;
    tdir = "RX";
  }
  else if (direction == LDP_SOCK_TX)
  {
    sk->saddr = li->addr->ip;
    tdir = "TX";
  }
  else
  {
    LLOG_HL_ERR("Unknown socket direction");
    return LDP_ERR;
  }

  sk->tos = IP_PREC_INTERNET_CONTROL;
  sk->rx_hook = ldp_hello_rx_hook;
  sk->err_hook = ldp_hello_err_hook;
  sk->iface = li->iface;
  sk->rbsize = LDP_SOCK_RX_BUFFER;
  sk->tbsize = LDP_SOCK_TX_BUFFER;
  sk->data = (void *) li;
  sk->flags = SKF_LADDR_RX;

  if (sk_open(sk) != 0)
    {
      LOG_HL_ERR("%s socket setup failed on inteface %s", tdir, li->iface_name);
      goto err;
    }

  LDBG("%s socket created, fd=%d", tdir, sk->fd);

  if ((direction == LDP_SOCK_RX) && (li->type == LDP_IF_MULTICAST))
    {
      /* XXX: Setup IPv6 group too */
      if (sk_join_group(sk, LDP_ALL_ROUTERS) < 0)
	{
	  LOG_HL_ERR("Failed to join multicast group on %s", li->iface_name);
	  goto err;
	}

      LOG_HL_DBG("RX Multicast configured on interface %s", li->iface_name);
    }

  if (direction == LDP_SOCK_TX)
    {
      if (sk_set_ttl(sk, 1) < 0)
	{
	  LOG_HL_ERR("Setting TTL failed on interface %s", li->iface_name);
	  goto err;
	}

      /* hellos are send via this socket */
      if (sk_setup_multicast(sk) < 0)
	{
	  LOG_HL_ERR("Failed to setup multicast socket on interface %s", li->iface_name);
	  goto err;
	}

      LOG_HL_DBG("TX Multicast configured on %s", li->iface_name);
      li->tx_sk = sk;
    }
  else if (direction == LDP_SOCK_RX)
    li->rx_sk = sk;

  return LDP_OK;

err:
  rfree(sk);
  return LDP_ERR;
}

/**
 * ldp_hello_find - Find hello adjacency
 * @pl: pointer to LDP protocol
 * @li: LDP interface hello received from
 * @remote: address hello received from
 * @space_id: labelspace (NOT USED)
 *
 * Searches all hello session for given neighbour
 *
 * Returns pointer to found hello adjacency or NULL
 */
static ldp_hello *
ldp_hello_find(struct proto_ldp *pl, struct ldp_iface *li, ip_addr remote, int space_id)
{
  ldp_hello *lh, *lh_next;

  WALK_LIST_DELSAFE(lh, lh_next, pl->hello_hash[ldp_hello_hash(remote)])
    {
      if ((ipa_equal(lh->remote, remote)) && (lh->li == li))
	return lh;
    }

  return NULL;
}

static int 
ldp_hello_hash(ip_addr addr)
{
  return ipa_hash(addr);
}

/**
 * ldp_hello_preprocess - makes PDU from hello
 * @li: pointer to interface hello received on
 * @remote: address hello received from
 * @data: pointer to hello data
 * @size: data size
 *
 * Creates Hello session for given neighbour, makes PDU from hello data
 * and calls ldp_dispatch_pdu()
 *
 * Returns length read from data buffer or -1 in case of error
 */
static int
ldp_hello_preprocess(struct ldp_iface *li, ip_addr remote, char *data, int size)
{
  struct proto_ldp *pl = li->pl;
  ldp_pdu_header lph;
  ldp_hello *lh;
  ldp_neighbour *ln;
  int length, res, new = 0;

  /* 
   * XXX: this should be changed since neighbour can announce multiple labelspaces 
   * so some packet inspection should be done (like extracting PDU header)
   */
  if ((lh = ldp_hello_find(pl, li, remote, 0)) == NULL)
    {
      /* Hello from new peer, add to list */
      if ((lh = ldp_hello_add(li, remote)) == NULL)
	{
	  LOG_HL_ERR("Failed to allocate hello data for packet from %I", remote);
	  return -1;
	}

      new = 1;
    }

  res = ldp_pdu_preprocess(pl, remote, LDP_MAXPDU_INITIAL, data, size,
    lh->data, LDP_MAXPDU_INITIAL, lh->dataoff, &length);
  lh->dataoff += length;

  if ((res == -1) || (res == 0))
  {
    /*
     * Invalid hello PDU (-1) or incomplete PDU (0).
     * Hello messages should be small and should fit in
     * 576 bytes (e.g. not fragmented). Here we drop
     * all hello packets not received within single
     * datagramm. RFC 5035 says "The criteria 
     * for determining whether a Hello is acceptable are
     * implementation dependent" (see 3.5.2.1. ).
     *
     * Do not bother collecting/aging part of PDUs
     * not bound to any neighbor.
     */
    ldp_hello_shutdown(lh);
    return res;
  }

  /* Complete PDU received */
  ldp_pdu_get_header(lh->data, &lph);

  /* Create/Link neighbour structure to global list if not exists */
  if ((ln = ldp_neighbour_find(pl, lph.lsr_id)) == NULL)
    ln = ldp_neighbour_add(pl, lph.lsr_id);

  if (new)
    {
      /* Fill in new LDP hello structure */
      lh->ln = ln;
      lh->space_id = lph.space_id;
      /* Add hello to neighbour hello list */
      add_tail(&ln->hello_list, &lh->n_lsr);
      LLOG_HL_DBG("Added hello %R:%d from %I to neighbour %R:*",
	lph.lsr_id, lph.space_id, remote, ln->lsr_id);
    }

  ldp_pdu_dispatch(ln, lh->data, lh->dataoff, LDP_PDU_HELLO, lh);

  /* Zero offset in hello structure */
  lh->dataoff = 0;
  return length;
}

/* VARIOUS HOOKS */

/**
 * ldp_hello_timeout_hook - Hook called on hello adjacency timeout
 * @t: pointer to timer 
 *
 * Shutdowns hello adjacency for which timer has expired
 */
static void
ldp_hello_hold_timeout_hook(timer *t)
{
  ldp_hello *lh = t->data;
  struct ldp_iface *li = lh->li;
  struct proto_ldp *pl = li->pl;
  ldp_neighbour *ln = lh->ln;

  DEBUG_PROTO(pl);

  ldp_hello_shutdown(lh);

  LOG_HL_VRB("LDP hello adjacency with %I is down on %s",
    ipa_to_u32(ln->lsr_id), li->iface_name);

  /* Check if we need to close LDP session, RFC 5038 Clause 2.5.5 */
  if (EMPTY_LIST(ln->hello_list))
    ldp_neighbour_shutdown(ln);

  UNDEBUG_PROTO();
}


/**
 * ldp_hello_rx_hook
 * @sk: socket we received the packet.
 * @size: size of the packet
 *
 * This is the entry point for HELLO messages from neighbors.
 */
static int
ldp_hello_rx_hook(sock *sk, int size)
{
  struct ldp_iface *li = (struct ldp_iface *) sk->data;
  struct proto_ldp *pl = li->pl;
  int length;

  DEBUG_PROTO(pl);

  LOG_HL_DB2("RX hook (iface %s, src %I, dst %I, size %d)",
    sk->iface->name, sk->faddr, sk->laddr, size);

  /* 
   * We want packets from sk->iface only. Unfortunately, on BSD we
   * cannot filter out other packets at kernel level and we receive
   * all packets on all sockets
   */
  if (sk->lifindex != sk->iface->index)
    goto done;

  /* Ignore packets on interface/sockets marked as down */
  if ((sk->type != SK_UDP) || (li->status != LDP_ISM_UP))
    goto done;

  /* Some simple parsing logic has to be kept here */
  while (size > 0)
    {
      if ((length = ldp_hello_preprocess(li, sk->faddr, sk->rbuf, size)) == -1)
	{
	  LOG_HL_DBG("Received invalid HELLO on interface %s", sk->iface->name);
	  goto done;
	}

      size -= length;  
      LOG_HL_DB2("Parsed %d bytes, %d remains", length, size);
    }
done:
  UNDEBUG_PROTO();
  return 1;
}

/**
 * ldp_hello_err_hook - dispatch interface socket error
 * @sk: errored socket
 * @err: system errno value
 *
 * Does nothing at the moment
 */
static void
ldp_hello_err_hook(sock *sk, int err)
{
  struct ldp_iface *li = (struct ldp_iface *) (sk->data);
  struct proto_ldp *pl = li->pl;

  DEBUG_PROTO(pl);

  /* 
   * Do not take any action since global things like
   * interface addresses change or interface going up/down
   * are handled in different places. This error can be temporary
   * so we issue debug message and ignore it
   */
  LOG_HL_ERR("Socket error: %M on interface %s", err, li->iface_name);

  UNDEBUG_PROTO();
}

/**
 * ldp_hello_shout_loud - sends periodic LDP hellos
 * @t: pointer to timer
 *
 * Sends hello PDUs on all enabled LDP interfaces.
 * XXX: Maybe we should do such sendings on per-interface
 * timers ?
 */
void
ldp_hello_shout_loud(timer *t)
{
  struct ldp_iface *li, *li_next;
  char *data;
  sock *sk;
  int length;
  struct proto_ldp *pl = (struct proto_ldp *) t->data;

  DEBUG_PROTO(pl);

  /* Send hello to all LDP-enabled interfaces */
  WALK_LIST_DELSAFE(li, li_next, pl->iface_list)
    {
      if (li->status != LDP_ISM_UP)
        continue;

      LOG_HL_DBG("Sending HELLO on %s", li->iface_name);
      sk = li->tx_sk;

      if (sk->tbuf != sk->tpos)
        LOG_HL_VRB("Old data overwritten on transmit on %s", li->iface_name);

      data = sk->tbuf;

      if ((length = ldp_hello_build(data, LDP_MAXPDU_INITIAL, pl)) <= 0)
	{
	  LLOG_HL_VRB("ldp_build_hello() failed");
	  break;
	}

      /* Okay, let's send it */
      sk_send_to(sk, length, LDP_ALL_ROUTERS, LDP_PORT);
    }

  UNDEBUG_PROTO();
}

/**
 * ldp_hello_shout_targeted - sends periodic LDP hellos
 * @t: pointer to timer
 *
 * Sends hello PDUs to all statically configured neigbours 
 * FIXME: No such functionality exists
 */
static void
ldp_hello_shout_targeted(timer *t)
{
  struct proto_ldp *pl = (struct proto_ldp *) t->data;
  //struct ldp_iface *li;

  DEBUG_PROTO(pl);

  LLOG_HL_DB2("Targeted hello timer hook in action");

  UNDEBUG_PROTO();
}

