/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * Network interfaces manipulation code.
 * We maintain current interface list matching user-supplied
 * patterns and interfaces statuses from LDP point of view
 *
 */

#undef LOCAL_DEBUG

#include "ldp.h"

/*
 *         LDP INTERFACE STATE MACHINE
 *
 *                +-----------+
 *         +----->|   INIT    |------+
 *         |      +-----------+      |
 *         |            |            v
 *   +-----------+      |      +-----------+
 *   |    DOWN   |<-----+------|     UP    |
 *   +-----------+      |      +-----------+
 *         ^            |             
 *         |            v             
 *         |      +-----------+       
 *         +------|  INVALID  |       
 *                +-----------+       
 *
 * DOWN: initial state, no session running, no socket is set up
 * INIT: setting up socket and other parameters
 * UP: setup completed, LDP is running
 * INVALID: socket setup failed or other error occured
 */

static int ldp_iface_reconfigure(struct ldp_iface *li, struct ldp_iface_patt *lip);
static void ldp_iface_add(struct proto_ldp *pl, struct iface *iface);
static void ldp_iface_del(struct ldp_iface *li);
static int ldp_iface_start(struct ldp_iface *li);
static void ldp_iface_shutdown(struct ldp_iface *li);
static int ldp_iface_classify(struct iface *ifa);
static struct ldp_iface_patt * ldp_iface_find_patt(struct ldp_config *c, struct iface *i);
struct ldp_iface * ldp_iface_find_by_ptr(struct proto_ldp *pl, struct iface *i);

/*
 * Return human-readable interface FSM state
 */
char *
ldp_iface_textstatus(enum LDP_ISM status)
{
  char *buf = "";

  switch (status)
  {
    case LDP_ISM_DOWN:
      buf = "Down";
      break;
    case LDP_ISM_INIT:
      buf = "Init";
      break;
    case LDP_ISM_UP:
      buf = "Up";
      break;
    case LDP_ISM_INVALID:
      buf = "Invalid";
      break;
    default:
      buf = "Unknown";
  }

  return buf;
}

/*
 * Sets current FSM status for interfasce
 */
void
ldp_iface_set_status(struct ldp_iface *li, enum LDP_ISM status)
{
  LOG_VRB("interface %s change state to %s", li->iface_name, ldp_iface_textstatus(status));
  li->status = status;
}

/*
 * Inits interfaces module
 */
int
ldp_ifaces_init(struct proto_ldp *pl)
{
  init_list(&pl->iface_list);
  return LDP_OK;
}

/*
 * Shuts all configured interfaces down
 */
void
ldp_ifaces_shutdown(struct proto_ldp *pl)
{
  struct ldp_iface *li, *li_next;

  WALK_LIST_DELSAFE(li, li_next, pl->iface_list)
    ldp_iface_del(li);
}

/*
 * Builds actual interface list from interface patterns.
 * Caches an IP address we are using on this interface.
 *
 * Called (non-directly!) from bird configuration/reconfiguration hooks
 */
void
ldp_ifaces_reconfigure(struct proto_ldp *pl, struct ldp_config *c)
{
  struct ldp_iface *li, *li_next;
  struct ldp_iface_patt *lip;
  struct iface *iface;

  /* Set deleted flag */
  WALK_LIST(li, pl->iface_list)
    li->flags |= LDP_FLAG_DELETED2;

  WALK_LIST(iface, iface_list)
    {
      /* Check if interface matches our _new_ pattern list */
      if ((lip = ldp_iface_find_patt(c, iface)) == NULL)
        continue;
  
      /*
       * XXX: In case of interface destroy/create - 
       * pointers will be different for old and new.
       * We will deal with that later
       */
      if ((li = ldp_iface_find_by_ptr(pl, iface)) != NULL)
	{
	  /* Interface already exists in our configuration */
	  li->flags &= ~LDP_FLAG_DELETED2;
	  if (ldp_iface_reconfigure(li, lip) == LDP_OK)
	    continue;
    
	  /* Hard restart */
	  ldp_iface_shutdown(li);
	  ldp_iface_del(li);
	}
  
      ldp_iface_add(pl, iface);
      LOG_VRB("Interface %s attached", iface->name);
    }

  /* Delete all interfaces not matching pattern */
  WALK_LIST_DELSAFE(li, li_next, pl->iface_list)
    {
      if ((li->flags & LDP_FLAG_DELETED2) == 0)
        continue;
  
      LOG_VRB("Interface %s detached", li->iface_name);
      
      /* Interface removed from configuration */
      li->flags &= ~LDP_FLAG_DELETED2;
      if (!(li->flags & LDP_FLAG_DELETED))
	{
	  ldp_iface_shutdown(li);
	  ldp_iface_del(li);
	}
    }
}


/*
 * Tries to do on-fly interface reconfiguration if possible.
 * Returns LDP_OK in that case.
 *
 */
static int
ldp_iface_reconfigure(struct ldp_iface *li, struct ldp_iface_patt *lip)
{
  LLOG_DBG("interface %s reconfiguration process requested", li->iface_name);

  /* soft-reconfigure */
  return LDP_OK;
}

/**
 * ldp_iface_update_state -- Handles interface changing state
 * @pl: pointer to LDP protocol instance
 * @i: pointer to interface
 * @flags: interface change flags
 */
void
ldp_iface_update_state(struct proto_ldp *pl, struct iface *i, unsigned flags)
{
  struct ldp_iface *li;

  /* Filter unnecessary flags */
  flags = flags & (IF_CHANGE_DOWN | IF_CHANGE_UP);

  /* XXX: In any case, we should generate event for all LDP neighbours */

  if (flags & IF_CHANGE_DOWN)
    {
      if ((li = ldp_iface_find_by_ptr(pl, i)) == NULL)
        goto done;
  
      /* Existing interface going down */
      ldp_session_announce_addrlist_change(pl, IANA_AFI_IP,
	(byte *)&li->addr->ip, 4, 1);
      ldp_iface_shutdown(li);
      goto done;
    }

  if (!(flags & IF_CHANGE_UP))
    return;

  /* Interface going up */
  if ((li = ldp_iface_find_by_ptr(pl, i)) == NULL)
    goto done;

  /* Existing interface goes up. */
  if (i->addr != NULL)
    {
      ldp_session_announce_addrlist_change(pl, IANA_AFI_IP,
	(byte *)&i->addr->ip, 4, 0);
      ldp_iface_start(li);
    }
done:
  /* Check if our global TCP address has changed */
  ldp_iface_check_tcpaddr(pl);
}

/**
 * ldp_iface_update_addrs -- Handles interface changing state
 * @pl: pointer to LDP protocol instance
 * @i: pointer to interface address
 * @flags: interface change flags
 *
 * Handles interface address add/delete.
 */
void
ldp_iface_update_addrs(struct proto_ldp *pl, struct ifa *a, unsigned flags)
{
  struct iface *i = a->iface;
  struct ldp_iface *li;

  LOG_DBG("address update requested for interface %s", a->iface->name);

  if (flags & IF_CHANGE_UP)
    LOG_HL_VRB("interface %s prefix %I/%d UP", i->name, a->prefix, a->pxlen);
  if (flags & IF_CHANGE_DOWN)
    LOG_HL_VRB("interface %s prefix %I/%d DOWN", i->name, a->prefix, a->pxlen);

  if (!(li = ldp_iface_find_by_ptr(pl, i)))
    return;

   /* Check if main interface address has changed */
  if (li->addr == i->addr)
    return;

  li->addr = i->addr;
    
  if (li->addr == NULL)
    {
      /*
       * Primary address removed, interface goes down.
       * We can do nothing and wait till ldp_iface_update_state
       * will be called, but, to be sure..
       */
      
      ldp_session_announce_addrlist_change(pl, IANA_AFI_IP,
        (byte *)&i->addr->ip, 4, 1);
      ldp_iface_shutdown(li);
    }
  else
    {
      /*
       * Primary address changed (or added).
       * XXX: We have to clear all hello adjacencies on this interface.
       * Anyway, let's try to restart interface (rebind sockets)
       */
      if (i->flags & IF_UP)
	{
	  if (li->status != LDP_ISM_DOWN)
	    {
	      ldp_session_announce_addrlist_change(pl,
	        IANA_AFI_IP, (byte *)&li->addr->ip, 4, 1);
	      ldp_iface_shutdown(li);
	    }

	  ldp_session_announce_addrlist_change(pl,
	    IANA_AFI_IP, (byte *)&i->addr->ip, 4, 1);
	  ldp_iface_start(li);
	}
    }
}

/*
 * Checks if our global TCP address is still valid.
 * If address changes we have to shut down all sessions/hellos.
 * This is done to avoid routing loops
 */
int
ldp_iface_check_tcpaddr(struct proto_ldp *pl)
{
  ldp_neighbour *ln, *ln_next;
  ip_addr new_addr;

  new_addr = ldp_session_get_saddr(pl);
  /* Check if router-id address is changed */
  if (!ipa_compare(pl->tcp_saddr, new_addr))
    return LDP_OK;

  LOG_ERR("LDP transport address changed from %I to %I, dropping all connections!",
    pl->tcp_saddr, new_addr);

  /* 
   * Close connections with neighbours.
   * neighbour shutdown assumes hello adjacency
   * is shutting down, too
   */
  WALK_LIST_DELSAFE(ln, ln_next, pl->neighbour_list)
    {
      /* XXX: we should specify shutdown reason here */
      ldp_neighbour_shutdown(ln);
    }

  /* Set source address to new value */
  pl->tcp_saddr = new_addr;

  return LDP_ERR;
}


/*
 * Adds new interface to list of LDP-enabled interfaces.
 * Interface is added in DOWN stage, after that ldp_iface_start
 * is called
 */
static void
ldp_iface_add(struct proto_ldp *pl, struct iface *iface)
{
  struct ldp_iface *li;

  if ((li = mb_allocz(MAIN_POOL(pl), sizeof(struct ldp_iface))) == NULL)
    {
      LLOG_ERR("Error allocating interface structure");
      return;
    }

  li->iface = iface;
  li->addr = iface->addr;
  li->pl = pl;

  /* Classify interface broadcast capabilities */
  li->type = ldp_iface_classify(iface);
  ldp_iface_set_status(li, LDP_ISM_DOWN);
  if (li->addr)
    LLOG_VRB("Adding interface %s, prefix: %I/%d", iface->name, li->addr->prefix, li->addr->pxlen);
  else
    LLOG_VRB("Adding interface %s", iface->name);

  /* Finally, add to list */
  add_tail(&pl->iface_list, NODE li);

  /* Start sockets */
  if ((li->addr != NULL) && (iface->flags & IF_UP))
    ldp_iface_start(li);
}

/**
 * ldp_iface_del - Deletes interfaces from list.
 * @li: pointer to interface.
 *
 * Function is called from 3 places:
 * 1) ldp_ifaces_shutdown (we need to to delete ALL interfaces)
 * 2) ldp_ifaces_reconfigure (in case of soft reconfiguration fails which is not the case)
 * 3) interface is removed from configuration
 *
 * Some hello adjacencies can exist on given interface
 * It is unreasonble to remove such adjacencies immediately.
 * Much better way is to stop receiving/sending hellos and wait till
 * adjacency times out. This will give more chances in cases like
 * interface re-creation or some human error.
 * As a result:
 *
 * We REMOVE interface from list, close all sockets but don't delete interface by itself. 
 * When interface is added back li pointer gets changed, so new hellos received on this 
 * interface will cause new hello adjacency to form. Old hello adjacecy will time out 
 * and decrease usage count.
 */
static void
ldp_iface_del(struct ldp_iface *li)
{
  LLOG_VRB("Removing interface %s from configuration", li->iface_name);

  if ((li->status != LDP_ISM_DOWN) && (li->status != LDP_ISM_INVALID))
    ldp_iface_shutdown(li);

  /* 
   * Do not kill hello adjacencies since
   * this can easily be down/up interface issue (RJ45 jack) or some interface reset
   *
   * Set deleted flag here. 
   */
  li->iface = NULL;
  li->flags |= LDP_FLAG_DELETED;
  ldp_iface_set_status(li, LDP_ISM_INVALID);

  /* 
   * Free interface IFF its usage count is 0 (e.g. no hello adjacencies are kept)
   * This is the case when we shuts protocol down
   */

  rem_node(NODE li);

  if (li->refcount == 0)
    mb_free(li);
  else
  {
    LLOG_VRB("Keeping interface %s since its usage count (%d) is not zero", li->iface_name, li->refcount);
    /*
     * XXX: add it to special GC list
     */
  }
}

/* INTERFACE FSM MANAGEMENT */

/*
 * Prepares interface for usage (sets up sockets, etc).
 * Sets interface status to LDP_ISM_UP
 * Return LDP_OK if everything is OK
 */
static int
ldp_iface_start(struct ldp_iface *li)
{
  LLOG_VRB("Starting LDP on interface %s", li->iface_name);

  if (li->addr == NULL)
    {
      LLOG_VRB("interface %s has no primary address, skipping", li->iface_name);
      ldp_iface_set_status(li, LDP_ISM_DOWN);
      return LDP_ERR;
    }

  if (li->status != LDP_ISM_DOWN)
    {
      LLOG_ERR("called with wrong state %s for interface %s, ignoring",
	ldp_iface_textstatus(li->status), li->iface_name);
      return LDP_ERR;
    }

  /* Set up multicast hello RX/TX sockets */
  if (ldp_hello_setup_socket(li, LDP_SOCK_RX) && ldp_hello_setup_socket(li, LDP_SOCK_TX))
    {
      /* Interface added, sockets set up and operational */
      li->flags |= LDP_FLAG_SOCKETUP;
      ldp_iface_set_status(li, LDP_ISM_UP);
      return LDP_OK;
    }
  else
    {
      LLOG_ERR("RX/TX socket setup failed for interface %s", li->iface_name);
      /* Socket setup failed. */
      ldp_iface_set_status(li, LDP_ISM_INVALID);
    }

  return LDP_ERR;
}

/*
 * Shuts down interface (closes sockets, etc..)
 * Sets interface satatus to LDP_ISM_DOWN
 */
static void
ldp_iface_shutdown(struct ldp_iface *li)
{
  LLOG_DBG("Shutting down LDP on interface %s", li->iface_name);

  /* Close interface sockets */
  rfree(li->rx_sk);
  rfree(li->tx_sk);
  li->rx_sk = li->tx_sk = NULL;

  li->flags &= ~LDP_FLAG_SOCKETUP;
  ldp_iface_set_status(li, LDP_ISM_DOWN);
}

/* VARIOUS HELPER ROUTINES */

/*
 * Get multicast/broadcast interfaces capabilities
 */
static int
ldp_iface_classify(struct iface *ifa)
{
  if (ifa->flags & IF_MULTICAST)
    return LDP_IF_MULTICAST;

  return LDP_IF_NBMA; /* XXX: not fully true */
}

/*
 * Get first matching interface pattern by interface address
 */
static struct ldp_iface_patt *
ldp_iface_find_patt(struct ldp_config *c, struct iface *i)
{
  return (struct ldp_iface_patt *) iface_patt_find(&c->if_patt_list, i, NULL);
}

/*
 * Find LDP interface by bird interface
 */
struct ldp_iface *
ldp_iface_find_by_ptr(struct proto_ldp *pl, struct iface *i)
{
  struct ldp_iface *li;
  WALK_LIST(li, pl->iface_list)
    if (li->iface == i)
      return li;
  return NULL;
}

/**
 * ldp_ifaces_fill_addrs - Fill supplied buffer with all our active interface addresses
 * @af: IANA address family
 * @data: pointer to data buffer
 * @datalen: data buffer length
 * @total: variable to store total buffer size
 *
 *
 * return LDP_OK if buffer size is sufficient or LDP_ERR
 * total buffer size (required to fill buffer with ALL addresses)
 * is stored in _total
 */
int
ldp_ifaces_fill_addrs(struct proto_ldp *pl, int af, byte *data, int datalen, int *total)
{
  struct ldp_iface *li, *li_next;
  int af_size = (af == IANA_AFI_IP6) ? 16 : 4;
  int count = 0;
  int result = LDP_OK;
  ip_addr *addr;

  WALK_LIST_DELSAFE(li, li_next, pl->iface_list)
  {
    /* Assume every interface is up & runnig */
    if (datalen < af_size)
    {
      result = LDP_ERR;
      /* Permit */
      count++;
      continue;
    }

    /* XXX: IPv6 */
    addr = &li->addr->ip;
    log(L_ERR "IADDRS: found %I on iface %s", *addr, li->iface->name);

    put_addr(data, (af == IANA_AFI_IP6) ? RT_IPV6 : RT_IPV4, addr);
    data += af_size;
    datalen -= af_size;
    count++;
  }
#if 0
  WALK_LIST(iface, iface_list)
  {
    /* We're interested in active interfaces only */
    if (!(iface->flags & IF_UP))
      continue;

    WALK_LIST(a, iface->addrs)
    {
      if (datalen < af_size)
      {
        result = LDP_ERR;
        /* Permit */
        count++;
        continue;
      }

      addr = &a->ip;
      log(L_ERR "IADDRS: found %I on iface %s", *addr, iface->name);

      put_addr(data, (af == IANA_AFI_IP6) ? RT_IPV6 : RT_IPV4, addr);
      data += af_size;
      datalen -= af_size;
      count++;
    }
  }
#endif

  *total = count * af_size;
  return result;
}


