/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/**
 * DOC: LDP
 *
 * LDP is a protocol defined for distributing labels. It is the set of
 * procedures and messages by which Label Switched Routers (LSRs)
 * establish Label Switched Paths (LSPs) through a network by mapping
 * network-layer routing information directly to data-link layer
 * switched paths.  These LSPs may have an endpoint at a directly
 * attached neighbor (comparable to IP hop-by-hop forwarding), or may
 * have an endpoint at a network egress node, enabling switching via all
 * intermediary nodes. This implementation is based on RFC5074.
 *
 */

/*
 * NAMING SCHEME:
 *
 * Function naming:
 * 'Globally' (e.g. between protocol files) visible functions names by:
 * ldp[_subsystem]_action_info. Examples:
 * ldp_session_check_addrs, ldp_tlv_fill_chello_params
 * 'Locally' visible (single file) use the same scheme and (in some caces) 
 * can be prefixed with underscore
 *
 * Some exceptions:
 * * protocol hooks
 * * _texttype functions (converting various enums to human-readable form)
 *
 *
 * Structure naming:
 * 'Globally' visible structure is named by
 * _ldp_name. Evey such structure should have a typedef to ldp_name
 *
 * Exception:
 * * proto_ldp structure
 *
 * Common variables naming:
 * struct protocol: p
 * struct proto_ldp: pl
 * ldp_interface: li
 * ...
 *
 * Function grouping:
 * externally visible functions first,
 * local - next
 * hooks - last
 *
 * FILE STRUCTURE:
 *
 * Globally visible functions should come first,
 * locally visible follows,
 * vairous hooks comes last
 *
 * General notes:
 * WALK_LIST_DELSAFE uses (xx, xx_next) variable naming
 *
 * DO NOT USE WALK_LIST. Really. Copy&Paste it to the place with SKIP_BACK and..
 *
 * If we free some structure field we set it to NULL
 *
 * EVERY hook (protocol, timer, socket, ...) SHOULD
 * have
 * DEBUG_PROTO(pl) in the beginning (if it is possible to assign protocol instance)
 * and 
 * UNDEBUG_PROTO() in the end.
 * This is needed for debugging stuff to work
 *
 * 
 *
 * MEMORY MANAGEMENT:
 * Items we allocate on pdu_lp linpool/stack if possible:
 * Every data needed for parsing received PDU
 *
 */

/*
 * General protocol setup and most external world callbacks (routing table
 * updates, interface updates, reconfiguration, global timers) are
 * located in this file.
 *
 */

#define LOCAL_DEBUG

#include "ldp.h"
#include "ldp_proto.h"
#include "lib/krt_mpls.h"
#include "nest/cli.h"

list	pl_list; /* list of LDP protocols */
int	init_done;

/* General protocol routines */
static void ldp_lib_rel_notify(struct proto *provider, struct proto *requester, struct service_reply *repl);
static int ldp_lib_init(struct proto_ldp *pl);
static void ldp_lib_free(struct proto_ldp *pl);
static void ldp_lib_open_timer(timer *t);
static int ldp_run(struct proto_ldp *pl);

/*
 * Starts protocol instance.
 */
static int
ldp_start(struct proto *p)
{
  struct proto_ldp *pl = (struct proto_ldp *)p;
  struct ldp_config *c = (void *)p->cf;
  timer *t;

  DEBUG_PROTO(pl);

  /* Lock tables anyway */
  rt_lock_table(p->table);
  pl->mpls_table = c->mpls_table->table;
  rt_lock_table(pl->mpls_table);
  pl->route_table = c->route_table->table;
  rt_lock_table(pl->route_table);

  /* Copy debugging mask at first */
  memcpy(&pl->debug_mask, &c->debug_mask, sizeof(pl->debug_mask));

  if (ldp_lib_init(pl) != SR_OK)
    {
      LLOG_ERR("LDP database open delayed");
  
      /* Start periodic protocol checks */
      t = tm_new(p->pool);
      t->hook = ldp_lib_open_timer;
      t->data = p;
      t->randomize = 1;
      tm_start(t, 2);
  
      UNDEBUG_PROTO();
      return PS_START;
    }

  UNDEBUG_PROTO();

  return ldp_run(pl);
}


/**
 * ldp_lib_rel_notify - dispatches forced LIB database close
 * @protived: protocol withdrawing service
 * @requester: our protocol
 * @repl: label information
 */
static void
ldp_lib_rel_notify(struct proto *provider, struct proto *requester, struct service_reply *repl)
{
  struct proto_ldp *pl = (struct proto_ldp *)requester;

  if ((repl->request != SRCLASS_LABEL) || (repl->subclass != SRSUB_LABEL_CLOSEDB))
    return;
/*
  if (repl->t != p->mpls_table) 
    return;
*/

  DBG("LIB database close request from %s\n", pl->lib_proto->name);

  /* Protocol requests us to free all labels */
  proto_notify_state(&pl->proto, PS_DOWN);
}

/*
 * Shuts protocol down to unconfigured state
 */
static int
ldp_shutdown(struct proto *p)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;
  int last;
  ldp_neighbour *ln, *ln_next;

  DEBUG_PROTO(pl);

  /* Remove protocol from LDP protocols list */
  rem_node(&pl->pl_list);
  last = EMPTY_LIST(pl_list) ? 1 : 0;

  /* Shut down neighbours first to permit notifications */
  WALK_LIST_DELSAFE(ln, ln_next, pl->neighbour_list)
    {
      ln = SKIP_BACK(ldp_neighbour, n_list, ln);
      ldp_neighbour_shutdown(ln);
    }

  /* Stop discovery timers and kill remaining hellos if any */
  ldp_hellos_shutdown(pl, last);
  
  /* Close remaining sessions, kill socket and timers */
  ldp_sessions_shutdown(pl, last);

  /* 
   * Free all interfaces.
   * This should be called AFTER ldp_hellos_shutdown()
   * to ensure no hellos holding reference to interfaces
   */
  ldp_ifaces_shutdown(pl);

  /* Free RIB/LFIB/addrmap databases */
  ldp_rib_free(pl, last);
  ldp_addrmap_shutdown(pl, last);

  /* Free resource pools */
  lp_flush(pl->pdu_lp);
  pl->pdu_lp = NULL;

  pl->tcp_saddr = IPA_NONE;

  LOG_DBG("Going down..");
  UNDEBUG_PROTO();

  return PS_DOWN;
}

static int
ldp_run(struct proto_ldp *pl)
{
  struct proto *p = &pl->proto;
  struct ldp_config *c = (void *)p->cf;

  DEBUG_PROTO(pl);

  /* Setup linear pool for structure parsing */
  if ((pl->pdu_lp = lp_new(p->pool, 1024)) == NULL)
    {
      LLOG_ERR("Error creating LDP linear pool");
      goto error;
    }

  /* Setup routing/forwarding database */
  if ((ldp_rib_init(pl) != LDP_OK) || (ldp_addrmap_init(pl) != LDP_OK))
    {
      LLOG_ERR("Error initializing databases");
      goto error;
    }

  /* Setup global listening socket/our addresses/timers */
  if ((ldp_ifaces_init(pl) != LDP_OK))
    {
      LLOG_ERR("Error initializing interface module");
      goto error;
    }

  /* Setup global listening socket/our addresses/timers */
  if ((ldp_sessions_init(pl) != LDP_OK))
    {
      LLOG_ERR("Error initializing sessions module");
      goto error;
    }

  /* Setup hello database / timers */
  if ((ldp_hellos_init(pl) != LDP_OK))
    {
      LLOG_ERR("Error initializing hello module");
      goto error;
    }

  pl->router_id = config->router_id;

  LOG_DBG("Router ID: %R, MPLS table %s, route table %s",
    pl->router_id, pl->mpls_table->name, pl->route_table->name);

  ldp_ifaces_reconfigure(pl, c);

  ldp_session_notify_reconfigure(pl);

  add_tail(&pl_list, &pl->pl_list);
  /* Zero protocol stats */
  bzero(&p->stats, sizeof(struct proto_stats));
  bzero(&pl->route_stats, sizeof(struct proto_stats));
  bzero(&pl->mpls_stats, sizeof(struct proto_stats));

  p->main_ahook = proto_add_announce_hook(p, p->table, &p->stats);

  /* XXX: proper limits */
  p->main_ahook->out_filter = p->cf->out_filter;
  p->main_ahook->in_filter = p->cf->in_filter;
  p->main_ahook->in_limit = p->cf->in_limit;

  pl->mpls_ahook = proto_add_announce_hook(p, pl->mpls_table, &pl->mpls_stats);
  pl->route_ahook = proto_add_announce_hook(p, pl->route_table, &pl->route_stats);

  UNDEBUG_PROTO();
  return PS_UP;

error:
  /* Free various resources */
  lp_flush(pl->pdu_lp);
  pl->pdu_lp = NULL;

  UNDEBUG_PROTO();
  return PS_STOP;
}

static int
ldp_lib_init(struct proto_ldp *pl)
{
  struct service_request req;
  struct service_reply repl;
  byte *prepl;
  size_t buflen;

  req.request = SRCLASS_LABEL;
  req.subclass = SRSUB_LABEL_OPENDB;
  req.t = pl->mpls_table;
  req.p = &pl->proto;
  req.target = NULL;
  req.relfunc = ldp_lib_rel_notify;

  buflen = sizeof(repl);
  prepl = (byte *)&repl;

  LOG_DBG("Opening LIB database..");

  if (proto_request_service(&req, &prepl, &buflen) != SR_OK)
    return 0;

  LOG_DBG("LIB database opened: provider %s", repl.p->name);
  pl->lib_proto = repl.p;

  return 1;
}

static void
ldp_lib_free(struct proto_ldp *pl)
{
  struct service_request req;

  if (!pl->lib_proto)
    return;

  req.request = SRCLASS_LABEL;
  req.subclass = SRSUB_LABEL_CLOSEDB;
  req.t = pl->mpls_table;
  req.p = &pl->proto;
  req.target = pl->lib_proto;

  DBG("Closing LIB database..\n");

  proto_request_service(&req, NULL, NULL);
  pl->lib_proto = NULL;
}

static void
ldp_lib_open_timer(timer *t)
{
  struct proto_ldp *pl = (struct proto_ldp *)t->data;

  DEBUG_PROTO(pl);

  LOG_DBG("LDP DB timer..");

  if (ldp_lib_init(pl) == SR_OK)
    {
      /* Found LDP protocol instance, let's start up */
      proto_notify_state(&pl->proto, ldp_run(pl));
      return;
    }
  else
    tm_start(t, 2);

  UNDEBUG_PROTO();
}

/*
 * Free locked tables.
 */
static void
ldp_cleanup(struct proto *p)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;

  ldp_lib_free(pl);

  rt_unlock_table(p->table);
  rt_unlock_table(pl->mpls_table);
  rt_unlock_table(pl->route_table);

  pl->mpls_table = NULL;
  pl->route_table = NULL;

  pl->mpls_ahook = NULL;
  pl->route_ahook = NULL;
}

/*
 * Does protocol reconfiguration.
 * Returns 1 if reconfiguration succeeded or
 * 0 if soft reconfiguration can'te be done
 */
static int
ldp_reconfigure(struct proto *p, struct proto_config *c)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;
  struct ldp_config *old = (struct ldp_config *) (p->cf);
  struct ldp_config *new = (struct ldp_config *) c;

  DEBUG_PROTO(pl);

  LOG_VRB("Protocol reconfiguration requested");

  if (old->distribute_mode != new->distribute_mode)
    {
      LOG_ERR("Distribute mode cannot be changed on fly");
      goto error;
    }

  if (old->retention_mode != new->retention_mode)
    {
      LOG_ERR("Retention mode cannot be changed on fly");
      goto error;
    }

  if (old->lspace_type != new->lspace_type)
    {
      LOG_ERR("Label space type cannot be changed on fly");
      goto error;
    }

  /* Checks completed, new configuration can be applied on fly */
  memcpy(&pl->debug_mask, &new->debug_mask, sizeof(pl->debug_mask));

  /* Reconfigure all participating interfaces */
  ldp_ifaces_reconfigure(pl, new);

  /* Notify sessions aboute some re-configuration taking place */
  ldp_session_notify_reconfigure(pl);

  return 1;
error:
  LOG_VRB("Protocol soft reconfiguration failed");
  UNDEBUG_PROTO();
  return 0;
}

/*
 * Dumps protocol state on debugging output.
 * A bit outdated
 *
 */
static void
ldp_dump(struct proto *p)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	struct ldp_iface *li;

	DEBUG_PROTO(pl);

	debug("Interfaces:\n");
	WALK_LIST(li, pl->iface_list)
	{
		debug("Interface: %s\n", li->iface->name);
		debug("LDP address: %I/%d\n", li->addr->prefix, li->addr->pxlen);
		debug("LDP status: %s\n", ldp_iface_textstatus(li->status));
	}

	UNDEBUG_PROTO();
}

/*
 * Interface state change notification.
 * Pass to iface layer.
 * Generally we update our local states and
 * checking if our local address is changed
 */
static void
ldp_if_notify(struct proto *p, unsigned flags, struct iface *i)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;

  DEBUG_PROTO(pl);
  LLOG_DBG("Called on interface %s, flags 0x%X", i->name, flags);
  ldp_iface_update_state(pl, i, flags);
  UNDEBUG_PROTO();
}

/*
 * Interface address change notification.
 * Pass to iface layer.
 */
void
ldp_ifa_notify(struct proto *p, unsigned flags, struct ifa *a)
{
  struct proto_ldp *pl = (struct proto_ldp *) p;

  DEBUG_PROTO(p);
  LLOG_DBG("Called on interface %s, flags 0x%X", a->iface->name, flags);
  ldp_iface_update_addrs(pl, a, flags);
  ldp_iface_check_tcpaddr(pl);
  UNDEBUG_PROTO();
}

/*
 * Checks parsed LDP config
 */
void
ldp_check(struct ldp_config *c)
{
  if ((c->mpls_table == NULL) || (c->mpls_table->addr_type != RT_MPLS))
    cf_error("Please set valid mpls_table for an LDP protocol instance");

  if (c->route_table == NULL)
    cf_error("Please set valid route_table for an LDP protocol instance");
}

/*
 * Create new protocol instance. initialize its callbacs/lists
 */
static struct proto *
ldp_init(struct proto_config *c)
{
  struct proto *p;

  /* Init protocol list */
  if (init_done == 0)
    {
      init_done = 1;
      init_list(&pl_list);
    }

  p = proto_new(c, sizeof(struct proto_ldp));

  p->if_notify = ldp_if_notify;
  p->ifa_notify = ldp_ifa_notify;
  p->accept_ra_types = RA_OPTIMAL;
  p->rt_notify = ldp_rt_notify;

  return p;
}

struct protocol proto_ldp = {
  name:		"LDP",
  template:	"ldp%d",
  multitable:	1,
  preference:	LDP_DEFAULT_PREF,
  init:		ldp_init,
  dump:		ldp_dump,
  start:	ldp_start,
  shutdown:	ldp_shutdown,
  cleanup:	ldp_cleanup,
  reconfigure:	ldp_reconfigure,
  get_route_info:	ldp_get_route_info
};

