/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

#ifndef _BIRD_LDP_DEBUG_H_
#define _BIRD_LDP_DEBUG_H_

extern struct proto_ldp *d_proto;
/* 
 * Those macroses have to be called in the beginning/end of EACH 
 * routine called from outer world (hooks, timer/socket callbacks)
 */
#undef DEBUG_HOOKS

#ifndef DEBUG_HOOKS
#define DEBUG_PROTO(x)		{ d_proto = (struct proto_ldp *)(x); } 
#define UNDEBUG_PROTO(x)	{ d_proto = NULL; }
#else
#define DEBUG_PROTO(x)		do { \
					d_proto = (struct proto_ldp *)(x); \
					log(L_ERR "SWITCH TO %s:%d", __FUNCTION__, __LINE__); \
				} while (0)

#define UNDEBUG_PROTO(x)	do {\
					log(L_ERR "END HOOK %s:%d", __FUNCTION__, __LINE__); \
					d_proto = NULL; \
				} while (0)
#endif

/* Debug classes */
#define LD_GENERAL	0x01	/* Everything not fit in other classes */
#define LD_HELLO	0x02	/* (HL) Hello send/recv */
#define LD_SESSION	0x03	/* (SS) Session data */
#define LD_DB		0x04	/* (DB) LIB/RIB databases */
#define LD_KERN		0x05	/* (KR) Kernel interaction */
#define LD_MAX		0x06	/* Maximum array size */

/* Severify flags */
#define LDP_NONE	0x00	/* No debugging */
#define LDP_ERROR	0x01	/* Protocol error */
#define LDP_VRB		0x02	/* Verbose */
#define LDP_WRN		0x04	/* Warning */
#define LDP_DBG		0x08	/* Debug */
#define LDP_DB2		0x10	/* More debug - packet dumps */


/*
 * Bird-style LDBG macro for printing data with function names`
 */
#if defined(LOCAL_DEBUG) || defined(GLOBAL_DEBUG)
#define LDBG(x, y...) debug("%s:%d " x "\n", __FUNCTION__, __LINE__, ##y)
#else
#define LDBG(x, y...) do { } while(0)
#endif

/* 
 * Debug macro naming: 
 * [L]LOG[_CLASS]_SEVERITY(fmt, data)
 * [L] prepends function:line (like 'ldp_rt_notify_pre:651 ')
 * Class is one of (HL, SS, DB, KR). If message belongs to
 * LD_GEN class its naming is omitted.
 * Severity is one of (ERR, VRB, WRN, DBG, DB2)
 *
 */

/* Map macro to class, severity and bird logger severity */

/* LD_GENERAL */
#define LOG_ERR(fmt, ...)		_LP_DEBUG(LD_GENERAL, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LOG_VRB(fmt, ...)		_LP_DEBUG(LD_GENERAL, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LOG_WRN(fmt, ...)		_LP_DEBUG(LD_GENERAL, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LOG_DBG(fmt, ...)		_LP_DEBUG(LD_GENERAL, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LOG_DB2(fmt, ...)		_LP_DEBUG(LD_GENERAL, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

#define LLOG_ERR(fmt, ...)		_LP_LDEBUG(LD_GENERAL, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LLOG_VRB(fmt, ...)		_LP_LDEBUG(LD_GENERAL, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LLOG_WRN(fmt, ...)		_LP_LDEBUG(LD_GENERAL, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LLOG_DBG(fmt, ...)		_LP_LDEBUG(LD_GENERAL, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LLOG_DB2(fmt, ...)		_LP_LDEBUG(LD_GENERAL, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

/* LD_HELLO */
#define LOG_HL_ERR(fmt, ...)		_LP_DEBUG(LD_HELLO, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LOG_HL_VRB(fmt, ...)		_LP_DEBUG(LD_HELLO, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LOG_HL_WRN(fmt, ...)		_LP_DEBUG(LD_HELLO, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LOG_HL_DBG(fmt, ...)		_LP_DEBUG(LD_HELLO, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LOG_HL_DB2(fmt, ...)		_LP_DEBUG(LD_HELLO, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

#define LLOG_HL_ERR(fmt, ...)		_LP_LDEBUG(LD_HELLO, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LLOG_HL_VRB(fmt, ...)		_LP_LDEBUG(LD_HELLO, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LLOG_HL_WRN(fmt, ...)		_LP_LDEBUG(LD_HELLO, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LLOG_HL_DBG(fmt, ...)		_LP_LDEBUG(LD_HELLO, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LLOG_HL_DB2(fmt, ...)		_LP_LDEBUG(LD_HELLO, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

/* LD_SESSION */
#define LOG_SS_ERR(fmt, ...)		_LP_DEBUG(LD_SESSION, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LOG_SS_VRB(fmt, ...)		_LP_DEBUG(LD_SESSION, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LOG_SS_WRN(fmt, ...)		_LP_DEBUG(LD_SESSION, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LOG_SS_DBG(fmt, ...)		_LP_DEBUG(LD_SESSION, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LOG_SS_DB2(fmt, ...)		_LP_DEBUG(LD_SESSION, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

#define LLOG_SS_ERR(fmt, ...)		_LP_LDEBUG(LD_SESSION, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LLOG_SS_VRB(fmt, ...)		_LP_LDEBUG(LD_SESSION, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LLOG_SS_WRN(fmt, ...)		_LP_LDEBUG(LD_SESSION, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LLOG_SS_DBG(fmt, ...)		_LP_LDEBUG(LD_SESSION, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LLOG_SS_DB2(fmt, ...)		_LP_LDEBUG(LD_SESSION, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

/* LD_DB */
#define LOG_DB_ERR(fmt, ...)		_LP_DEBUG(LD_DB, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LOG_DB_VRB(fmt, ...)		_LP_DEBUG(LD_DB, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LOG_DB_WRN(fmt, ...)		_LP_DEBUG(LD_DB, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LOG_DB_DBG(fmt, ...)		_LP_DEBUG(LD_DB, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LOG_DB_DB2(fmt, ...)		_LP_DEBUG(LD_DB, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

#define LLOG_DB_ERR(fmt, ...)		_LP_LDEBUG(LD_DB, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LLOG_DB_VRB(fmt, ...)		_LP_LDEBUG(LD_DB, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LLOG_DB_WRN(fmt, ...)		_LP_LDEBUG(LD_DB, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LLOG_DB_DBG(fmt, ...)		_LP_LDEBUG(LD_DB, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LLOG_DB_DB2(fmt, ...)		_LP_LDEBUG(LD_DB, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

/* LD_KERN */
#define LOG_KR_ERR(fmt, ...)		_LP_DEBUG(LD_KERN, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LOG_KR_VRB(fmt, ...)		_LP_DEBUG(LD_KERN, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LOG_KR_WRN(fmt, ...)		_LP_DEBUG(LD_KERN, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LOG_KR_DBG(fmt, ...)		_LP_DEBUG(LD_KERN, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LOG_KR_DB2(fmt, ...)		_LP_DEBUG(LD_KERN, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

#define LLOG_KR_ERR(fmt, ...)		_LP_LDEBUG(LD_KERN, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define LLOG_KR_VRB(fmt, ...)		_LP_LDEBUG(LD_KERN, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define LLOG_KR_WRN(fmt, ...)		_LP_LDEBUG(LD_KERN, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define LLOG_KR_DBG(fmt, ...)		_LP_LDEBUG(LD_KERN, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define LLOG_KR_DB2(fmt, ...)		_LP_LDEBUG(LD_KERN, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

/* CUSTOM */
#define CLOG_ERR(dtype, fmt, ...)	_LP_DEBUG(dtype, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define CLOG_VRB(dtype, fmt, ...)	_LP_DEBUG(dtype, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define CLOG_WRN(dtype, fmt, ...)	_LP_DEBUG(dtype, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define CLOG_DBG(dtype, fmt, ...)	_LP_DEBUG(dtype, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define CLOG_DB2(dtype, fmt, ...)	_LP_DEBUG(dtype, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)

#define CLLOG_ERR(dtype, fmt, ...)	_LP_LDEBUG(dtype, LDP_ERROR, L_ERR, fmt, ##__VA_ARGS__)
#define CLLOG_VRB(dtype, fmt, ...)	_LP_LDEBUG(dtype, LDP_VRB, L_INFO, fmt, ##__VA_ARGS__)
#define CLLOG_WRN(dtype, fmt, ...)	_LP_LDEBUG(dtype, LDP_WRN, L_WARN, fmt, ##__VA_ARGS__)
#define CLLOG_DBG(dtype, fmt, ...)	_LP_LDEBUG(dtype, LDP_DBG, L_DEBUG, fmt, ##__VA_ARGS__)
#define CLLOG_DB2(dtype, fmt, ...)	_LP_LDEBUG(dtype, LDP_DB2, L_DEBUG, fmt, ##__VA_ARGS__)


/* General debug of given type associated with protocol */
#define _LP_DEBUG(_dtype, _dseverity, _bseverity, _fmt, ...) \
	if ((d_proto) && (d_proto->debug_mask[_dtype] & _dseverity)) \
		log(_bseverity "%s: " _fmt, d_proto->proto.name, ##__VA_ARGS__)
#define _LP_LDEBUG(_dtype, _dseverity, _bseverity, _fmt, ...) \
	if ((d_proto) && (d_proto->debug_mask[_dtype] & _dseverity)) \
	 	log_msg(_bseverity "%s: %s:%d " _fmt, d_proto->proto.name, (char *)__FUNCTION__, __LINE__, ##__VA_ARGS__)


/* LDP-specific packet dumping macro */
extern int dump_packets;
#define DEBUG_PACKET(dtype)		{ dump_packets = d_proto->debug_mask[dtype] & LDP_DB2; }
#define UNDEBUG_PACKET()		{ dump_packets = 0; }
#define DUMP_PACKET			dump_packets

#define TDUMP(_fmt, ...)		log_msg(_fmt, ##__VA_ARGS__)
#define TLINE()	

#endif

