/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */
#ifndef _BIRD_LDP_PROTO_H_
#define _BIRD_LDP_PROTO_H_

#define LDP_PROTOCOL_VERSION		0x01 /* RFC 5036 */
//#define LDP_MAXPDU_INITIAL		4096
#define LDP_PORT			646

#define LDP_ALL_ROUTERS			ipa_from_u32(0xe0000002)
#define LDP_ALL_ROUTERS6		_MI(0xFF020000, 0, 0, 2)

/* RFC 5036, Clause 3.7 */
#define LDP_MSG_NOTIFICATION	0x0001
#define LDP_MSG_HELLO		0x0100
#define LDP_MSG_INIT		0x0200
#define LDP_MSG_KEEPALIVE	0x0201
#define LDP_MSG_ADDR		0x0300
#define LDP_MSG_ADDRWITH	0x0301
#define LDP_MSG_LMAP		0x0400
#define LDP_MSG_LREQ		0x0401
#define LDP_MSG_LWITH		0x0402
#define LDP_MSG_LREL		0x0403
#define LDP_MSG_LABRT		0x0404

/* RFC 5036, Clause 3.8 */
#define LDP_TLV_FEC		0x0100
#define LDP_TLV_ADDRLIST	0x0101
#define LDP_TLV_HOPCOUNT	0x0103
#define LDP_TLV_PATHVECTOR	0x0104
#define LDP_TLV_GLABEL		0x0200
#define LDP_TLV_ATMLABEL	0x0201
#define LDP_TLV_FRLABEL		0x0202
#define LDP_TLV_STATUS		0x0300
#define LDP_TLV_EXTSTATUS	0x0301
#define LDP_TLV_RPDU		0x0302
#define LDP_TLV_RMSG		0x0303
#define LDP_TLV_CHELLOPAR	0x0400
#define LDP_TLV_IP4TRADDRESS	0x0401
#define LDP_TLV_CONFSEQ		0x0402
#define LDP_TLV_IP6TRADDRESS	0x0403
#define LDP_TLV_CSESSPAR	0x0500
#define LDP_TLV_ATMSESSPAR	0x0501
#define LDP_TLV_FRSESSPAR	0x0502
#define LDP_TLV_LREQMSGID	0x0600
#define LDP_TLV_EXPROUTE	0x0800

/* RFC 5036, Clause 3.9 */
#define LDP_STATUS_SUCCESS		0x00
#define LDP_STATUS_BAD_LDP_ID		0x01
#define LDP_STATUS_BAD_PROTO_VER	0x02
#define LDP_STATUS_BAD_PDU_LEN		0x03
#define LDP_STATUS_UNKN_MSG_TYPE	0x04
#define LDP_STATUS_BAD_MSG_LEN		0x05
#define LDP_STATUS_UNKN_TLV_TYPE	0x06
#define LDP_STATUS_BAD_TLV_LEN		0x07
#define LDP_STATUS_BAD_TLV_VAL		0x08
#define LDP_STATUS_HOLD_TIMER		0x09
#define LDP_STATUS_SHUTDOWN		0x0A
#define LDP_STATUS_LOOP_DETECTED	0x0B
#define LDP_STATUS_UNKN_FEC		0x0C
#define LDP_STATUS_NO_ROUTE		0x0D
#define LDP_STATUS_NO_LABELRES		0x0E
#define LDP_STATUS_LABELRES_OK		0x0F
#define LDP_STATUS_SESS_REJ_HELLO	0x10
#define LDP_STATUS_SESS_REJ_PADV	0x11
#define LDP_STATUS_SESS_REJ_MAXPDU	0x12
#define LDP_STATUS_SESS_REJ_LRANGE	0x13
#define LDP_STATUS_KEEPALIVE_TIMER	0x14
#define LDP_STATUS_LREQ_ABORT		0x15
#define LDP_STATUS_MISS_MSG_PARAMS	0x16
#define LDP_STATUS_UNSUPP_AF		0x17
#define LDP_STATUS_SESS_REJ_KEEPALIVE	0x18
#define LDP_STATUS_INT_ERROR		0x19

#define LDP_BIT_NONE		0x00
#define LDP_BIT_UNKNOWN		0x01

#define LDP_LABEL_ADV_UNSOLICITED	0x00
#define LDP_LABEL_ADV_DEMAND		0x01

/* http://www.iana.org/assignments/mpls-label-values/mpls-label-values.xml */
#define LDP_LABEL_IPV4EXPNULL		0x00
#define LDP_LABEL_IPV6EXPNULL		0x02
#define LDP_LABEL_IMPNULL		0x03

/* Network protocol structures */
typedef struct _ldp_pdu_header {
	uint16_t	version;	/* Protocol version number */
	uint16_t	length;		/* PDU length */
	uint32_t	lsr_id;		/* LDP Identifier: LSR id */
	uint16_t	space_id;	/* LDP Identifier: label space id */
} ldp_pdu_header;
#define LDP_PDU_HEADER_SIZE		10

typedef struct _ldp_pdu {
	char		*data;		/* pointer to PDU data */
	ip_addr		remote;		/* remote party address */
	int		flags;
	struct proto_ldp	*pl;	/* Pointer to protocol (memory pool, lalels) */
	ldp_pdu_header	*lph;		/* PTR to PDU converted header */
} ldp_pdu;

typedef struct _ldp_message_header {
	uint16_t	unknown:1;	/* Unknown bit */
	uint16_t	type:15;	/* Message type */
	uint16_t	length;		/* Message length */
	uint32_t	id;		/* Message id */
} ldp_message_header;
#define LDP_MESSAGE_HEADER_SIZE		8

typedef struct _ldp_tlv_header {
	uint16_t	unknown:1;	/* Unknown bit */
	uint16_t	forward:1;	/* Forward bit */
	uint16_t	type:14;	/* TLV type */
	uint16_t	length;		/* TLV value length */
} ldp_tlv_header;
#define LDP_TLV_HEADER_SIZE		4

/* Structure for storing TLVs which needs to be forwarded */
typedef struct _ldp_tlv_forward {
	node		n;
	uint16_t	type;
	uint16_t	length;
} ldp_tlv_forward;

/* Base structure for storing interpreted TLVs */
typedef struct _ldp_tlv {
	node		n; /* This should be the first record not to force SKIP_BACK() in tlv_list */
	ldp_tlv_header	base;
} ldp_tlv;
#define _base		base.base

/* Custom TLV headers */
typedef struct _ldp_tlv_ipv4transport {
	ldp_tlv		base;
	uint32_t	addr;
} ldp_tlv_ipv4transport;
#define LDP_TLV_IPV4TRANSPORT_SIZE	LDP_TLV_HEADER_SIZE + 4

typedef struct _ldp_tlv_chello_params {
	ldp_tlv		base;
	uint16_t	hold_time;
	uint16_t	thello:1;
	uint16_t	rthello:1;
	uint16_t	reserved:14;
} ldp_tlv_chello_params;
#define LDP_TLV_CHELLO_PARAMS_SIZE	LDP_TLV_HEADER_SIZE + 4

/* RFC 5036 3.5.3. Initialization Message */
typedef struct _ldp_tlv_csess_params {
	ldp_tlv		base;
	uint16_t	pversion;
	uint16_t	keep_interval;
	uint8_t		adiscipline:1;
	uint8_t		loopdetect:1;
	uint8_t		reserved:6;
	uint8_t		pvlimit;
	uint16_t	max_pdu;
	uint32_t	lsr_id;
	uint16_t	space_id;
} ldp_tlv_csess_params;
#define LDP_TLV_CSESS_PARAMS_SIZE	LDP_TLV_HEADER_SIZE + 4 + 2 + 2 + 6

/* RFC 5036 3.4.6. Status TLV */
typedef struct _ldp_tlv_status_params {
	ldp_tlv		base;
	uint32_t	fatal:1;
	uint32_t	forward:1;
	uint32_t	statuscode:30;
	uint32_t	msgid;
	uint16_t	msgtype;
} ldp_tlv_status_params;
#define LDP_TLV_STATUS_PARAMS_SIZE	LDP_TLV_HEADER_SIZE + 10

/* RFC 5036 3.4.3. Address list TLV */
typedef struct _ldp_tlv_addrlist {
	ldp_tlv		base;
	uint16_t	iana_af;
} ldp_tlv_addrlist;
/* This is known size, actually */
#define LDP_TLV_ADDRLIST_SIZE		LDP_TLV_HEADER_SIZE + 2

/* RFC 5036 3.4.1. FEC TLV */
typedef struct _ldp_tlv_fec {
	ldp_tlv		base;
	int		count;	/* Non-standart field */
} ldp_tlv_fec;
#define LDP_TLV_FEC_SIZE		LDP_TLV_HEADER_SIZE

#define LDP_FEC_ELEMENT_WILD	0x01
#define LDP_FEC_ELEMENT_PREFIX	0x02
/* Pure host-format structure */
typedef struct _ldp_tlv_fec_element {
	uint16_t	family; /* Address family */
	uint8_t		plen; /* Prefix length */
	uint8_t		type; /* Prefix type */
	ip_addr		prefix; /* Prefix */
} ldp_tlv_fec_element;


/* RFC 5036 Generic Label TLV */
typedef struct _ldp_tlv_generic_label {
	ldp_tlv		base;
	uint32_t	label; /* RFC 3032 Label stack encoding */
} ldp_tlv_generic_label;
#define LDP_TLV_GENERIC_LABEL_SIZE	LDP_TLV_HEADER_SIZE +4


#endif
