/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011-2013 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * BIRD routing table hooks / manipulation routines.
 *
 *
 * @rte flags description:
 *
 *  -- flags, describing RTE source --
 * LDP_RE_DIRECT - direct route (comes from RTD_DEVICE announce).
 *  We don't propagate them to nexthop table, but announce such routes
 *  to other LDP neighbours with MPLS_LABEL_IPV4NULL (if LDP_RE_BEST is set).
 *
 * LDP_RE_NEXTHOP - route with gateway (comes from RTD_ROUTER announces).
 *  Main candidate to propagate to nexthop table.
 *
 * LDP_RE_LDP - rte propagated via LDP. This flag implies "lt" field is set
 *  to valid structure describing remote label and session.
 *
 * -- flags, describing RTE status --
 * LDP_RE_HASLABEL - route has local label associated with it. Label itself
 *  is stored in lt->me.local_label. This flag implies that local label is
 *  installed to MPLS table. Note that is does NOT nesessary means that re is
 *  installed in nexthop table (think of direct routes advertised by LDP
 *  speakers with explicit NULL label).
 *
 * LDP_RE_BEST - route is currently the best one in source route table.
 *  Best route is always relinked to the rtes head in @rg.
 *
 * LDP_RE_INSTALLED - route is installed into nexthop table. This implies
 *  LDP_RE_NEXTHOP|LDP_RE_LDP|LDP_RE_HASLABEL are all set.
 *
 *
 */

#define LOCAL_DEBUG

#include "ldp.h"
#include "ldp_database.h"
#include "ldp_proto.h" /* LDP_LABEL_IMPNULL */
#include "lib/krt.h"
#include "lib/krt_mpls.h"

static int ldp_rt_announce_label(struct proto_ldp *pl, ldp_rib_entry *re);
static void ldp_rt_withdraw_label(struct proto_ldp *pl, ldp_rib_entry *re);
static void ldp_rt_announce_rte(struct proto_ldp *pl, ldp_rib_entry *re);
static void ldp_rt_withdraw_rte(struct proto_ldp *pl, ldp_rib_entry *re);

static int ldp_rt_install_re(struct proto_ldp *pl, ldp_rib_entry *re);

static int ldp_rt_request_label(struct proto_ldp *pl, ldp_rib_entry *re);


/*
 * Handles route announce for main IGP route table.
 */
void
ldp_rt_notify(struct proto *p, rtable *tbl, net *n, rte *new, rte *old, ea_list *attrs)
{
	struct proto_ldp *pl = (struct proto_ldp *) p;
	int dest_type, found = 0, flag;
	ldp_rib_entry *re = NULL;
	ldp_rib_group *rg;

	DEBUG_PROTO(pl);

	/* We're interested ONLY in main table announces */
	if ((tbl != p->table) || (!new && !old))
	{
		UNDEBUG_PROTO();
		return;
	}

	/* Check if we have something on given prefix */
	rg = ldp_rib_search_group(pl, FPREFIX_IP(&n->n), n->n.pxlen);

#if 0
	if (old && new && (old->attrs->dest == new->attrs->dest) && (dest_type = RTD_ROUTER))
	{
		/* Fast update can be handled here */
		return;
	}
#endif

	if (old && rg)
	{
		/* Remove flag from best entry if set */
		WALK_LIST(re, rg->entries)
		{
			if (re->flags & LDP_RE_BEST)
			{
				found = 1;
				break;
			}
		}

		/* XXX */
		if (!found)
			return;

		/* Withdraw announces if any */
		ldp_rt_withdraw_re(pl, re);

		/* Remove flags & nexthop */
		re->flags &= ~(LDP_RE_BEST|LDP_RE_DIRECT|LDP_RE_NEXTHOP);
		re->nexthop = 0;
		re->iface = NULL;

		/* check if we can delete unused re now */
		if (!new)
			ldp_rib_gc_entry(pl, re);
	}

	if (new)
	{
		/* Announce new prefix */

		dest_type = new->attrs->dest;
		flag = (dest_type == RTD_ROUTER) ? LDP_RE_NEXTHOP : LDP_RE_DIRECT;

		if (rg)
			re = ldp_rib_findlink_nh(pl, rg,
			    (dest_type == RTD_ROUTER) ? new->attrs->gw: 0);

		if (!re)
		{
			/* Create new re */
			re = ldp_rib_add_entry(pl, FPREFIX_IP(&n->n), n->n.pxlen, NULL);
		}

		/* Mark re with appropriate flag */
		re->flags |= flag;

		/* Mark as best and relink */
		ldp_rib_set_best(pl, re);

		if (flag == LDP_RE_NEXTHOP)
		{
			/* Save nhop information */
			re->nexthop = new->attrs->gw;
			re->iface = new->attrs->iface;

			/* check if we can install given route */
			if (re->flags & LDP_RE_LDP)
			{
				/* Allocate and announce label */
				ldp_rt_install_re(pl, re);
			}
		}
	}

	UNDEBUG_PROTO();
}

/*
 * Handle FEC update from LDP lmap message
 */
void
ldp_rt_handle_fec(struct proto_ldp *pl, ip_addr addr, int plen, int label, ldp_session *ls, int withdraw)
{
	ldp_rib_entry *re;

	re = ldp_rib_findlink(pl, &addr, plen, ls);

	/*
	 * Remote host withdraws label for given FEC.
	 */
	if (withdraw)
	{
		/* Matched RE not found, do nothing */
		if (!re)
			return;

		/* withdraw prefix and label */
		ldp_rt_withdraw_re(pl, re);
		/* Unlink re from given session */
		ldp_rib_unlink_session(pl, re);
		/* re may be unused now, let's check and delete */
		ldp_rib_gc_entry(pl, re);
		return;
	}

       /* New or updated RIB entry */

	if (re != NULL)
	{
		/* Check if anything is changed */
		if (re->me.stack[0] == label)
			return;
	} else {
		/* Allocate new re linked to given session */
		re = ldp_rib_add_entry(pl, &addr, plen, ls);
	}

	/* Update/set remote label */
	re->me.depth = 1;
	re->me.stack[0] = label;

	/*
	 * Do nothing if we're not the best entry.
	 * Be paranoid: check if @re is not for directly connected route.
	 */
	if ((re->flags & (LDP_RE_BEST|LDP_RE_NEXTHOP)) != (LDP_RE_BEST|LDP_RE_NEXTHOP))
	{
		LOG_SS_DBG("remote label %s for %F: INACTIVE",
			krt_print_mplslbl(label), FPREFIX(&re->rg->n));
		return;
	}

	/*
	 * Best route update or installation.
	 * RFC 5036 recommends us to use delete/install update method:
	 * See A.1.2 and LMp.10a for more info.
	 */

	LOG_SS_DBG("remote label %s for %F: INSTALLING",
		krt_print_mplslbl(label), FPREFIX(&re->rg->n));

	/* Withdraw old resources if set */
	ldp_rt_withdraw_re(pl, re);
	
	/* .. And install new */
	ldp_rt_install_re(pl, re);
}

/**
 * ldp_session_install_prefix - creates local label mapping for given prefix and send updated routes to kernel
 * @pl: pointer to LDP protocol
 * @re: pointer to RIB entry
 * @n: pointer to prefix
 * flags: announcement flags
 *
 * Creates local label mapping for given prefix, links @re with new mapping, adds label to session label list.
 * if LDP_FLAG_INSTALL_RIB is set, kernel route update for IP prefix is executed.
 * if LDP_FLAG_INSTALL_LFIB is set, label is added to AF_MPLS family.
 * Assumes all data already checked. @re MUST contain:
 * * remote label in me.stack[0]
 * * valid pointer to LDP session
 * * valid next hop set
 *
 * Returns standart LDP message dispathing status
 *
 */
static int
ldp_rt_install_re(struct proto_ldp *pl, ldp_rib_entry *re)
{
	ldp_rib_group *rg = re->rg;
	ldp_session *ls = re->ls;

	/* Allocate new label */
	if (ldp_rt_request_label(pl, re) != LDP_OK)
	{
		LOG_DB_ERR("Label allocation failed for %F", FPREFIX(&rg->n));
		return LDP_ERR;
	}

	/*
	 * Announce given label.
	 * Old label (if any) gets destroyed by KRT code automatically.
	 */

	if (ldp_rt_announce_label(pl, re) != LDP_OK)
	{
		LOG_DB_ERR("Label %s installation failed for %F",
			krt_print_mplslbl(re->me.local_label), FPREFIX(&rg->n));
		return LDP_ERR;
	}

	/* Everything OK, announce LMAP */
	ldp_session_announce_lmap(pl, &re->rg->n, re->me.local_label, ls, LM_NONE, 0);

	/* Add labeled route to IGP table */
	if (re->me.stack[0] != MPLS_LABEL_IMPLNULL)
	{
		LOG_DB_DBG("Installing remote label %s for %F",
			krt_print_mplslbl(re->me.stack[0]), FPREFIX(&rg->n));

		ldp_rt_announce_rte(pl, re);
	}

	return LDP_OK;
}

static int
ldp_rt_request_label(struct proto_ldp *pl, ldp_rib_entry *re)
{
  struct service_request *req;
  struct krt_mpls_label_request lr;
  struct krt_mpls_label_reply lrepl;
  size_t buflen;
  byte *rbuf;

  req = &lr.req;

  req->request = SRCLASS_LABEL;
  req->subclass = SRSUB_LABEL_REQ;
  req->t = pl->mpls_table;
  req->p = &pl->proto;
  req->target = pl->lib_proto;
  req->relfunc = NULL;

  /*
   * We perform label request in single case:
   * when remote next-hop is LDP-capable and
   * remote label is not implicit/explicit NULL.
   */
  lr.op = KMPLS_OP_PUSH;
  lr.flags = 0;
  lr.data = 0;

  memset(&lrepl, 0, sizeof(lrepl));
  buflen = sizeof(lrepl);
  rbuf = (byte *)&lrepl;

  LOG_DBG("Requesting label..");

  if (proto_request_service((struct service_request *)&lr, &rbuf, &buflen) != SR_OK)
    return LDP_ERR;

  LOG_DBG("LIB database returned label %d", lrepl.label);

  re->me.local_label = lrepl.label;

  return LDP_OK;
}

/*
 * ldp_rt_withdraw_re - remove @re prefix & label from route tables
 * @pl: pointer to LDP protocol instance
 * @re: LDP rib entry
 */
void
ldp_rt_withdraw_re(struct proto_ldp *pl, ldp_rib_entry *re)
{

#if 0
	/* Remote label change or withdraw  */
	LOG_DB_DBG("Withdrawing RIB entry %F with label %s",
		FPREFIX(&rg->n), krt_print_mplslbl(re->me.local_label));
#endif
	/* Withdraw RIB announce */
	if (re->flags & LDP_RE_INSTALLED)
		ldp_rt_withdraw_rte(pl, re);

	/* Remove local label */
	if (re->flags & LDP_RE_HASLABEL)
	{
		/* Announce label being released to other neigbours. */
		ldp_session_announce_lmap(pl, &re->rg->n,
			re->me.local_label, re->ls, LM_NONE, 1);
		/* Withdraw label from MPLS fib */
		ldp_rt_withdraw_label(pl, re);
	}
}

/*
 * Install @re to lfib table. Set  HASLABEL flag.
 */
static int
ldp_rt_announce_label(struct proto_ldp *pl, ldp_rib_entry *re)
{
	rta *a, A;
	net *n;
	bird_mpls_entry lme, *m;
	u32 label;
	eattr *ea;
	struct adata *ad;
	ea_list *tmpa;
	rte *e;
	
	label = re->me.local_label;
/*
	if ((label < LDP_LABEL_MIN) || (label > LDP_LABEL_MAX))
	{
		LOG_DB_ERR("Label %s insert request is invalid. Ignoring", buf);
		return;
	}
*/

	/* Prepare attribute */
	bzero(&A, sizeof(A));
	A.proto = &pl->proto;
	A.source = RTS_LDP;
	A.scope = SCOPE_UNIVERSE;
	A.cast = RTC_UNICAST;
	A.dest = RTD_ROUTER;
	A.flags = 0;
	A.gw = re->nexthop;
	A.iface = re->iface;
	/* Prepare label extended attribute */
	
	/* Allocate temporary attribute for LA_MPLSENTRY */	
	tmpa = alloca(sizeof(ea_list) + sizeof(eattr));
	memset(tmpa, 0, sizeof(ea_list) + sizeof(eattr));
	tmpa->count = 1;

	/* Setup external storage for sockaddr  */
	ad = alloca(sizeof(struct adata) + sizeof(struct bird_mpls_entry));
	ad->length = sizeof(struct bird_mpls_entry);
	m = (bird_mpls_entry *)(ad + 1);

	/* fill in valid label structure */
	memcpy(m, &re->me, sizeof(struct bird_mpls_entry));
	m->local_label = label;
	if (re->me.stack[0] == MPLS_LABEL_IMPLNULL)
	{
		m->depth = 0;
		m->action = KMPLS_OP_POP;
	}
	else
	{
		m->depth = 1;
		m->action = KMPLS_OP_SWAP;
	}

	/* Finish attribute setup */
	ea = (eattr *)(tmpa + 1);
	ea->id = EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL);
	ea->type = EAF_TEMP | EAF_TYPE_OPAQUE | EAF_ORIGINATED;
	ea->u.ptr = ad;

	A.eattrs = tmpa;

	a = rta_lookup(&A);

	/* Get new prefix */
	memset(&lme, 0, sizeof(lme));
	lme.local_label = label;
	n = fib_get(&pl->mpls_table->fib, &lme, 0);

	e = rte_get_temp(a);
	e->net = n;
	e->pflags = 0;

	/* Setup protocol-specific attributes */
	e->u.ldp.router_id = re->ls->lsr_id;
	e->u.ldp.ltype = m->action; /* XXX: convert to wider class */

	LOG_DB_DBG("Installing label %s with nexthop %I",
		krt_print_mplslbl(label), re->nexthop);

	/* Do announce */
	rte_update2(pl->mpls_ahook, n, e, &pl->proto);

	re->flags |= LDP_RE_HASLABEL;

	return LDP_OK;
}

/*
 * Withdraw @re label from lfib table. Remove HASLABEL flag.
 */
static void
ldp_rt_withdraw_label(struct proto_ldp *pl, ldp_rib_entry *re)
{
	net *n;
	u32 label = re->me.local_label;
	bird_mpls_entry lme;

	re->me.local_label = 0;
	re->flags &= ~LDP_RE_HASLABEL;

	if ((label < LDP_LABEL_MIN) || (label > LDP_LABEL_MAX))
	{
		LOG_DB_ERR("Label %u withdraw request is invalid. Ignoring", label);
		return;
	}

	memset(&lme, 0, sizeof(lme));
	lme.local_label = label;
	if ((n = fib_find(&pl->mpls_table->fib, &lme, 0)) == NULL)
	{
		LOG_DB_VRB("Label %u does not exist in LFIB, unable to withdraw", label);
		return;
	}

	/* Do withdraw*/
	rte_update2(pl->mpls_ahook, n, NULL, &pl->proto);
}

/*
 * Announce @rte to nexthop table. Set INSTALLED flag.
 */
static void
ldp_rt_announce_rte(struct proto_ldp *pl, ldp_rib_entry *re)
{
	rta *a, A;
	net *n;
	bird_mpls_entry *m;
	eattr *ea;
	struct adata *ad;
	ea_list *tmpa;
	rte *e;
	u32 label = re->me.stack[0];

	if (label == MPLS_LABEL_IMPLNULL)
	{
		LOG_DB_VRB("Not installing labeled prefix %F: remote label is %s",
			&re->rg->n, krt_print_mplslbl(label));
		return;
	}

	/* Prepare attribute */
	bzero(&A, sizeof(A));
	A.proto = &pl->proto;
	A.source = RTS_LDP;
	A.scope = SCOPE_UNIVERSE;
	A.cast = RTC_UNICAST;
	A.dest = RTD_ROUTER;
	A.flags = 0;
	A.gw = re->nexthop;
	A.iface = re->iface;
	/* Prepare label extended attribute */
	
	/* Allocate temporary kernel MPLS attribute */
	tmpa = alloca(sizeof(ea_list) + sizeof(eattr));
	memset(tmpa, 0, sizeof(ea_list) + sizeof(eattr));
	tmpa->count = 1;

	/* Setup external storage for sockaddr  */
	ad = alloca(sizeof(struct adata) + sizeof(struct bird_mpls_entry));
	ad->length = sizeof(struct bird_mpls_entry);
	m = (struct bird_mpls_entry *)(ad + 1);

	/* Setup MPLS attribute */
	memcpy(m, &re->me, sizeof(bird_mpls_entry));
	m->local_label = 0;
	m->action = KMPLS_OP_PUSH;

	/* Finish attribute setup */
	ea = (eattr *)(tmpa + 1);
	ea->id = EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL);
	ea->type = EAF_TEMP | EAF_TYPE_OPAQUE | EAF_ORIGINATED;
	ea->u.ptr = ad;

	A.eattrs = tmpa;

	a = rta_lookup(&A);

	n = fib_get(&pl->route_table->fib, FPREFIX(&re->rg->n), re->rg->n.pxlen);

	e = rte_get_temp(a);
	e->net = n;
	e->pflags = 0;

	e->u.ldp.router_id = 0;
	e->u.ldp.ltype = 0;

	LOG_DB_DBG("Installing label %s for %F with nexthop %I",
		krt_print_mplslbl(label), &re->rg->n, re->nexthop);

	/* Do announce */
	rte_update2(pl->route_ahook, n, e, &pl->proto);

	re->flags |= LDP_RE_INSTALLED;
}

/*
 * Withdraw @re from nexthop table. Remove INSTALLED flag.
 */
static void
ldp_rt_withdraw_rte(struct proto_ldp *pl, ldp_rib_entry *re)
{
	ldp_rib_group *rg = re->rg;
	net *n;

	n = fib_find(&pl->route_table->fib, FPREFIX(&rg->n), rg->n.pxlen);
	if (n && n->routes)
		rte_update2(pl->route_ahook, n, NULL, &pl->proto);

	re->flags &= ~LDP_RE_INSTALLED;
}


