/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * LDP TLV conversion/generation routines.
 *
 * For conversion TLVS: we're getting data in network format
 * and linpool to allocate from. We return new TLV in host format.
 * In some cases like FEC we convert all prefixes in host format
 * structures and put them immediately after new TLV
 *
 */

#undef LOCAL_DEBUG

#include "ldp.h"
#include "ldp_proto.h"
#include "lib/krt_mpls.h"

static int ldp_tlv_get_ipv4transport(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);
static int ldp_tlv_get_chello_params(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);
static int ldp_tlv_get_csess_params(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);
static int ldp_tlv_get_status_params(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);
static int ldp_tlv_get_addrlist(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);
static int ldp_tlv_get_fec(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);
static int ldp_tlv_get_generic_label(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);
static int ldp_tlv_get_unknown(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv);


#define TLV_CHECK_LENGTH(x)	{ if (lth->length + LDP_TLV_HEADER_SIZE != (x)) return 0; }

void
ldp_tlv_put_header(byte *data, int unknown, int forward, int type, int length)
{
	u8 *x = (u8 *)data;
	unknown = (unknown != 0) ? 1 : 0;
	forward = (forward != 0) ? 1 : 0;
	*x++ = (unknown << 7) | (forward << 6) | ((type >> 8) & 0x3F);
	*x++ = type & 0xFF;
	put_u16(data + 2, length - LDP_TLV_HEADER_SIZE);
}

void
ldp_tlv_get_header(byte *data, ldp_tlv_header *lth)
{
	u8 *x = (u8 *)data;
	int unknown, forward, type;

	unknown = *x >> 7;
	forward = (*x >> 6) & 0x01;
	type = (*x & 0x3F) << 8;
	type += *++x;

	lth->unknown = unknown;
	lth->forward = forward;
	lth->type = type;
	lth->length = get_u16(data + 2);

	if (DUMP_PACKET)
	{
		TDUMP(" %s TLV, flags: %c%c, length: %d", ldp_tlv_texttype(lth->type), 
			lth->unknown ? 'U' : '_', lth->forward ? 'F' : '_', lth->length);
		TLINE();
	}
}

int
ldp_tlv_put_ipv4transport(byte *data, struct proto_ldp *pl)
{
	ldp_tlv_put_header(data, 0, 0, LDP_TLV_IP4TRADDRESS, LDP_TLV_IPV4TRANSPORT_SIZE);
	put_u32(data + LDP_TLV_HEADER_SIZE, pl->tcp_saddr);

	return LDP_TLV_IPV4TRANSPORT_SIZE;
}

/**
 * ldp_tlv_get_ipv4transport - get IPv4 Transport Address TLV
 *
 * @data: pointer to byffer to read from
 * @lth: converted TLV header
 * @lp: linpool to allocate from
 * @_tlv: pointer to put allocated structure
 *
 */
int
ldp_tlv_get_ipv4transport(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	ldp_tlv_ipv4transport *tlv;

	TLV_CHECK_LENGTH(LDP_TLV_IPV4TRANSPORT_SIZE);

	/* Allocate data */
	if ((tlv = lp_alloc(lp, sizeof(*tlv))) == NULL)
		return LDP_OBJ_FAIL; /* Do not return fatal error not to force session closing */

	memcpy(&tlv->_base, lth, sizeof(ldp_tlv_header));
	tlv->addr = get_u32(data + LDP_TLV_HEADER_SIZE);
	if (DUMP_PACKET)
	{
		TDUMP("  Address: %I", ipa_from_u32(tlv->addr));
		TLINE();
	}

	*_tlv = (ldp_tlv *)tlv;

	return LDP_OBJ_OK;
}

/*                        Common Hello Parameters TLV
 *
 *      0                   1                   2                   3
 *      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |0|0| Common Hello Parms(0x0400)|      Length                   |
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |      Hold Time                |T|R| Reserved                  |
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */



/**
 * ldp_tlv_put_chello_params - write Common Hello Parms header
 *
 * @data: pointer to byffer to write to
 * @pl: poingter to protocol instance
 * @targeted: T/R bit flags
 *
 * Returns bytes written
 */
int
ldp_tlv_put_chello_params(byte *data, struct proto_ldp *pl, int targeted)
{
	u8 *x;
	struct ldp_config *cf = PL_TO_CF(pl);

	ldp_tlv_put_header(data, 0, 0, LDP_TLV_CHELLOPAR, LDP_TLV_CHELLO_PARAMS_SIZE);
	put_u16(data + LDP_TLV_HEADER_SIZE, targeted ? cf->thold_interval : cf->dhold_interval);

	x = (u8 *)(data + LDP_TLV_HEADER_SIZE + 2);
	*x++ = (targeted & 0x3) << 6;
	*x = 0; /* Reserved field */

	return LDP_TLV_CHELLO_PARAMS_SIZE;
}

/**
 * ldp_tlv_get_chello_params - get Common Hello Parameters TLV
 *
 * @data: pointer to byffer to read from
 * @lth: converted TLV header
 * @lp: linpool to allocate from
 * @_tlv: pointer to put allocated structure
 *
 */
int
ldp_tlv_get_chello_params(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	u8 *x;
	ldp_tlv_chello_params *tlv;

	TLV_CHECK_LENGTH(LDP_TLV_CHELLO_PARAMS_SIZE);

	/* Allocate data */
	if ((tlv = lp_alloc(lp, sizeof(*tlv))) == NULL)
		return LDP_OBJ_FAIL; /* Do not return fatal error not to force session closing */

	memcpy(&tlv->_base, lth, sizeof(ldp_tlv_header));
	tlv->hold_time = get_u16(data + LDP_TLV_HEADER_SIZE);
	x = (u8 *)(data + LDP_TLV_HEADER_SIZE + 2);
	tlv->thello =  *x >> 7;
	tlv->rthello =  (*x >> 6) & 0x01;
	tlv->reserved = (*x++ & 0x3F) << 8;
	tlv->reserved += *x;

	if (DUMP_PACKET)
	{
		TDUMP("  Hold: %d, flags: %s%s", tlv->hold_time, tlv->thello ? "T" : "", tlv->rthello ? "R" : "");
		TLINE();
	}

	*_tlv = (ldp_tlv *)tlv;

	return LDP_OBJ_OK;
}

/*                        Common Session Parameters TLV
 *
 *      0                   1                   2                   3
 *      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |0|0| Common Sess Parms (0x0500)|      Length                   |
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     | Protocol Version              |      KeepAlive Time           |
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |A|D|  Reserved |     PVLim     |      Max PDU Length           |
 *     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |                 Receiver LDP Identifier                       |
 *     +                               +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *     |                               |
 *     -+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-++
 */

/**
 * ldp_tlv_put_csess_params - write Common Session Parms header
 *
 * @data: pointer to byffer to write to
 * @pl: poingter to protocol instance
 * @targeted: T/R bit flags
 *
 * Returns bytes written
 */
int
ldp_tlv_put_csess_params(byte *data, ldp_session *ls)
{
	u8 *x;

	ldp_tlv_put_header(data, 0, 0, LDP_TLV_CSESSPAR, LDP_TLV_CSESS_PARAMS_SIZE);

	put_u16(data + LDP_TLV_HEADER_SIZE, LDP_PROTOCOL_VERSION);
	put_u16(data + LDP_TLV_HEADER_SIZE + 2, ls->keep_interval);
	x = (u8 *)(data + LDP_TLV_HEADER_SIZE + 4);
	*x++ = (LDP_LABEL_ADV_UNSOLICITED ? 0 : 1 << 7); /* XXX: No loop detection */
	*x = 0; /* PVLim, 0 if loop detec disabled */
	put_u16(data + LDP_TLV_HEADER_SIZE + 6, ls->max_pdu);
	put_u32(data + LDP_TLV_HEADER_SIZE + 8, ls->lsr_id);
	put_u16(data + LDP_TLV_HEADER_SIZE + 12, ls->space_id);

	return LDP_TLV_CSESS_PARAMS_SIZE;
}

/**
 * ldp_tlv_get_csess_params - get Common Session Parameters TLV
 *
 * @data: pointer to byffer to read from
 * @lth: converted TLV header
 * @lp: linpool to allocate from
 * @_tlv: pointer to put allocated structure
 *
 */
static int
ldp_tlv_get_csess_params(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	u8 *x;
	ldp_tlv_csess_params *tlv;

	TLV_CHECK_LENGTH(LDP_TLV_CSESS_PARAMS_SIZE);

	/* Allocate data */
	if ((tlv = lp_alloc(lp, sizeof(*tlv))) == NULL)
		return LDP_OBJ_FAIL; /* Do not return fatal error not to force session closing */

	memcpy(&tlv->_base, lth, sizeof(ldp_tlv_header));
	tlv->pversion = get_u16(data + LDP_TLV_HEADER_SIZE);
	tlv->keep_interval = get_u16(data + LDP_TLV_HEADER_SIZE + 2);
	x = (u8 *)(data + LDP_TLV_HEADER_SIZE + 4);
	tlv->adiscipline = *x >> 7;
	tlv->loopdetect = (*x >> 6) & 0x01;
	tlv->reserved = *x & 0x3F;
	tlv->pvlimit = *(x + 1);
	tlv->max_pdu = get_u16(data + LDP_TLV_HEADER_SIZE + 6);
	tlv->lsr_id = get_u32(data + LDP_TLV_HEADER_SIZE + 8);
	tlv->space_id = get_u16(data + LDP_TLV_HEADER_SIZE + 12);

	if (DUMP_PACKET)
	{
		TDUMP("  Version: %d, Keepalive: %d, flags: %s%s", tlv->pversion, tlv->keep_interval, tlv->adiscipline ? "D" : "U", tlv->loopdetect ? "L" : "");
		TLINE();
		TDUMP("  MAXPDU: %d LSR: %I:%d", tlv->max_pdu, tlv->lsr_id, tlv->space_id);
		TLINE();
	}

	*_tlv = (ldp_tlv *)tlv;

	return LDP_OBJ_OK;
}

/*                  STATUS TLV
 *
 *    0                   1                   2                   3
 *    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |U|F| Status (0x0300)           |      Length                   |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |                     Status Code                               |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |                     Message ID                                |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |      Message Type             |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

/**
 * ldp_tlv_put_status_params - write Status TLV
 *
 * @data: pointer to byffer to write to
 * @ls: pointer to LDP session
 *
 * Returns bytes written
 */
int
ldp_tlv_put_status_params(byte *data, ldp_session *ls)
{
	ldp_tlv_put_header(data, 0, 0, LDP_TLV_STATUS, LDP_TLV_STATUS_PARAMS_SIZE);

	return LDP_TLV_CSESS_PARAMS_SIZE;
}

/**
 * ldp_tlv_get_status_params - get Status TLV
 *
 * @data: pointer to byffer to read from
 * @lth: converted TLV header
 * @lp: linpool to allocate from
 * @_tlv: pointer to put allocated structure
 *
 */
static int
ldp_tlv_get_status_params(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	u8 *x;
	u32 statuscode;
	ldp_tlv_status_params *tlv;

	TLV_CHECK_LENGTH(LDP_TLV_STATUS_PARAMS_SIZE);

	/* Allocate data */
	if ((tlv = lp_alloc(lp, sizeof(*tlv))) == NULL)
		return LDP_OBJ_FAIL; /* Do not return fatal error not to force session closing */

	memcpy(&tlv->_base, lth, sizeof(ldp_tlv_header));

	x = (u8 *)(data + LDP_TLV_HEADER_SIZE);
	tlv->fatal = *x >> 7;
	tlv->forward = (*x >> 6) & 0x01;
	statuscode = (*x++ & 0x3F) << 24;
	statuscode += *x++ << 16;
	statuscode += *x++ << 8;
	statuscode += *x;
	tlv->statuscode = statuscode;

	tlv->msgid = get_u32(data + LDP_TLV_HEADER_SIZE + 4);
	tlv->msgtype = get_u16(data + LDP_TLV_HEADER_SIZE + 8);


	if (DUMP_PACKET)
	{
		TDUMP("  Flags: %s%s, Code: %d", tlv->fatal ? "E" : "", tlv->forward ? "F" : "", tlv->statuscode);
		TLINE();
		TDUMP(" MsgId: %X MsgType: %s (%X)", tlv->msgid, tlv->msgtype > 0 ? ldp_tlv_texttype(tlv->msgtype) : "", tlv->msgtype);
		TLINE();
	}

	*_tlv = (ldp_tlv *)tlv;

	return LDP_OBJ_OK;
}

/*  Address List TLV
 *  
 *    0                   1                   2                   3
 *    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |0|0| Address List (0x0101)     |      Length                   |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |     Address Family            |                               |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               |
 *   |                                                               |
 *   |                        Addresses                              |
 *   ~                                                               ~
 *   |                                                               |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

/**
 * ldp_tlv_put_addrlist - write Address List TLV
 *
 * @data: pointer to buffer to write to
 * @datasize: size of buffer
 * @af: IANA address family of nexthops following
 * @addrbuf: pointer to array of our addresses
 * @addrbufsize: array size
 * @bufused: variable to store used data amount, can be NULL
 *
 * Returns bytes written
 */
int
ldp_tlv_put_addrlist(byte *data, size_t datasize, int af, byte *addrbuf, size_t addrbufsize, size_t *bufused)
{
	int af_size = (af == IANA_AFI_IP6) ? 16 : 4;

	if (datasize < LDP_TLV_ADDRLIST_SIZE)
		return 0;

	datasize -= LDP_TLV_ADDRLIST_SIZE;

	/* Let's check how much we can send */
	datasize = MIN(datasize, addrbufsize);
	/* Check alignment */
	if ((datasize % af_size) != 0)
		datasize -= datasize % af_size;

	/* Write TLV data header */
	put_u16(data + LDP_TLV_HEADER_SIZE, af);
	memcpy(data + LDP_TLV_HEADER_SIZE + 2, addrbuf, datasize);

	if (bufused)
		*bufused = datasize;

	ldp_tlv_put_header(data, 0, 0, LDP_TLV_ADDRLIST, LDP_TLV_ADDRLIST_SIZE + datasize);
	return LDP_TLV_ADDRLIST_SIZE + datasize;
}

/**
 * ldp_tlv_get_addrlist - get Address list TLV
 *
 * @data: pointer to buffer to read from
 * @lth: converted TLV header
 * @lp: linpool to allocate from
 * @_tlv: pointer to put allocated structure
 *
 *
 * Address data is located at BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN) offset
 *
 */
static int
ldp_tlv_get_addrlist(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	ldp_tlv_addrlist *tlv;
	int length, iana_af, af_size, count = 0;
	u32 *addr;

	if (lth->length + LDP_TLV_HEADER_SIZE < LDP_TLV_ADDRLIST_SIZE)
		return LDP_OBJ_FAIL;

	iana_af = get_u16(data + LDP_TLV_HEADER_SIZE);

	af_size = (iana_af == IANA_AFI_IP6) ? 16 : 4;
	length = lth->length - 2;

	/* XXX: This can change after bird[6] merge */
	if (af_size != sizeof(ip_addr))
	{
		LOG_SS_VRB("Received Unsupported address family in Addresslist TLV");
		return LDP_OBJ_FAIL;
	}

	if (length % af_size)
	{
		LOG_SS_VRB("Incorrect Addresslist TLV received");
		return LDP_OBJ_FAIL;
	}

	/* Allocate data */
	if ((tlv = lp_alloc(lp, BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN) + length)) == NULL)
		return LDP_OBJ_FAIL; /* Do not return fatal error not to force session closing */

	memcpy(&tlv->_base, lth, sizeof(ldp_tlv_header));
	tlv->iana_af = iana_af;
	data += LDP_TLV_HEADER_SIZE + 2;

	addr = (u32 *)((byte *)tlv + BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN));

	while (length > 0)
	{
		*addr++ = get_u32(data);
		data += af_size;
		length -= af_size;
		count++;
	}

	if (DUMP_PACKET)
	{
		TDUMP("  Family: %s, Address count: %d", (iana_af == IANA_AFI_IP6) ? "AF_INET6" : "AF_INET", count);
		TLINE();

		addr = (u32 *)((byte *)tlv + BIRD_ALIGN(sizeof(ldp_tlv_addrlist), CPU_STRUCT_ALIGN));
		while (count > 0)
		{
			TDUMP("   Address: %I", *addr++);
			TLINE();
			count--;
		}
	}

	*_tlv = (ldp_tlv *)tlv;

	return LDP_OBJ_OK;
}

/*                          Prefix FEC Element
 *
 *     0                   1                   2                   3
 *     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |  Prefix (2)   |     Address Family            |     PreLen    |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |                     Prefix                                    |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

/**
 * ldp_tlv_fec_get_prefix - Extract prefix from data buffer
 * @data: pointer to data buffer
 * @plen: prefix length
 * @dataread: value to store amount of data read to
 *
 * Assume this is not wildcard prefix. Assume ip_addr
 * is from the same AF
 *
 */
ip_addr
ldp_tlv_fec_get_prefix(byte *data, int plen, int *dataread)
{
	ip_addr addr;
	plen = (plen / 8) + ((plen % 8) ? 1 : 0);
	_ZERO(addr);
	if (plen <= sizeof(ip_addr))
		memcpy(&addr, data + 4, plen);
	ipa_ntoh(addr);

	*dataread = plen + 4;

	return addr;
}

/**
 * ldp_tlv_fec_put_prefix -- write IPv4 prefix in FEC format
 *
 * @data: data buffer
 * @datasize: data buffer length
 * @addr: prefix
 * @mask: prefix mask
 *
 */
int
ldp_tlv_fec_put_prefix(byte *data, int datasize, ip_addr addr, int mask)
{
	int len = (mask / 8) + ((mask % 8) ? 1 : 0);

	/* Check if we have sufficient space */
	if (datasize < len + 4)
		return 0;

	/* Fill in header */
	*data = LDP_FEC_ELEMENT_PREFIX;
	put_u16(data + 1, IANA_AFI_IP);
	*(data + 3) = mask;

	ipa_hton(addr);
	memcpy(data + 4, &addr, len);

	return len + 4;
}

/*                               FEC TLV
 *
 *     0                   1                   2                   3
 *     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |0|0| FEC (0x0100)              |      Length                   |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |                        FEC Element 1                          |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |                                                               |
 *    ~                                                               ~
 *    |                                                               |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |                        FEC Element n                          |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

/**
 * ldp_tlv_put_fec - write FEC TLV header
 *
 * @data: pointer to buffer to write to
 * @length: address data length
 *
 * Returns total TLV length
 */
int
ldp_tlv_put_fec(byte *data, int length)
{
	ldp_tlv_put_header(data, 0, 0, LDP_TLV_FEC, LDP_TLV_FEC_SIZE + length);
	return LDP_TLV_FEC_SIZE + length;
}

/**
 * ldp_tlv_get_fec - get FEC elememtn
 *
 * @data: pointer to buffer to read from
 * @lth: converted TLV header
 * @lp: linpool to allocate from
 * @_tlv: pointer to put allocated structure
 *
 * list of FECs is located at BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN) offset
 *
 */
static int
ldp_tlv_get_fec(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	ldp_tlv_fec *tlv;
	int length, plen, len, family, count = 0;
	ldp_tlv_fec_element *fe;
	ip_addr addr;

	if (lth->length == 0)
		return LDP_OBJ_FAIL;

	/*
	 * We're allocating single buffer for header and all FEC elements
	 * in host format at once. Predict how much we need to allocate:
	 * Minimal (non-wildcard FEC) consimes 5 bytes (4 header and 1 data).
	 * So approppriate memory size is: (length / 5 + 1) * sizeof(ldp_tlv_fec_element)
	 * This gives ratio of 1.6 for IPv4 and 6.4 for IPv6. Assuming LMAP full of FEC
	 * and default size of 4096 lp should be at least 8192 for effective operations
	 *
	 */

	/* Allocate data */
	if ((tlv = lp_alloc(lp, BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN) + (lth->length / 5 + 1) * sizeof(ldp_tlv_fec_element))) == NULL)
		return LDP_OBJ_FAIL; /* Do not return fatal error not to force session closing */

	memcpy(&tlv->_base, lth, sizeof(ldp_tlv_header));

	fe = (ldp_tlv_fec_element *)((byte *)tlv + BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN));
	data += LDP_TLV_FEC_SIZE;
	length = lth->length;

	/* Check special case: wildcard prefix */
	if (*data == LDP_FEC_ELEMENT_WILD)
	{
		memset(fe, 0, sizeof(ldp_tlv_fec_element));
		fe->type = LDP_FEC_ELEMENT_WILD;

		if (length != 1)
		{
			/* Must be the only FEC */
			return LDP_OBJ_FAIL;
		}

		tlv->count = 1;
		return LDP_OBJ_OK;
	}

	/* Cycle thru all data */
	while (length >= 4)
	{
		if (*data != LDP_FEC_ELEMENT_PREFIX)
		{
			/* XXX: Notify lower layer */
			return LDP_OBJ_FAIL;
		}

		if ((family = get_u16(data + 1)) != IANA_AFI_IP)
		{
			/* XXX: We support IPv4 only at me moment */
			return LDP_OBJ_FAIL;
		}

		plen = *(data + 3);
		len = (plen / 8) + ((plen % 8) ? 1 : 0);
		if (len > 4)
		{
			/* XXX: IPv4 */
			return LDP_OBJ_FAIL;
		}

		/* Fill in resulting FE structure */
		fe->plen = plen;
		fe->family = family;
		fe->type = LDP_FEC_ELEMENT_PREFIX;
		memset(&addr, 0, sizeof(addr));
		memcpy(&addr, data + 4, len);
		ipa_ntoh(addr);

		/* Check if prefix is valid. */
		if (addr != ipa_and(addr, ipa_mkmask(plen)))
		{
			/* Bad combination of prefix & mask */
			return LDP_OBJ_FAIL;
		}

		fe->prefix = addr;

		/* Prepare for next cycle.. */
		fe++;
		count++;
		length -= len + 4;
	}

	if (length != 0)
	{
		/* XXX: we have to warn lower layer here */
		return LDP_OBJ_FAIL;
	}

	tlv->count = count;

	if (DUMP_PACKET)
	{
		TDUMP("  FEC count: %d", count);
		TLINE();

		fe = (ldp_tlv_fec_element *)((byte *)tlv + BIRD_ALIGN(sizeof(*tlv), CPU_STRUCT_ALIGN));

		while (count > 0)
		{
			TDUMP("   Prefix: %I/%d", fe->prefix, fe->plen);
			TLINE();
			fe++;
			count--;
		}
	}

	*_tlv = (ldp_tlv *)tlv;

	return LDP_OBJ_OK;
}

/*                          Generic Label TLV
 *
 *     0                   1                   2                   3
 *     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |0|0| Generic Label (0x0200)    |      Length                   |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    |     Label                                                     |
 *    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */

/**
 * ldp_tlv_put_generic_label - write Generic Label TLV
 *
 * @data: pointer to buffer to write to
 * @datasize: size of buffer
 * @label: MPLS label to write
 *
 * Returns bytes written
 */
int
ldp_tlv_put_generic_label(byte *data, size_t datasize, int label)
{

	if (datasize < LDP_TLV_GENERIC_LABEL_SIZE)
		return 0;

	/*
	 * XXX: it is completely undocumented in RFC3036 and RFC5036
	 * but cisco (at least) assumes label not to follow RFC3032 encoding.
	 * Field is read as uin32_t in network format
	 */
	put_u32(data + LDP_TLV_HEADER_SIZE, label);
	ldp_tlv_put_header(data, 0, 0, LDP_TLV_GLABEL, LDP_TLV_GENERIC_LABEL_SIZE);

	return LDP_TLV_GENERIC_LABEL_SIZE;
}

/**
 * ldp_tlv_get_ipv4transport - get IPv4 Transport Address TLV
 *
 * @data: pointer to byffer to read from
 * @lth: converted TLV header
 * @lp: linpool to allocate from
 * @_tlv: pointer to put allocated structure
 *
 */
static int
ldp_tlv_get_generic_label(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	ldp_tlv_generic_label *tlv;

	TLV_CHECK_LENGTH(LDP_TLV_GENERIC_LABEL_SIZE);

	/* Allocate data */
	if ((tlv = lp_alloc(lp, sizeof(*tlv))) == NULL)
		return LDP_OBJ_FAIL; /* Do not return fatal error not to force session closing */

	memcpy(&tlv->_base, lth, sizeof(ldp_tlv_header));
	tlv->label = get_u32(data + LDP_TLV_HEADER_SIZE);

	if (DUMP_PACKET)
	{
		TDUMP("  Label: %s", krt_print_mplslbl(tlv->label));
		TLINE();
	}

	*_tlv = (ldp_tlv *)tlv;

	return LDP_OBJ_OK;
}

#define MAX_BYTES 16
/**
 * ldp_tlv_get_unknown - unkhown TLV handling
 * @tlv - TLV header pointer (already converted to host format)
 *
 * ldp_parse_tlv() returns the following values:
 * 1 - U flag not set
 * 2 - U flags set
 *
 * Inspired by RFC5036 Clause 3.3
 *
 */
static int
ldp_tlv_get_unknown(byte *data, ldp_tlv_header *lth, linpool *lp, ldp_tlv **_tlv)
{
	char buf[512];
	u8 *x;
	int i, ii;

	if (DUMP_PACKET)
	{
		TDUMP("Unknown TLV, hexdump follows:");
		TLINE();
	
		i = 0;
		x = (u8 *)(data + LDP_TLV_HEADER_SIZE);
		while (i < lth->length)
		{
			_ZERO(buf);
			for (ii = 0; ((ii < MAX_BYTES) && (i < lth->length)); ii++, i++)
				snprintf(&buf[3 * ii], 4, "%02X ", x[i]);
			TDUMP("%s", buf);
			TLINE();
		}
	}

	if (lth->unknown)
		return LDP_OBJ_IGNORE;
	
	return LDP_OBJ_FAIL;
}

#undef MAX_BYTES

/**
 * ldp_tlv_parse - check TLV for valid data and convert fields to host format
 * @lpl: pointer to LDP protocol
 * @lth: TLV header pointer (already converted to host format)
 * @data: pointer to data buffer
 * @tlength: length of TLV
 * @tlv: pointer to pointer to TLV header
 *
 *
 * Function passes most of its parameters to TLV-specific parsing function which
 * allocated nesessary data size on pl linear pool and passes it back to caller
 *
 * ldp_tlv_parse() returns the following values:
 * 0 - Parsing done, TLV successfully converted to host format
 * 1 - Parsing failed (unknown TLV?), message needs to be ignored and notification has to be sent
 * 2 - Parsing failed, ignore TLV
 *
 */
int
ldp_tlv_parse(struct proto_ldp *pl, ldp_tlv_header *lth, byte *data, int length, ldp_tlv **tlv)
{
	int res;
	linpool *lp = pl->pdu_lp;

	switch (lth->type)
	{
		case LDP_TLV_CHELLOPAR:
			res = ldp_tlv_get_chello_params(data, lth, lp, tlv);
			break;

		case LDP_TLV_CSESSPAR:
			res = ldp_tlv_get_csess_params(data, lth, lp, tlv);
			break;
	
		case LDP_TLV_IP4TRADDRESS:
			res = ldp_tlv_get_ipv4transport(data, lth, lp, tlv);
			break;

		case LDP_TLV_STATUS:
			res = ldp_tlv_get_status_params(data, lth, lp, tlv);
			break;

		case LDP_TLV_ADDRLIST:
			res = ldp_tlv_get_addrlist(data, lth, lp, tlv);
			break;

		case LDP_TLV_FEC:
			res = ldp_tlv_get_fec(data, lth, lp, tlv);
			break;

		case LDP_TLV_GLABEL:
			res = ldp_tlv_get_generic_label(data, lth, lp, tlv);
			break;

		default:
			res = ldp_tlv_get_unknown(data, lth, lp, tlv);
			break;
	}

	return res;
}

ldp_tlv *
_ldp_tlv_find(list *l, int type)
{
	ldp_tlv *tlv;

	WALK_LIST(tlv, *l)
	{
		if (tlv->base.type == type)
			return tlv;
	}

	return NULL;
}

char *
ldp_tlv_texttype(int type)
{
	char *buf;
	static char xbuf[50];


	switch (type)
	{
		case LDP_TLV_FEC:
			buf = "FEC";
			break;
		case LDP_TLV_ADDRLIST:
			buf = "Address List";
			break;
		case LDP_TLV_HOPCOUNT:
			buf = "Hop Count";
			break;
		case LDP_TLV_PATHVECTOR:
			buf = "Path Vector";
			break;
		case LDP_TLV_GLABEL:
			buf = "Generic Label";
			break;
		case LDP_TLV_ATMLABEL:
			buf = "ATM Label";
			break;
		case LDP_TLV_FRLABEL:
			buf = "Frame Relay Label";
			break;
		case LDP_TLV_STATUS:
			buf = "Status";
			break;
		case LDP_TLV_EXTSTATUS:
			buf = "Extended Status";
			break;
		case LDP_TLV_RPDU:
			buf = "Returned PDU";
			break;
		case LDP_TLV_RMSG:
			buf = "Returned Message";
			break;
		case LDP_TLV_CHELLOPAR:
			buf = "Common Hello Params";
			break;
		case LDP_TLV_IP4TRADDRESS:
			buf = "IPv4 Transport Address";
			break;
		case LDP_TLV_CONFSEQ:
			buf = "Configuration Sequence Number";
			break;
		case LDP_TLV_IP6TRADDRESS:
			buf = "IPv6 Transport Address";
			break;
		case LDP_TLV_CSESSPAR:
			buf = "Common Session Params";
			break;
		case LDP_TLV_ATMSESSPAR:
			buf = "ATM Session Params";
			break;
		case LDP_TLV_FRSESSPAR:
			buf = "Frame Relay Session Params";
			break;
		case LDP_TLV_LREQMSGID:
			buf = "Label Request Message ID";
			break;
		case LDP_TLV_EXPROUTE:
			buf = "Explicit Route TLV";
			break;
		default:
			buf = xbuf;

			if ((type >= 0x3E00) && (type <= 0x3EFF))
			{
				snprintf(xbuf, sizeof(xbuf), "Vendor-Private: 0x%X", type);
				break;
			}

			if ((type >= 0x3F00) && (type <= 0x3FFF))
			{
				snprintf(xbuf, sizeof(xbuf), "Experimental: 0x%X", type);
				break;
			}

			snprintf(xbuf, sizeof(xbuf), "Unknown: 0x%X", type);
	}

	return buf;
}

