/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2013 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * MPLS label database (LIB) is managed here
 *
 *
 * MPLS table (lfib) and request/release API operations:
 *
 *
 * Open database:
 * Protocol (`user`) wishing to use MPLS labels initially requests access
 * to LIB database (`lib`) via SRSUB_LABEL_OPENDB request supplying its
 * `resource release` function pointer.
 *
 * If database is currently operating (e.g. krt protocol (`krt`) associated
 * with given table is UP and is not shutting down), SR_OK is returned.
 * If database is not operational (this can happen, for example, on daemon
 * startup) protocol should delay its startup, shedule timer to check if
 * database can be opened, and return PS_START. When lib is opened, protocol
 * should proceed startup.
 *
 * Request label:
 * User than saves krt pointer to use in all subsequent requests to lib.
 * When user needs new label it requests it via SRSUB_LABEL_REQ request to
 * lib. User protocol MUST announce allocated label immediately.
 *
 * Release label:
 * Label withdrawal is done automatically in krt rt_notify hook.
 * Route updates (if old & new label are the same) are handled correctly.
 *
 * Close database:
 * Prototol should close database at cleanup state (when routes, originated by
 * its protocol are flushed at _flush_ state).
 *
 *
 * krt protocol going down:
 * krt notifies each of its consumers using `resource release` function and waits
 * until all consumers closes database. Typically (for protocols which can
 * request large number of labels) it is better to request shutdown inside callback.
 * 
 * krt itself does the following:
 * * check if lib can be closed immediately (not started or no cunsumers) and does so if true.
 * * Shedule krt_shutdown_notify_hook hook and return PS_STOP
 * * Given hook executes protocols shutdown notification.
 * After that, protocol state changes are triggered by SRSUB_LABEL_CLOSEDB:
 * * When last consumer frees lib, krt_shutdown_finalize_hook hook is scheduled.
 */

#undef LOCAL_DEBUG

#include "nest/bird.h"
#include "nest/route.h"
#include "nest/protocol.h"
#include "lib/event.h"
#include "lib/alloca.h"

#include "unix.h"
#include "krt.h"
#include "krt_mpls.h"

static u32 krt_lib_hash(struct fib *f, void *a);
static u32 krt_ptr_hash(struct fib *f, void *a);
static krt_lib_proto *krt_lib_find_proto(krt_lib_database *lib, struct proto *p);
static u32 krt_lib_add_label(struct krt_proto *k, krt_lib_proto *lp, struct krt_mpls_label_request *lreq);
static int krt_lib_del_label(struct krt_proto *k, krt_lib_proto *lp, u32 label);
static int krt_dispatch_service_request(struct proto *p, struct service_request *req, byte **buf, size_t *buflen);

/*
 * Initialize LIB database
 */
int
krt_lib_init(struct krt_proto *k)
{
  krt_lib_database *lib;

  k->label_min = MPLS_LABEL_MIN;
  k->label_max = MPLS_LABEL_MAX;

  if ((lib = mb_alloc(k->p.pool, sizeof(krt_lib_database))) == NULL)
    {
      log(L_ERR "LIB database initialization failed!");
      return 0;
    }

  fib2_init(&lib->f, k->p.pool, sizeof(krt_lib_label), RT_MPLS, sizeof(u32),
  	0, NULL, krt_lib_hash);
  fib2_init(&lib->pr, k->p.pool, sizeof(krt_lib_proto), 0xFF, sizeof(uintptr_t),
  	0, NULL, krt_ptr_hash);

  lib->pr_count = 0;
  lib->current_label = k->label_min;
  init_list(&lib->free_items);

  k->lib = lib;

  /* Configure protocol as service provider */
  lib->sh = proto_add_service_hook(&k->p, SRCLASS_LABEL,
 	(service_func)&krt_dispatch_service_request);

  log(L_DEBUG "LIB database started");

  return 1;
}

static u32
krt_lib_hash(struct fib *f, void *a)
{
  return (*((u32 *)a) & 0xFFFFF);
}

/*
 * Hash protocol pointers
 */
static u32
krt_ptr_hash(struct fib *f, void *a)
{
  return (*((uintptr_t *)a) % 65521);
}

static krt_lib_proto *
krt_lib_find_proto(krt_lib_database *lib, struct proto *p)
{

  return (krt_lib_proto *)fib_find(&lib->pr, &p, 0);
}

/*
 * Free LIB database.
 */
void
krt_lib_free(struct krt_proto *k, int last UNUSED)
{
  krt_lib_database *lib = k->lib;
  krt_lib_free_label *fl, *fl_next;

  if (!lib)
	  return;

  /* Stop serving requests */
  proto_del_service_hook(lib->sh);

  /* Free unused items list */
  WALK_LIST_DELSAFE(fl, fl_next, lib->free_items)
    {
      rem_node(NODE fl);
      mb_free(fl);
    }

  /* Free database fib/itself */
  fib_free(&lib->f);
  fib_free(&lib->pr);
  mb_free(lib);
  k->lib = NULL;
}

/*
 * Returns 0 if database can be
 * closed immediately
 */
int
krt_lib_opened(struct krt_proto *k)
{
  krt_lib_database *lib = k->lib;

  return (lib ? lib->pr_count : 0);
}

/*
 * Notify every customer DB is going down.
 */
int
krt_lib_notify_down(struct krt_proto *k)
{
  krt_lib_database *lib = k->lib;
  krt_lib_proto *lp;
  struct service_reply repl;

  /* Protocol is not started or database is not used */
  if (!lib || !lib->pr_count)
    return 0;

  /* Prepare common fields for repl */
  repl.request = SRCLASS_LABEL;
  repl.subclass = SRSUB_LABEL_CLOSEDB;
  repl.t = k->mpls_table;
  repl.p = &k->p;

  /*
   * We need to notify every subscribed protocol
   * that we are going down. Typically protocol should shut down
   * itself do withdraw allocated labels (if any) and call
   * SRSUB_LABEL_CLOSEDB function which can lead us to possible
   * use-after-free fib walk after last protocol calls.
   * We handle this by running our shutdown in delayed event.
   */
  FIB_WALK(&lib->pr, ltmp)
    {
      lp = (krt_lib_proto *)ltmp;
      lp->relfunc(&k->p, lp->proto, &repl);
    }
  FIB_WALK_END;

  log(L_DEBUG "LIB shutdown: notified %d protocol(s)", 1);

  return 1;
}

/**
 * krt_lib_add_label - Allocate new LDP label of given type from LDP database
 * @pl: pointer to LDP protocol
 * @type: LIB entry type
 * @lp: pointer to LIB protocol record
 * @adata: Additional data for given @type if necessary
 *
 * Returns: krt_lib_label structure or NULL in case of failure
 */
static u32
krt_lib_add_label(struct krt_proto *k, krt_lib_proto *lp, struct krt_mpls_label_request *lreq)
{
  u32 label = 0;
  krt_lib_label *ll;
  krt_lib_database *lib = k->lib;
  krt_lib_free_label *fl, *fl_next;

  /* 
   * Find free label number to allocate
   * Labels are allocated in ascending order.
   * We keep track of maximum allocated number and
   * all free labels in pl->label_min .. lib->current_label - 1 space
   *
   * On label allocation request we try to eliminate all free items at first
   * (taking care on possible changed config label space)
   * IFF free items list is empty we try to increment current_label counter
   *
   * On label release we check if we can decrement current_label counter 
   * (in case of last allocated label is released) and in case of failure - 
   * push free label in free label list
   */

  /* Check if configured label space has changed */
  if (lib->current_label < k->label_min)
    lib->current_label = k->label_min;

  if (!EMPTY_LIST(lib->free_items))
    {
      /* Walk through free labels list */
      WALK_LIST_DELSAFE(fl, fl_next, lib->free_items)
	{
	  if ((fl->label < k->label_min) || (fl->label > k->label_max))
	    {
	      /* Not available in current labelspace */
	      rem_node(NODE fl);
	      mb_free(fl);
	      continue;
  	    }

	  /* Found */
	  label = fl->label;
	  rem_node(NODE fl);
	  mb_free(fl);
	  log(L_DEBUG "Found label %d in empty space", label);
	  break;
  	}
    }

  /* If label is still not found => free items list is (now) empty */
  if (label == 0)
    {
      if (lib->current_label > k->label_max)
	{
  	  log(L_ERR "No free labels left, consider increasing label_max");
  	  return 0;
	}

  	label = lib->current_label++;
    }

  log(L_DEBUG "Allocated type/%s label %d", krt_print_mplsop(lreq->op), label);

  /* Request label from FIB */
  ll = fib_get(&lib->f, &label, 0);

  /* Fill in all data we know */
  ll->type = 0;
  ll->me.action = lreq->op;
  ll->me.depth = 1;
  ll->me.local_label = label;

  /* Increase label count */
  lp->labels++;
  
  return label;
}

/**
 * krt_lib_del_label - Deletes label from label database
 * @pl: pointer to protocol 
 * @label: MPLS label number
 * @p: protocol instance label is allocated by
 *
 * Finds and deletes label from LDP database
 * SIDE EFFECTS:
 * Depending on label types, label structure is removed from appropriate lists, too
 *
 */
static int
krt_lib_del_label(struct krt_proto *k, krt_lib_proto *lp, u32 label)
{
  krt_lib_label *ll;
  krt_lib_free_label *fl;
  krt_lib_database *lib = k->lib;

  log(L_DEBUG "Freeing label %u", label);

  /* Kernel L3vpn label window check */
  if ((k->l3bhole_min > 0) && (k->l3bhole_min <= label) && (k->l3bhole_max >= label))
  {
  	KRT_TRACE(k, D_ROUTES, "Removing kernel VRF label %u", label);
  	return 1;
  }

  if ((ll = fib_find(&lib->f, &label, 0)) == NULL)
  {
  	KRT_TRACE(k, D_ROUTES, "Label %u not found in LIB", label);
  	return 0;
  }

  /* Decrease labels used by protocol */
  lp->labels--;

  /* Assume label is already removed from other lists and we can safely remove it */
  fib_delete(&lib->f, ll);

  /* Save label to free list for faster allocation */
  log(L_DEBUG "%d < %d < %d, current label is %d", k->label_min, label, k->label_max, lib->current_label);

  /* If label is outside of configured limit scope, do nothing */
  if ((label < k->label_min) || (label > k->label_max))
  	return 1;

  /* Check if re're freeing last allocated label */
  if (label + 1 == lib->current_label)
  {
  	--lib->current_label;
  	return 1;
  }

  KRT_TRACE(k, D_ROUTES, "Adding label %u to free list", label);
  /* Add label to free list */
  fl = mb_allocz(k->p.pool, sizeof(krt_lib_free_label));
  fl->label = label;
  add_tail(&lib->free_items, NODE fl);

  return 1;
}

void
krt_lib_expire_label(struct krt_proto *k, rte *old, rte *new)
{
  rta *a;
  eattr *ea;
  struct adata *ad;
  u32 old_label = 0, new_label = 0;
  bird_mpls_entry *me;
  krt_lib_proto *lp;

  if (!old || (!(a = old->attrs)))
    return;
  /* Ignore pipelike protocols */
  if (old->sender->proto != a->proto)
    return;
  /* Ignore routes without mandatory attribute */
  if (!(ea = ea_find(a->eattrs, EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL))))
    return;

  ad = ea->u.ptr;
  me = (bird_mpls_entry *)(ad + 1);

  old_label = me->local_label;

  /* Ignore 'special' labels */
  if (old_label < MPLS_LABEL_MIN)
    return;

  /* Check for new label if exists */
  if (new && (a = new->attrs) && (ea = ea_find(a->eattrs, EA_CODE(EAP_KRT, EA_KRT_MPLS_LABEL))))
    {
      ad = ea->u.ptr;
      me = (bird_mpls_entry *)(ad + 1);
      new_label = me->local_label;
    }

  if (old_label == new_label)
    return;

  /* Check if originator has opened our LIB. */
  if (!(lp = krt_lib_find_proto(k->lib, old->sender->proto)))
    return;

    krt_lib_del_label(k, lp, old_label);
}

static int
krt_dispatch_service_request(struct proto *p, struct service_request *req, byte **buf, size_t *buflen)
{
  struct krt_proto *k = (struct krt_proto *) p;
  krt_lib_database *lib = k->lib;
  krt_lib_proto *lp;
  int directed = req->target ? 1 : 0;
  int error = SR_FAIL;
  u32 label;
  struct krt_mpls_label_request *lreq;
  struct krt_mpls_label_reply *lrepl;
  struct service_reply *repl;

  KRT_TRACE(k, D_EVENTS, "Service request: %d/%d from protocol %s",
    req->request, req->subclass, req->p->name);

  if (req->request != SRCLASS_LABEL)
    return (directed ? SR_FAIL : SR_NEXT);

  switch (req->subclass)
    {
    case SRSUB_LABEL_OPENDB:
      /* Dispatch open database request */

      /* Deny opens if we're going down */
      if (k->shutdown_ev)
	break;

      /* work with supplied buffer only */
      if ((*buf == NULL) || (*buflen < sizeof(struct service_reply)))
	break;

      lp = (krt_lib_proto *)fib_get(&lib->pr, &req->p, 0);
      lp->proto = req->p;
      lp->relfunc = req->relfunc;
      lp->labels = 0;
      lib->pr_count++;
      KRT_TRACE(k, D_EVENTS, "LIB database opened by %s [%p] [%p]",
	req->p->name, req->p, lp);

      /* Fill in reply */
      repl = (struct service_reply *)(*buf);
      repl->p = &k->p;
      *buflen = sizeof(struct service_reply);

      error = SR_OK;
      break;

    case SRSUB_LABEL_CLOSEDB:
      /* Dispatch close database request */
      lp = krt_lib_find_proto(lib, req->p);
      log(L_ERR "LIB close request from %p (found %p)", req->p, lp);
      if (!lp)
	return (directed ? SR_FAIL : SR_NEXT);
      /* This can happen on daemon shutdown or krt restart */
      if (lp->labels > 0)
	  log(L_DEBUG "LIB close received from protocol %s while holding %d label(s)",
	    req->p->name, lp->labels);

      fib_delete(&lib->pr, lp);
      lib->pr_count--;
      KRT_TRACE(k, D_EVENTS, "LIB database closed by %s", req->p->name);
      error = SR_OK;

      /*
       * Check if we're going down and last consumer has closed database.
       * Call shutdown finalization if true.
       */
      if (k->shutdown_ev && !lib->pr_count)
	ev_schedule(k->shutdown_ev);
      break;

    case SRSUB_LABEL_REQ:
      /* Dispatch label request */

      KRT_TRACE(k, D_EVENTS, "subclass: Label request");

      /* work with supplied buffer only */
      if ((*buf == NULL) || (*buflen < sizeof(struct krt_mpls_label_reply)))
	break;

      lreq = (struct krt_mpls_label_request *)req;
      lp = krt_lib_find_proto(lib, lreq->req.p);

      if (!lp)
	return (directed ? SR_FAIL : SR_NEXT);

      if ((label = krt_lib_add_label(k, lp, lreq)) == 0)
	break;

      /* Increase label count */

      /* Fill in reply */
      lrepl = (struct krt_mpls_label_reply *)(*buf);
      lrepl->repl.p = &k->p;
      lrepl->label = label;
      *buflen = sizeof(struct krt_mpls_label_reply);

      KRT_TRACE(k, D_ROUTES, "Allocated label %u", label);
      error = SR_OK;
      break;

    }

  /* Do not return fatal error for non-unicast request */
  if ((error == SR_FAIL) && !directed)
    error = SR_NEXT;

  return error;
}

char *
krt_print_mplsop(enum kmpls_op op)
{
  char *buf;
  static char xbuf[50];

  switch (op)
  {
    case KMPLS_OP_NONE:
      buf = "NONE";
      break;

    case KMPLS_OP_POP:
      buf = "POP";
      break;

    case KMPLS_OP_PUSH:
      buf = "PUSH";
      break;

    case KMPLS_OP_SWAP:
      buf = "SWAP";
      break;

    case KMPLS_OP_L3VPN:
      buf = "L3VPN";
      break;

    case KMPLS_OP_NETGRAPH:
      buf = "NETGRAPH";
      break;

    default:
      bsnprintf(xbuf, sizeof(xbuf), "UNKN:%d", op);
      buf = xbuf;
  }

  return buf;
}

char *
krt_print_mplslbl(u32 label)
{
  char *buf;
  static char xbuf[50];

  switch(label)
  {
    case MPLS_LABEL_IPV4NULL:
      buf = "IPv4NULL";
      break;

    case MPLS_LABEL_IPV6NULL:
      buf = "IPv6NULL";
      break;

    case MPLS_LABEL_IMPLNULL:
      buf = "ImpNULL";
      break;

    default:
      buf = xbuf;

      if (label < MPLS_LABEL_RESERVED_MAX)
      {
        bsnprintf(xbuf, sizeof(xbuf), "Reserved(%d)", label);
        break;
      }

      bsnprintf(xbuf, sizeof(xbuf), "%d", label);
  }

  return buf;
}


