/*
 *	BIRD -- Label distribution protocol
 *
 *	(c) 2011 Alexander V. Chernikov <melifaro@ipfw.ru>
 *
 *	Can be freely distributed and used under the terms of the 
 *	GNU GPL or Simplified BSD License.
 */

/*
 * Low-level PDU/Message parsing/generating functions.
 *
 */

#define LOCAL_DEBUG

#include "ldp.h"
#include "ldp_proto.h"

static int ldp_message_dispatch(ldp_neighbour *ln, ldp_message_header *lmh, byte *data, int msgtype, void *xptr);

/**
 * ldp_pdu_get_header - converts PDU header to host format
 * @data: pointer to unaligned buffer with PDU header in network format
 * @lph: pointer to PDU header structure
 */
void
ldp_pdu_get_header(byte *data, ldp_pdu_header *lph)
{
	lph->version = get_u16(data);
	lph->length = get_u16(data + 2);
	lph->lsr_id = get_u32(data + 4);
	lph->space_id = get_u16(data + 8);
}

/**
 * ldp_pdu_put_header - Fills PDU header in given buffer
 * @data - pointer to buffer need to be filled in
 * @length - length of data inside PDU
 * @space_id - Messages label space id
 *
 * Fills in length, LSR id
 * Returns total PDU length
 */
int
ldp_pdu_put_header(byte *data, int length, int space_id)
{
	put_u16(data, LDP_PROTOCOL_VERSION);
	put_u16(data + 2, length + 6); /* sizeof(Lsr_id) */
	put_u32(data + 4, config->router_id);
	put_u16(data + 8, space_id);

	return length + LDP_PDU_HEADER_SIZE;
}

/**
 * ldp_message_get_header - converts message header to host format
 * @data: pointer to unaligned buffer with message header in network format
 * @lph: pointer to message header structure
 */
void
ldp_message_get_header(byte *data, ldp_message_header *lmh)
{
	u8 x1, x2;

	x1 = *((u8 *)data);
	x2 = *((u8 *)(data + 1));
	lmh->unknown = x1 >> 7;
	lmh->type = ((x1 & 0x7F) << 8) + x2;
	lmh->length = get_u16(data + 2);
	lmh->id = get_u32(data + 4);
}

/**
 * ldp_message_put_header - Fills message header in given buffer
 * @dara - pointer to header need to be filled in
 * @ls: pointer to LDP session or NULL
 * @unknown - unknown bit
 * @type - message type
 * @length - length of data inside message
 * @id - pointer to generated message unique id
 *
 * Fills in all appropriate fields
 * Returns total message length
 */
int
ldp_message_put_header(byte *data, ldp_session *ls, int unknown, int type, int length, int *id)
{
	u8 *x = data;

	*x++ = ((unknown & 0x01) << 7) | (type >> 8);
	*x = type & 0xFF;
	put_u16(data + 2, length + 4); /* sizeof(Message ID)*/
	/* 
	 * XXX: We should save id <> lmh bindings here for some time to permit 
	 * remote notifictions dispatching
	 */
	put_u32(data + 4, ls ? ls->seq++ : 0);
	
	return length + LDP_MESSAGE_HEADER_SIZE;
}


/**
 * ldp_pdu_preprocess - read PDUs from given buffer
 * @pl - pointer to protocol instance
 * @remote - neighbour address used in debug messages
 * @maxlength - maximum PDU length (i.e. LDP_MAXPDU_INITIAL for hello)
 * @sock_data - pointer to data received from socket
 * @sock_size - amount of data received from socket
 * @prev_data - pointer to buffer with previously read data
 * @prev_size - buffer size
 * @prev_off - data offset in buffer (amount of data already in buffer)
 * @read_size - amount of data read will be put here
 *
 * Reads PDUs from socket.
 * Returns 1 in case of complete PDU in buffer,
 * 0 in case of successful read
 * -1 in case of error
 * read_size is set in all cases
 */
int
ldp_pdu_preprocess(struct proto_ldp *pl, ip_addr remote, int maxlength, char *sock_data, int sock_size, char *prev_data, int prev_size, int prev_off, int *read_size)
{
	ldp_pdu_header lph;
	int length = 0;

	if (prev_off < LDP_PDU_HEADER_SIZE)
	{
		/* Try to read header */
		length = LDP_PDU_HEADER_SIZE - prev_off;

		if (length > sock_size)
		{
			/* Some probe/exploit? */
			memcpy(&prev_data[prev_off], sock_data, sock_size);
			*read_size = sock_size;
			return 0;
		}

		/* Read header remainder */
		memcpy(&prev_data[prev_off], sock_data, length);
		/* Do not update counters, this will break code after if {} ending */
	}

	/* We have full LDP header read in lh->data. Check now much can we read */
	ldp_pdu_get_header(prev_data, &lph);

	/* Check version */
	if (lph.version != LDP_PROTOCOL_VERSION)
	{
		LOG_VRB("Unsupported LDP version %d in PDU from %I", lph.version, remote);
		*read_size = length;
		return -1;
	}

	/* Check size */
	if (lph.length > maxlength) {
		LOG_DBG("Skipping PDU from %I: length %d is too big", remote, lph.length);
		*read_size = length;
		return -1;
	}

	/* Determine how much data is remaining */
	length = lph.length + 4 - prev_off; /* Include version and length fields */
	if (length <= sock_size)
	{
		/* Most common case, 1 read == 1 PDU */
		memcpy(&prev_data[prev_off], sock_data, length);
		*read_size = length;
		return 1;
	}
	else
	{
		/* 
		 * Should not happen for hello 
		 * XXX: We should check bird socket api
		 * for this case
		 */
		memcpy(&prev_data[prev_off], sock_data, sock_size);
		*read_size = sock_size;
		return 0;
	}

	/* MOTREACHED */
}

/**
 * ldp_pdu_dispatch - begin PDU message processing
 * @ln:  pointer to LDP neighbour
 * @data: pointer to PDU data
 * @length: data length
 * @type: PDU type (session/hello)
 * @xptr: type-dependent pointer (ldp_session or ldp_hello)
 *
 * Reads messages from PDU supplied in @data. 
 *
 */
void
ldp_pdu_dispatch(ldp_neighbour *ln, byte *data, int length, int msgtype, void *xptr)
{
	int off, mlength, dtype;
	ldp_message_header lmh;
	int space_id, res;
	byte *msgdata;

	ldp_session *ls = NULL;
	ldp_hello *lh = NULL;

	if (msgtype == LDP_PDU_SESSION)
	{
		ls = (ldp_session *)xptr;
		space_id = ls->space_id;

		dtype = LD_SESSION;
		CLOG_DBG(dtype, "Got session %R:%d PDU size %d", ln->lsr_id, space_id, length);

		/* Restart session hold timer on valid PDU reception */
		tm_stop(ls->hold_timer);
		tm_start(ls->hold_timer, ls->keep_interval);
	}
	else
	{
		/* This is hello PDU */
		lh = (ldp_hello *)xptr;
		space_id = lh->space_id;
		dtype = LD_HELLO;
		
		CLOG_DBG(dtype, "Got hello %R:%d PDU size %d from %I", ln->lsr_id, space_id, length, lh->remote);
	}

	/* Set appropriate variables for TDUMP() to work */
	DEBUG_PACKET((msgtype == LDP_PDU_SESSION) ? LD_SESSION : LD_HELLO);

	for (off = LDP_PDU_HEADER_SIZE, msgdata = &data[off]; off < length; off += mlength, msgdata += mlength)
	{
		if (off + LDP_MESSAGE_HEADER_SIZE > length)
		{
			CLOG_DBG(dtype, "Too short PDU from %R:*", ln->lsr_id);
			break;
		}

		ldp_message_get_header(msgdata, &lmh);
		mlength = lmh.length + 4; /* type & length is not accounted in length field */

		if ((mlength + off > length) || (mlength < LDP_MESSAGE_HEADER_SIZE))
		{
			CLOG_DBG(dtype, "Invalid %s message length %d in PDU from %R:%d", ldp_message_texttype(lmh.type), mlength, ln->lsr_id, space_id);
			break;
		}

		if ((res = ldp_message_dispatch(ln, &lmh, msgdata, msgtype, xptr)) != 0)
		{
			CLOG_VRB(dtype, "Error parsing message %s length %d from %R:%d", ldp_message_texttype(lmh.type), mlength, ln->lsr_id, space_id);
			break;
		}
	}

	UNDEBUG_PACKET();
}


int
ldp_message_parse_tlvs(struct proto_ldp *pl, ldp_neighbour *ln, ldp_message_header *lmh, byte *data, list *tlvs, list *fwd, int dtype)
{
	int length, off, res, tlength;
	ldp_tlv_header lth;
	ldp_tlv_forward *tlf;
	ldp_tlv *tlv;

	/* Get all TLVs from message */
	for (off = LDP_MESSAGE_HEADER_SIZE, length = lmh->length + 4; off < length; off += tlength)
	{
		ldp_tlv_get_header(&data[off], &lth);

		tlength = lth.length + LDP_TLV_HEADER_SIZE;
		if ((tlength + off > length) || (tlength < 4))
		{
			CLOG_VRB(dtype, "Incorrect TLV data in %s message from neighbour %I",
				ldp_message_texttype(lmh->type), ln->remote);

			return LDP_ERR;
		}

		/* Put TLV on forwarding list if it is forwardable */
		if (lth.unknown && lth.forward)
		{
			tlf = lp_alloc(pl->pdu_lp, BIRD_ALIGN(sizeof(ldp_tlv_forward), CPU_STRUCT_ALIGN) + tlength);
			tlf->type = lth.type;
			tlf->length = lth.length;
			memcpy((byte *)tlf + BIRD_ALIGN(sizeof(ldp_tlv_forward), CPU_STRUCT_ALIGN), &data[off], tlength);
			add_tail(fwd, NODE tlf);
		}

		res = ldp_tlv_parse(pl, &lth, &data[off], tlength, &tlv);

		/* Check parse result */
		switch (res)
		{
			case LDP_OBJ_FAIL:
				CLOG_VRB(dtype, "Unknown/mailformed TLV %s received from %I, ignoring message",
					ldp_tlv_texttype(lth.type), ln->remote);
				/* XXX: Send NOTIFICATION message */
				return LDP_ERR;
			case LDP_OBJ_IGNORE:
				CLOG_DBG(dtype, "Unknown TLV %X received from %I, ignoring locally",
					ldp_tlv_texttype(lth.type), ln->remote);
				continue;
		}

		/*
		 * Parsing succeeded, we've got converted header (allocated in linpool) in
		 * *header. Add it to list 
		 */
		add_tail(tlvs, &tlv->n);

		CLOG_DBG(dtype, "Got TLV %s(0x%X) size %d", ldp_tlv_texttype(lth.type), lth.type, lth.length);
	}

	return LDP_OK;
}


/**
 * ldp_message_dispatch - dispatches new LDP message
 * @ln: pointer to LDP neighbour
 * @lmh: pointer to converted message header
 * @data: pointer to data buffer
 * @msgtype: message type (session or hello)
 * @xptr: msgtype-dependent data (ldp_session or ldp_hello)
 *
 * Function check general message validness for given communication type,
 * makes list of interpreted TLVs and list of unmodified TLVs to forward.
 * In case of no error all this data is passed to message-specific handler
 *
 * Function returns 0 in case of no error
 *
 */
static int
ldp_message_dispatch(ldp_neighbour *ln, ldp_message_header *lmh, byte *data, int msgtype, void *xptr)
{
	struct proto_ldp *pl = ln->pl;
	int res, dtype, error = 0;
	int space_id;
	list tlv_list, forward_list;
	ldp_session *ls = NULL;
	ldp_hello *lh = NULL;

	if (msgtype == LDP_PDU_HELLO)
	{
		dtype = LD_HELLO;
		lh = (ldp_hello *)xptr;
		space_id = lh->space_id;

		if (lmh->type != LDP_MSG_HELLO)
		{
			LOG_HL_DBG("Message type %s received in hello PDU from %I, ignoring", ldp_message_texttype(lmh->type), ln->remote);
			return 1;
		}
	}
	else
	{
		dtype = LD_SESSION;
		ls = (ldp_session *)xptr;
		space_id = ls->space_id;

		if (lmh->type == LDP_MSG_HELLO)
		{
			LOG_SS_DBG("Message type %s received in session PDU from %I, ignoring", ldp_message_texttype(lmh->type), ln->remote);
			return 1;
		}
	}

	CLOG_DBG(dtype, "Got LDP message type %s (0x%X) size %d msgid %u from %I (lsr %R:%d)", 
			ldp_message_texttype(lmh->type), lmh->type, lmh->length, lmh->id, 
			ln->remote, ln->lsr_id, space_id);

	/* Dump packet header if requested */
	if (DUMP_PACKET)
	{
		TDUMP("===============================================================");
		TLINE();
		TDUMP("LDP message type %s (0x%X) size %d msgid %u from %I (lsr %R:%d)", 
				ldp_message_texttype(lmh->type), lmh->type, lmh->length, lmh->id, 
				ln->remote, ln->lsr_id, space_id);
	
		TLINE();
	}

	/* Init TLV interpreted and forwarding lists */
	init_list(&tlv_list);
	init_list(&forward_list);

	if (ldp_message_parse_tlvs(pl, ln, lmh, data, &tlv_list, &forward_list, dtype) != LDP_OK)
	{
		error = 1;
		goto done;
	}

	/* Increase received message counter for session */
	if (ls != NULL)
		ls->messages_in++;

	/* 
	 * We've got
	 * 1) Unknown TLVs to forward in forward_list (entirely in stack)
	 * 2) Host-format TLVs in tlv_list (allocated from linpool)
	 *
	 * Check for valid msgtype for given communication type and
	 * ls/lh pointer validness are done at the beginning of this
	 * function
	 *
	 * Let's call specific message dispatcher from here
	 */
	switch (lmh->type)
	{
		case LDP_MSG_HELLO:
			res = ldp_dispatch_hello(ln, lh, lmh, &tlv_list);
			break;

		case LDP_MSG_INIT:
			res = ldp_dispatch_session_init(ln, ls, lmh, &tlv_list);
			break;

		case LDP_MSG_KEEPALIVE:
			res = ldp_dispatch_session_keepalive(ln, ls, lmh, &tlv_list);
			break;

		case LDP_MSG_ADDR:
		case LDP_MSG_ADDRWITH:
			res = ldp_dispatch_session_addrlist(ln, ls, lmh, &tlv_list);
			break;

		case LDP_MSG_LMAP:
		case LDP_MSG_LWITH:
			res = ldp_dispatch_session_lmap(ln, ls, lmh, &tlv_list, &forward_list);
			break;

		case LDP_MSG_LREQ:
			res = ldp_dispatch_session_lreq(ln, ls, lmh, &tlv_list, &forward_list);
			break;

		case LDP_MSG_LREL:
			res = ldp_dispatch_session_lrel(ln, ls, lmh, &tlv_list, &forward_list);
			break;

		case LDP_MSG_NOTIFICATION:
			res = ldp_dispatch_session_notify(ln, ls, lmh, &tlv_list, &forward_list);
			break;

		default:
			if (lmh->type < 0x8000)
			{
				/* Error */
				CLOG_DBG(dtype, "Unknown message type: %X received from %I",
					lmh->type, ln->remote);
			}

			if (lmh->type >= 0x8000)
			{
				/* discard */
				CLOG_DBG(dtype, "Unknown message type: %X received from %I",
					lmh->type, ln->remote);
			}
	}

	if (res != LDP_STATUS_SUCCESS)
	{
		/* We should prepare a reply */
	}

	/* Finally, free allocated structures */

done:
	if (DUMP_PACKET)
	{
		TDUMP("===============================================================");
		TLINE();
	}

	/* Flush linear pool */
	lp_flush(pl->pdu_lp);

	return error;
}



char *
ldp_message_texttype(int type)
{
	char *buf;
	static char xbuf[50];


	switch(type)
	{
		case LDP_MSG_NOTIFICATION:
			buf = "Notification";
			break;
		case LDP_MSG_HELLO:
			buf = "Hello";
			break;
		case LDP_MSG_INIT:
			buf = "Initialization";
			break;
		case LDP_MSG_KEEPALIVE:
			buf = "KeepAlive";
			break;
		case LDP_MSG_ADDR:
			buf = "Address";
			break;
		case LDP_MSG_ADDRWITH:
			buf = "Address Withdraw";
			break;
		case LDP_MSG_LMAP:
			buf = "Label Mapping";
			break;
		case LDP_MSG_LREQ:
			buf = "Label Request";
			break;
		case LDP_MSG_LWITH:
			buf = "Label Withdraw";
			break;
		case LDP_MSG_LREL:
			buf = "Label Release";
			break;
		case LDP_MSG_LABRT:
			buf = "Label Abort Req";
			break;

		default:
			buf = xbuf;

			if ((type >= 0x3E00) && (type <= 0x3EFF))
			{
				snprintf(xbuf, sizeof(xbuf), "Vendor-Private: 0x%X", type);
				break;
			}

			if ((type >= 0x3F00) && (type <= 0x3FFF))
			{
				snprintf(xbuf, sizeof(xbuf), "Experimental: 0x%X", type);
				break;
			}

			snprintf(xbuf, sizeof(xbuf), "Unknown: 0x%X", type);
	}

	return buf;
}


